#include "RVLVTK.h"
#include "RVLCore2.h"
#include "Util.h"
#include "Space3DGrid.h"
#include "Graph.h"
#include "Mesh.h"
#include "Visualizer.h"
#include "SceneSegFile.hpp"
#include "SurfelGraph.h"
#include "ObjectGraph.h"
#include "PlanarSurfelDetector.h"
#include "RVLRecognition.h"
#include "PSGMCommon.h"
#include "CTISet.h"
#include "VertexGraph.h"
#include "TG.h"
#include "TGSet.h"
#include "PSGM.h"
#include "ObjectDetector.h"
#include <Eigen\Eigenvalues>
#include "VN.h"
#include "VNClassifier.h"
#include "VNLocalizer.h"

using namespace RVL;

VNLocalizer::VNLocalizer()
{
}


VNLocalizer::~VNLocalizer()
{
}

void VNLocalizer::CreateParamList()
{

}

void VNLocalizer::Create(char *cfgFileName)
{
	// Set parameters.

	CreateParamList();

	paramList.LoadParams(cfgFileName);
}

void VNLocalizer::UpdateMap(Mesh *pMesh)
{
	// Visualization

	Visualizer *pVisualizer = NULL;

	bool bVisualizeSurfels = true;

	uchar SelectionColor[] = { 0, 255, 0 };

	bool bVisualizerCallbackFunctionsDefined;

	//if (bVisualizeConvexClusters || bVisualizeConcaveClusters || bVisualizeSurfels || bVisualizeAllClusters)
	if (bVisualizeSurfels)
	{
		pVisualizer = new Visualizer;

		pVisualizer->Create();

		bVisualizerCallbackFunctionsDefined = pSurfels->DisplayData.bCallbackFunctionsDefined;

		pSurfels->DisplayData.bCallbackFunctionsDefined = false;
	}

	// Detect surfels.

	pSurfels->Init(pMesh);

	pSurfelDetector->Init(pMesh, pSurfels, pMem);

	printf("Segmentation to surfels...");

	pSurfelDetector->Segment(pMesh, pSurfels);

	printf("completed.\n");

	int nSurfels = pSurfels->NodeArray.n;

	printf("No. of surfels = %d\n", nSurfels);

	pSurfels->DetectVertices(pMesh);

	pSurfels->DetectOcclusionVertices(pMesh, camera);

	// Surfel visualization.

	if (bVisualizeSurfels)
	{
		pSurfels->NodeColors(SelectionColor);

		pSurfels->InitDisplay(pVisualizer, pMesh, pSurfelDetector);

		pSurfels->Display(pVisualizer, pMesh);

		pSurfels->DisplayEdgeFeatures();

		pVisualizer->Run();

		pVisualizer->renderer->RemoveAllViewProps();
	}

	// Delete visualizer.

	if (pVisualizer)
	{
		delete pVisualizer;

		pSurfels->DisplayData.bCallbackFunctionsDefined = bVisualizerCallbackFunctionsDefined;
	}
}