// RVLLocalizationDemo.cpp : Defines the entry point for the console application.
//

#include <vtkAutoInit.h>
VTK_MODULE_INIT(vtkRenderingOpenGL2);
VTK_MODULE_INIT(vtkInteractionStyle);
VTK_MODULE_INIT(vtkRenderingFreeType);
#include "RVLVTK.h"
#include <vtkTriangle.h>		// Remove after completion of MarchingCubes.
#include "RVLCore2.h"
#include "Util.h"
#include "Space3DGrid.h"
#include "Graph.h"
#include "Mesh.h"
#include "Visualizer.h"
#include "SceneSegFile.hpp"
#include "SurfelGraph.h"
#include "ObjectGraph.h"
#include "PlanarSurfelDetector.h"
#include "RVLRecognition.h"
//#include "RFRecognition.h"
//#include "RVLMeshNoiser.h"
#include "PSGMCommon.h"
#include "CTISet.h"
#include "VertexGraph.h"
#include "TG.h"
#include "TGSet.h"
#include "PSGM.h"
#include <pcl/common/common.h>
//#include <pcl/registration/registration.h>
#include <pcl/PolygonMesh.h>
//#include <pcl/io/openni2_grabber.h>
//#include <pcl/io/openni2/openni2_metadata_wrapper.h>
#include "PCLTools.h"
#include "RGBDCamera.h"
#include "PCLMeshBuilder.h"
#include "MarchingCubes.h"
#include "VN.h"
#include "VNClassifier.h"
#include "VNLocalizer.h"


#define RVLLOCALIZATION_DEMO_FLAG_SAVE_PLY				0x00000001

using namespace RVL;

void CreateParamList(
	CRVLParameterList *pParamList,
	CRVLMem *pMem,
	char **pSceneSequenceFileName,
	char **pResultsFolder,
	DWORD &flags,
	Camera &camera)
{
	pParamList->m_pMem = pMem;

	RVLPARAM_DATA *pParamData;

	pParamList->Init();

	pParamData = pParamList->AddParam("SceneSequenceFileName", RVLPARAM_TYPE_STRING, pSceneSequenceFileName);
	pParamData = pParamList->AddParam("ResultsFolder", RVLPARAM_TYPE_STRING, pResultsFolder);
	pParamData = pParamList->AddParam("Save PLY", RVLPARAM_TYPE_FLAG, &flags);
	pParamList->AddID(pParamData, "yes", RVLLOCALIZATION_DEMO_FLAG_SAVE_PLY);
	pParamData = pParamList->AddParam("Camera.fu", RVLPARAM_TYPE_FLOAT, &(camera.fu));
	pParamData = pParamList->AddParam("Camera.fv", RVLPARAM_TYPE_FLOAT, &(camera.fv));
	pParamData = pParamList->AddParam("Camera.uc", RVLPARAM_TYPE_FLOAT, &(camera.uc));
	pParamData = pParamList->AddParam("Camera.vc", RVLPARAM_TYPE_FLOAT, &(camera.vc));
	pParamData = pParamList->AddParam("Camera.w", RVLPARAM_TYPE_INT, &(camera.w));
	pParamData = pParamList->AddParam("Camera.h", RVLPARAM_TYPE_INT, &(camera.h));
}

int main(int argc, char ** argv)
{
	// Create memory storage.

	CRVLMem mem0;	// permanent memory

	mem0.Create(1000000000);

	CRVLMem mem;	// cycle memory

	mem.Create(1000000000);

	// Read parameters from a configuration file.

	char cfgSelectionFileName[] = "RVLLocalizationDemo.cfg";

	char *cfgFileName = ReadConfigurationFile(cfgSelectionFileName);

	if (cfgFileName == NULL)
		return 1;

	char *sceneSequenceFileName = NULL;
	char *resultsFolder = NULL;
	Camera camera;

	DWORD flags = 0x00000000;

	CRVLParameterList ParamList;

	CreateParamList(&ParamList,
		&mem0,
		&sceneSequenceFileName,
		&resultsFolder,
		flags,
		camera);

	ParamList.LoadParams(cfgFileName);

	// Create mesh builder.

	PCLMeshBuilder meshBuilder;

	meshBuilder.CreateParamList(&mem0);

	meshBuilder.ParamList.LoadParams(cfgFileName);

	meshBuilder.camera.depthFu = camera.fu;
	meshBuilder.camera.depthFv = camera.fv;
	meshBuilder.camera.depthUc = camera.uc;
	meshBuilder.camera.depthVc = camera.vc;

	pcl::PointCloud<pcl::PointXYZRGBA>::Ptr PC(new pcl::PointCloud<pcl::PointXYZRGBA>(camera.w, camera.h));

	meshBuilder.PC = PC;

	Mesh mesh;

	// Create file sequence handleer.

	FileSequenceLoader sceneSequence;
	char filePath[200];

	sceneSequence.Init(sceneSequenceFileName);

	// Generate scene PLY files.

	if (flags & RVLLOCALIZATION_DEMO_FLAG_SAVE_PLY)
	{
		Mesh mesh;

		while (sceneSequence.GetNextPath(filePath))
			LoadMesh(&meshBuilder, filePath, &mesh, true);

		RVL_DELETE_ARRAY(sceneSequenceFileName);
		RVL_DELETE_ARRAY(resultsFolder);

		return 0;
	}

	// Initialize surfel detection.

	SurfelGraph surfels;

	surfels.pMem = &mem;

	surfels.CreateParamList(&mem0);

	surfels.ParamList.LoadParams(cfgFileName);

	PlanarSurfelDetector surfelDetector;

	surfelDetector.CreateParamList(&mem0);

	surfelDetector.ParamList.LoadParams(cfgFileName);

	// Initialize localization.

	VNLocalizer localizer;

	localizer.pMem0 = &mem0;
	localizer.pMem = &mem;
	localizer.pSurfels = &surfels;
	localizer.pSurfelDetector = &surfelDetector;
	localizer.camera = camera;

	// Initialize visualization.

	unsigned char SelectionColor[3];

	SelectionColor[0] = 0;
	SelectionColor[1] = 255;
	SelectionColor[2] = 0;

	Visualizer visualizer;

	visualizer.Create();

	/// main loop

	while (sceneSequence.GetNextPath(filePath))
	{
		printf("Scene: %s:\n", filePath);

		// Load mesh.

		LoadMesh(&meshBuilder, filePath, &mesh, false);

		// Reset memory.

		mem.Clear();

		// Localization.

		localizer.UpdateMap(&mesh);
	}

	///
	
	// Free memory.

	RVL_DELETE_ARRAY(sceneSequenceFileName);
	RVL_DELETE_ARRAY(resultsFolder);

	return 0;
}

