// RVLMesher.cpp : Defines the entry point for the console application.
//

//#include "stdafx.h"
#include <vtkAutoInit.h>
//VTK_MODULE_INIT(vtkRenderingOpenGL);
VTK_MODULE_INIT(vtkRenderingOpenGL2);
VTK_MODULE_INIT(vtkInteractionStyle);
VTK_MODULE_INIT(vtkRenderingFreeType);
#include <vtkOBJReader.h>
#include "RVLVTK.h"
#include "RVLCore2.h"
#include "Util.h"
#include "Space3DGrid.h"
#include "Graph.h"
#include "Mesh.h"
#include "Visualizer.h"

#define RVLMESHER_FLAG_VISUALIZATION	0x00000001
#define RVLMESHER_METHOD_3DDELAUNAY		0x00000000
#define RVLMESHER_METHOD_VOXELRG		0x00000001

using namespace RVL;

void CreateParamList(
	CRVLParameterList *pParamList,
	CRVLMem *pMem,
	char **pSceneSequenceFileName,
	DWORD &method,
	DWORD &flags,
	float &SDFSurfaceValue,
	float &resolution,
	bool &bRoom)
{
	pParamList->m_pMem = pMem;

	RVLPARAM_DATA *pParamData;

	pParamList->Init();

	pParamData = pParamList->AddParam("SceneSequenceFileName", RVLPARAM_TYPE_STRING, pSceneSequenceFileName);
	pParamData = pParamList->AddParam("Method", RVLPARAM_TYPE_ID, &method);
	pParamList->AddID(pParamData, "3DDELAUNAY", RVLMESHER_METHOD_3DDELAUNAY);
	pParamList->AddID(pParamData, "VOXELRG", RVLMESHER_METHOD_VOXELRG);
	pParamData = pParamList->AddParam("Visualization", RVLPARAM_TYPE_FLAG, &flags);
	pParamList->AddID(pParamData, "yes", RVLMESHER_FLAG_VISUALIZATION);
	pParamData = pParamList->AddParam("Mesher.SDFSurfaceValue", RVLPARAM_TYPE_FLOAT, &SDFSurfaceValue);
	pParamData = pParamList->AddParam("Mesher.resolution", RVLPARAM_TYPE_FLOAT, &resolution);
	pParamData = pParamList->AddParam("Mesher.room", RVLPARAM_TYPE_BOOL, &bRoom);
}

int main(int argc, char ** argv)
{
	CRVLMem mem0;

	mem0.Create(10000000);

	char *sceneSequenceFileName = NULL;
	DWORD method;
	float SDFSurfaceValue = 0.0f;
	float resolution = 0.005f;
	DWORD flags = 0x00000000;
	bool bRoom = false;

	CRVLParameterList ParamList;

	CreateParamList(&ParamList,
		&mem0,
		&sceneSequenceFileName,
		method,
		flags,
		SDFSurfaceValue,
		resolution,
		bRoom);

	ParamList.LoadParams("RVLMesher.cfg");

	if (sceneSequenceFileName == NULL)
	{
		printf("No scene specified!\n");

		return 1;
	}

	MESH::Tetrahedrons tetrahedrons;

	Visualizer visualizer;

	visualizer.Create();

	FileSequenceLoader sceneSequence;

	sceneSequence.Init(sceneSequenceFileName);

	char filePath[200];
	Mesh mesh;

	while (sceneSequence.GetNextPath(filePath))
	{
		printf("Scene: %s:\n", filePath);

		if (method == RVLMESHER_METHOD_3DDELAUNAY)
		{
			MESH::LoadTetrahedrons(filePath, tetrahedrons);

			printf("Creating visible surface mesh...");

			mesh.pPolygonData = MESH::CreateVisibleSurfaceMeshFromTetrahedrons(tetrahedrons, resolution, SDFSurfaceValue, bRoom);
		}
		else if (method == RVLMESHER_METHOD_VOXELRG)
		{
			char *fileExtension = RVLGETFILEEXTENSION(filePath);

			if (strcmp(fileExtension, "obj") == 0)
			{
				vtkSmartPointer<vtkOBJReader> reader = vtkSmartPointer<vtkOBJReader>::New();
				reader->SetFileName(filePath);
				reader->Update();
				vtkSmartPointer<vtkPolyData> pPolygonDataSrc = reader->GetOutput();

				mesh.pPolygonData = MESH::CreateVisibleSurfaceMesh2(pPolygonDataSrc, resolution, SDFSurfaceValue);
			}

			else if (strcmp(fileExtension, "ply") == 0)
			{
			vtkSmartPointer<vtkPLYReader> reader = vtkSmartPointer<vtkPLYReader>::New();
			reader->SetFileName(filePath);
			reader->Update();
			vtkSmartPointer<vtkPolyData> pPolygonDataSrc = reader->GetOutput();

			mesh.pPolygonData = MESH::CreateVisibleSurfaceMesh2(pPolygonDataSrc, resolution, SDFSurfaceValue);
		}

		}

		printf("completed.\n");

		if (flags & RVLMESHER_FLAG_VISUALIZATION)
		{
			vtkSmartPointer<vtkPolyDataMapper> mapper = vtkSmartPointer<vtkPolyDataMapper>::New();
			mapper->SetInputData(mesh.pPolygonData);
			vtkSmartPointer<vtkActor> actor = vtkSmartPointer<vtkActor>::New();
			actor->SetMapper(mapper);

			visualizer.renderer->AddActor(actor);

			visualizer.Run();

			visualizer.renderer->RemoveAllViewProps();
		}

		char *PLYFileName = RVLCreateFileName(filePath, ".ply", -1, "_vs.ply");

		printf("Saving mesh to %s...", PLYFileName);

		mesh.SavePolyDataToPLY(PLYFileName);

		printf("completed.\n");

		delete[] PLYFileName;
	}

	if (sceneSequenceFileName)
		delete[] sceneSequenceFileName;

	return 0;
}

