//#include "stdafx.h"
#include "RVLVTK.h"
#include "RVLCore2.h"
#include "Util.h"
#include "Space3DGrid.h"
#include "Graph.h"
#include "MSTree.h"

using namespace RVL;
using namespace GRAPH;

MSTree::MSTree()
{
	memSize = 1000;

	T.NodeArray.n = 0;

	E = NULL;
	Q = NULL;
	bProcessed = NULL;
	T_ = NULL;
}


MSTree::~MSTree()
{
	Clear();
}

void MSTree::Init(int memSizeIn)
{
	memSize = memSizeIn;

	Clear();

	RVL_DELETE_ARRAY(T.NodeMem);
	RVL_DELETE_ARRAY(T.EdgeMem);
	RVL_DELETE_ARRAY(T.EdgePtrMem);

	T.NodeMem = new Node[memSize];
	T.NodeArray.Element = T.NodeMem;
	T.EdgeMem = new Edge[memSize];
	T.EdgePtrMem = new EdgePtr<Edge>[2 * memSize];
	E = new float *[memSize];
	Q = new UpdateMSTreeInsertData[memSize];
	bProcessed = new bool[memSize];
	T_ = new Pair<int, int>[memSize];
}

void MSTree::Clear()
{
	int i;

	for (i = 1; i < T.NodeArray.n; i++)
		delete[] E[i];
	RVL_DELETE_ARRAY(E);
	RVL_DELETE_ARRAY(Q);
	RVL_DELETE_ARRAY(bProcessed);
	RVL_DELETE_ARRAY(T_);
	T.NodeArray.n = 0;
}

// Function UpdateMSTree is an implementation of the method proposed in chin_JCSS78.

void MSTree::Update(float *ENew)
{
	Node *pNode;
	QList<EdgePtr<Edge>> *pEdgeList;

	if (T.NodeArray.n == 0)
	{
		T.NodeArray.n = 1;

		pNode = T.NodeArray.Element;

		pNode->idx = 0;

		pEdgeList = &(pNode->EdgeList);

		RVLQLIST_INIT(pEdgeList);

		return;
	}

	if (ENew == NULL)
	{
		T.NodeArray.Element[T.NodeArray.n].idx = T.NodeArray.n;

		pNode = T.NodeArray.Element + T.NodeArray.n;

		pEdgeList = &(pNode->EdgeList);

		RVLQLIST_INIT(pEdgeList);

		T.NodeArray.n++;

		return;
	}

	// Add ENew to E.

	bool bMemReallocation;
	Edge *newEdgeMem;
	EdgePtr<Edge> *newEdgePtrMem;

	if (bMemReallocation = (T.NodeArray.n == memSize))
	{
		int newMemSize = memSize * 2;

		float **newEBlock = new float *[newMemSize];

		memcpy(newEBlock, E, newMemSize * sizeof(float *));

		delete[] E;

		E = newEBlock;

		delete[] Q;

		Q = new UpdateMSTreeInsertData[newMemSize];

		delete[] bProcessed;

		bProcessed = new bool[newMemSize];

		delete[] T_;

		T_ = new Pair<int, int>[newMemSize];

		newEdgeMem = new Edge[newMemSize];

		newEdgePtrMem = new EdgePtr<Edge>[newMemSize];

		memSize = newMemSize;
	}

	E[T.NodeArray.n] = ENew;

	// Compute new minimum spanning tree T_.

	memset(bProcessed, 0, (T.NodeArray.n + 1) * sizeof(bool));

	UpdateMSTreeInsertData *pPut = Q;

	pPut->q = NULL;
	pPut->r = 0;
	pPut->state = 0;

	pPut++;

	UpdateMSTreeInsertData *pProcess = Q;

	int nT_ = 0;

	int z = T.NodeArray.n;

	int r, w;
	Pair<int, int> t, k, h;
	float ct, cwr, ck, cm;
	bool bInsert;

	while (pProcess)
	{
		r = pProcess->r;

		if (pProcess->state == 0)
		{
			bProcessed[r] = true;

			pProcess->m.a = r;
			pProcess->m.b = z;

			pProcess->pEdgePtr = T.NodeArray.Element[r].EdgeList.pFirst;
		}
		else
		{
			w = RVLPCSEGMENT_GRAPH_GET_OPPOSITE_NODE(pProcess->pEdgePtr);

			ct = (t.a > t.b ? E[t.a][t.b] : E[t.b][t.a]);
			cwr = (w > r ? E[w][r] : E[r][w]);

			if (ct > cwr)
			{
				k = t;
				h.a = w;
				h.b = r;
				ck = ct;
			}
			else
			{
				k.a = w;
				k.b = r;
				h = t;
				ck = cwr;
			}

			T_[nT_++] = h;

			cm = (pProcess->m.a > pProcess->m.b ? E[pProcess->m.a][pProcess->m.b] : E[pProcess->m.b][pProcess->m.a]);

			if (ck < cm)
				pProcess->m = k;

			pProcess->pEdgePtr = pProcess->pEdgePtr->pNext;
		}

		bInsert = false;

		while (pProcess->pEdgePtr)
		{
			w = RVLPCSEGMENT_GRAPH_GET_OPPOSITE_NODE(pProcess->pEdgePtr);

			if (!bProcessed[w])
			{
				pProcess->state = 1;

				pPut->q = pProcess;
				pPut->r = w;
				pPut->state = 0;

				pProcess = pPut;

				pPut++;

				bInsert = true;

				break;
			}

			pProcess->pEdgePtr = pProcess->pEdgePtr->pNext;
		}

		if (!bInsert)
		{
			t = pProcess->m;

			pProcess = pProcess->q;
		}
	}

	T_[nT_++] = t;

	T.NodeArray.Element[T.NodeArray.n].idx = T.NodeArray.n;

	T.NodeArray.n++;

	// Copy T_ to T.

	if (bMemReallocation)
	{
		delete[] T.EdgeMem;

		T.EdgeMem = newEdgeMem;

		T.EdgePtrMem = newEdgePtrMem;
	}

	int i;

	for (i = 0; i < T.NodeArray.n; i++)
	{
		pNode = T.NodeArray.Element + i;

		pEdgeList = &(pNode->EdgeList);

		RVLQLIST_INIT(pEdgeList);
	}

	Edge *pEdge = T.EdgeMem;
	EdgePtr<Edge> *pEdgePtr = T.EdgePtrMem;

	for (i = 0; i < nT_; i++)
	{
		ConnectNodes<Node, Edge, EdgePtr<Edge>>(T_[i].a, T_[i].b, T.NodeArray, pEdge, pEdgePtr);

		pEdge++;
		pEdgePtr += 2;
	}
}

void RVL::GRAPH::MSTreeTest()
{
	// MSTree test.

	MSTree MST;

	MST.Init();

	int i, j;

	float P[8][2] = {
		{ 50, 230 },
		{ 70, 150 },
		{ 10, 260 },
		{ 35, 250 },
		{ 30, 205 },
		{ 80, 210 },
		{ 30, 160 },
		{ 60, 190 } };

	//FILE *fp = fopen("C:\\RVL\\Debug\\P.txt", "w");

	//for (i = 0; i < 8; i++)
	//	fprintf(fp, "%f\t%f\n", P[i][0], P[i][1]);

	//fclose(fp);

	float dP[2];
	float *E;

	for (i = 0; i < 8; i++)
	{
		if (i > 0)
			E = new float[i];
		else
			E = NULL;

		for (j = 0; j < i; j++)
		{
			dP[0] = P[j][0] - P[i][0];
			dP[1] = P[j][1] - P[i][1];

			E[j] = dP[0] * dP[0] + dP[1] * dP[1];
		}

		MST.Update(E);
	}
}