#pragma once

#define RVLVN_GET_CLOSEST_CONVEX_TEMPLATE_ELEMENT(convexTemplateLUT, uS, R, uM, i, j, k, iCorrespondence)\
{\
	RVLMULMX3X3TVECT(R, uS, uM)\
	i = (int)(uM[0] / convexTemplateLUT.quant + convexTemplateLUT.halfRange);\
	j = (int)(uM[1] / convexTemplateLUT.quant + convexTemplateLUT.halfRange);\
	k = (int)(uM[2] / convexTemplateLUT.quant + convexTemplateLUT.halfRange);\
	iCorrespondence = convexTemplateLUT.LUT.Element[RVL3DARRAY_ELEMENT_INDEX(convexTemplateLUT.LUT, i, j, k)];\
}

#define RVLVN_MODELDB_NONE		0
#define RVLVN_MODELDB_3DNET		1
#define RVLVN_MODELDB_OSD		2
#define RVLVN_MODELDB_SHAPENET	3

namespace RVL
{
	class VNClassifier;

	namespace RECOG
	{
		namespace VN_
		{
			struct VisualizationData
			{
				float resolution;
				float SDFSurfaceValue;
				bool bVisualizeConvexClusters;
				bool bVisualizeConcaveClusters;
				bool bVisualizeSurfels;
				bool bVisualizeAllClusters;
				bool bVisualizeCTIs;
				bool bVisualizeSceneFitting;
				VNClassifier *pClassifier;
				Mesh *pMesh;
				SurfelGraph *pSurfels;
				Visualizer *pVisualizer;
				bool bClusters;
				unsigned char selectionColor[3];
				int iSelectedCluster;
				unsigned char *clusterColor;
			};

			void _3DNetDatabaseClasses(VNClassifier *pClassifier);
			void OSDDatabaseClasses(VNClassifier *pClassifier);
			bool keyPressUserFunction(
				Mesh *pMesh,
				SurfelGraph *pSurfels,
				std::string &key,
				void *vpData);
			bool mouseRButtonDownUserFunction(
				Mesh *pMesh,
				SurfelGraph *pSurfels,
				int iSelectedPt,
				int iSelectedSurfel,
				void *vpData);
		}	// namespace VN_
	}	// namespace RECOG

	class VNClassifier
	{
	public:
		VNClassifier();
		virtual ~VNClassifier();
		void Create(char *cfgFileName);
		void Init(Mesh *pMesh);
		void Clear();
		void CreateParamList();
		void FreeTmpMem();
		void ComputeDescriptor(
			Mesh *pMesh,
			float *RIn,
			float *tIn,
			float *&dS,
			bool *&bdS,
			Box<float> &SBoundingBox,
			int iModel = -1);
		void ComputeDescriptor(
			Array<int> iVertexArray,
			Array<SURFEL::NormalHullElement> NHull,
			float *R,
			float *d,
			uchar *bd,
			float *NS,
			int *iV,
			float o = 1.0f,
			bool bNormalTest = true,
			bool bVisibilityTest = true);
		void ComputeDescriptor(
			Mesh *pConvexHull,
			float *R,
			float *d,
			uchar *bd,
			Array<int> *piVertexArray = NULL,
			float o = 1.0f,
			int *iV = NULL);
		void ComputeDescriptors(
			Array<int> iSurfelArray,
			Array<int> iVertexArray,
			float *d,
			uchar *bd,
			float *dU = NULL,
			uchar *bdU = NULL);
		void ComputeDescriptors(
			Array<int> iSurfelArray,
			Array<int> iVertexArray,
			float **pd,
			uchar **pbd);
		void ComputeDescriptors(
			float *dSrc,
			float *dTgt,
			uchar *bd);
		void ComputeDescriptors(
			Mesh *pConvexHull,
			float *R,
			float *d,
			uchar *bd,
			Point **pVertex);
		void MergeDescriptors(
			float *dSrc1,
			uchar *bdSrc1,
			float *dSrc2,
			uchar *bdSrc2,
			float o,
			float *dTgt,
			uchar *bdTgt);
		void Learn(
			char *modelSequenceFileName,
			int iClass = -1,
			Visualizer *pVisualizer = NULL); //Vidovic
		void LearnPrimitives(char *primitiveFileName);
		void Interpret(
			Mesh *pMesh,
			int iClass);
		void Interpretation();
		void GreedyInterpretation();
		bool Classify(
			Mesh *pMesh,
			SURFEL::ObjectGraph *pObjects,
			Array<int> *piConvexObjects,
			Array<int> iSegmentArray,
			RECOG::VN_::FitData *fitData,
			QList<RECOG::VN_::Hypothesis> *pHypothesisList,
			float *d,
			uchar *bd,
			float *f,
			float *R,
			CRVLMem *pMem);
		bool Classify2(
			Mesh *pMesh,
			SURFEL::ObjectGraph *pObjects,
			Array<int> *piConvexObjects,
			Array<int> iSegmentArray,
			RECOG::VN_::FitData *fitData,
			QList<RECOG::VN_::Hypothesis> *pHypothesisList,
			float *d,
			uchar *bd,
			float *f,
			float *R,
			CRVLMem *pMem);
		float InterpretationCost(bool *X);
		void InitInterpretationCostComputation();
		void DeallocateInterpretationCostComputationMem();
		void SegmentToParts(Mesh *pMesh);
		void DetectPrimitives(
			Mesh *pMesh,
			RECOG::VN_::FitData *fitData,
			QList<RECOG::VN_::Hypothesis> *pHypothesisList,
			CRVLMem *pMem);
		void ImageSegmentation(
			Mesh *pMesh,
			Array2D<int> &objectMap);
		void SampleSO3();
		float ProjectToLatentSubspace(
			float *d,
			uchar *bd,
			float *dU,
			uchar *bdU,
			RECOG::ClassData *pClass,
			float *q,
			int &iR);
		float ProjectToLatentSubspace(
			float *d,
			uchar *bd,
			RECOG::ClassData *pClass,
			RECOG::VN_::FitData *pFitData,
			float *q,
			int &iR,
			int &nb,
			bool &bAllDoFsVisible);
		void ProjectToLatentSubspace(
			float *d,
			uchar *bd,
			RECOG::ClassData *pClass,
			RECOG::VN_::FitData *pFitData,
			std::vector<std::tuple<float, float*, int, int, bool>> *allbuffer);	//Filko (tuple<E, q, iR, nb, bAllDoFsVisible>)
		bool FitLS(
			float *d,
			uchar *bd,
			RECOG::ClassData *pClass,
			RECOG::VN_::FitData *pData,
			float &E);
		bool FitRotLS(	
			Array<int> iVertexArray,
			float o,
			float *NM,
			RECOG::ClassData *pClass,
			float *RIn,
			int nIterations,
			bool bNormalTest,
			bool bVisibilityTest,
			RECOG::VN_::FitData *pData,
			float *qOpt,
			float *ROpt,
			float &E);
		float FitCTI(
			Mesh *pMesh,
			Array<int> iSurfelArray, 
			Array<int> iVertexArray, 
			Camera camera,
			int nIterations,
			float *R,
			float *d,
			uchar *bd);
		void InitNormalAlignment(
			Array<MESH::Face *> faces,
			Array<int> iFaceArray,
			float *U);
		float FitCTI(
			Mesh *pConvexHull,
			float *U,
			float *RInit,
			bool *mask,
			int nIterations,
			float *R,
			int *faceCorrespondence,
			Array<int> *piModelNormalArray = NULL);
		float Align(
			Array<MESH::Face *> faceArray,
			Array<MESH::Face *> refFaceArray,
			float *R);
		void SetSceneFileName(char *sceneFileName_);
		void SaveDescriptor(
			FILE *fp,
			float *d,
			bool *bd,
			int iModel,
			int iMetaModel);
		void SaveDescriptor(
			FILE *fp,
			float *d,
			uchar *bd,
			float *R,
			int iCluster,
			int type);
		void SaveDescriptors(
			FILE *fp,
			float *d,
			uchar *bd,
			int iCluster,
			int type);
		void SaveDescriptors(
			FILE *fp,
			float *d,
			uchar *bd,
			int nDescriptors,
			int iCluster,
			int type);
		void LoadDescriptor(
			FILE *fp,
			float *&d,
			bool *&bd,
			int &iModel,
			int &iMetaModel);
		void SaveLatentVector(
			FILE *fp,
			float *q,
			int iClass);
		void LoadLatentVector(
			FILE *fp,
			float *&q,
			int &iClass);
		void LoadShapeSpace(char *shapeSpaceFileName);
		void InitDisplay(
			Visualizer *pVisualizer,
			Mesh *pMesh,
			unsigned char *selectionColor);
		void DisplayClusters();
		void ResetClusterColor(int iCluster);
		void DisplaySelectedCluster(int iCluster);
		void ComputeClassPropertiesForGrasping(int iClass, VN *pModel);
		void SaveVertices(FILE *fp, RECOG::PSGM_::Cluster *pSCluster);
		float ComputeDescriptor(
			Mesh *pMesh,
			float *RIn,
			float *tIn,
			float *&dS,
			bool *&bdS,
			Box<float> &SBoundingBox,
			int iModel,
			float score);
		//bool ReferenceFrames(Mesh convexHull);
	public:
		CRVLMem *pMem0;
		CRVLMem *pMem;
		CRVLParameterList paramList;
		DWORD mode;
		DWORD problem;
		int trainingMethod;
		DWORD modelDB;
		SurfelGraph *pSurfels;
		SURFEL::ObjectGraph *pObjects;
		PlanarSurfelDetector *pSurfelDetector;
		Array2D<float> sampledUnitSphere;
		PSGM alignment;
		PSGM convexClustering;
		PSGM concaveClustering;
		Camera camera;
		std::vector<VN *> models;
		std::vector<VN *> metaModels;
		float kMaxMatchCost;
		float tangentOrientationTolerance;
		float kTangentDistanceTolerance;
		float clusteringTolerance;
		int maxnSCClusters;
		int maxnSUClusters;
		int maxnSTClusters;
		float voxelSize;
		int sampleVoxelDistance;
		float connectedComponentMaxDist;
		int connectedComponentMinSize;
		float maxDistFromNHull;
		float maxedmax;
		float maxBoundingSphereRadius;
		float kZMin2;
		float alignmentSupportThr;
		int maxOutlierPerc;
		float maxz;
		int collisionPercThr;
		float convexityErrorCoeff;
		float invisibleComponentPenal;
		float unassignedPointCost;
		float multiplePointAssignmentCost;
		float outlierCost;
		float sigmaR;
		float sigmaRdeg;
		int simulatedAnnealingnIterationsPerHypothesis;
		int simulatedAnnealingnSamples;
		RECOG::VN_::FitParams fitParams;
		bool b3DNetTestSet;
		bool b3DNetVOI;
		bool bGndConstraint;
		bool bGenerateModelCTIs;
		bool bCreateAlignmentTree;
		bool bPrimitiveLayer;
		bool bLoadCTIDataBase;
		bool bVisualization;
		bool bVisualizeBestInstance;
		bool bSingleClass;
		bool bCTIAlignment;
		bool bDepthImage;
		bool bOwnsObjectGraph;
		bool bAlignWithConvexTemplate;
		bool bHorizontalBottomAssumption;
		bool bShapeSpace;
		char *modelDataBase; //Vidovic
		char *modelsInDataBase; //Vidovic
		char *VNShapeSpaceFileName;
		char *primitiveDataBase;
		char *sceneFileName; //Vidovic
		char *validFeaturesFileName;
		char *resultsFolder;
		void *vpMeshBuilder;
		bool(*LoadMesh)(void *vpMeshBuilder,
			char *FileName,
			Mesh *pMesh,
			bool bSavePLY);
		RECOG::VN_::VisualizationData visualizationData;
		RECOG::VN_::SceneObject sceneObject;
		float NGnd[3];
		float dGnd;
		bool bGnd;
		RECOG::VN_::Instance refModel;
		Array<RECOG::ClassData> classArray;
		int *hypClass;
		int *clusterMap;
		Array<RECOG::SceneCluster> SClusters;
		Array3D<float> SO3Samples;
		CRVLMem descriptorMem;
		bool bTrainingStructure;
		FILE *fpDebug;
		float timeInterpret;
		RECOG::PSGM_::ConvexTemplateLookUpTable convexTemplateLUT;
		RECOG::PSGM_::CTIDescriptorMapping<float> CTIDescriptorMapingData;
		Eigen::MatrixXf convexTemplate;
		int debug1, debug2;
		Mesh convexHull;
		Array<RECOG::VN_::Hypothesis *> hypotheses;
		Array<int> interpretation;
		Array<SortIndex<float>> sortedHypotheses;
		RECOG::VN_::FitData *fitData;
		float *D;
		uchar *bD;
		float *f;
		float *R;
		bool bLearnMetamodels;
		RECOG::VN_::PrimitiveLayerOutput primitiveLayerOutput;

		//struct convexHullFacesPairs
		//{
		//	Pair < int, int > *faces;
		//	float cost;

		//};

	private:
		Array<int> iSurfelArray;
		Array<int> iVertexArray;
		bool *bSurfelInArray;
		bool *bVertexInArray;
		Array<int> iConcavitySurfelArray;
		Array<int> iConcavityVertexArray;
		Array2D<float> convexHullVertices;
		Array2D<float> concavityVertices;
		CRVLMem memConvexHull;
		Mesh CTIMesh;
		CRVLMem memCTIMesh;
		int CTIMeshVertexMemSize;
		QList<VN> VNList;
		int *nSurfelCellHypotheses;
	};
}

