//#include "stdafx.h"

#define RVLVN_TIME_MESUREMENT //Vidovic
#define RVLVN_RECORD_RESULTS
#ifdef RVLVN_TIME_MESUREMENT
#undef RVLVN_VERBOSE
#include <Windows.h>
#endif

#include <algorithm>
#include "RVLVTK.h"
#include <vtkTriangle.h>
#include <vtkVertexGlyphFilter.h>
#include "RVLCore2.h"
#include "Util.h"
#include "Space3DGrid.h"
#include "Graph.h"
#include "Mesh.h"
#include "MSTree.h"
#include "Visualizer.h"
#include "SceneSegFile.hpp"
#include "SurfelGraph.h"
#include "ObjectGraph.h"
#include "PlanarSurfelDetector.h"
#include "RVLRecognition.h"
#include "PSGMCommon.h"
#include "CTISet.h"
#include "VertexGraph.h"
#include "TG.h"
#include "TGSet.h"
#include "PSGM.h"
#include "ObjectDetector.h"
#include <Eigen\Eigenvalues>
#include "VN.h"
#include "VNClassifier.h"
#include "Gripper.h"
#include <ctime>

//#define RVLVN_CLASSIFY_CTI_VERSION0
#define RVLVN_CLASSIFY_CTI_VERSION1
#define RVLVN_CLASSIFY_CTI_VERSION1B
//#define RVLVN_CLASSIFY_HYPOTHESIS_CTI_LOG

using namespace RVL;
using namespace RECOG;

VNClassifier::VNClassifier()
{
	mode = RVLRECOGNITION_MODE_TRAINING;
	problem = RVLRECOGNITION_PROBLEM_CLASSIFICATION;
	trainingMethod = 0;
	voxelSize = 0.01f;
	sampleVoxelDistance = 2;
	visualizationData.resolution = 0.01f;
	visualizationData.SDFSurfaceValue = 0.0f;
	visualizationData.bVisualizeConvexClusters = false;
	visualizationData.bVisualizeConcaveClusters = false;
	visualizationData.bVisualizeSurfels = false;
	visualizationData.bVisualizeAllClusters = false;
	visualizationData.bVisualizeCTIs = false;
	modelDB = RVLVN_MODELDB_NONE;
	modelDataBase = NULL; //Vidovic
	modelsInDataBase = NULL; //Vidovic
	VNShapeSpaceFileName = NULL;
	primitiveDataBase = NULL;
	sceneFileName = NULL;
	validFeaturesFileName = NULL;
	classArray.Element = NULL;
	classArray.n = 0;
	sceneObject.sampleArray.Element = NULL;
	refModel.d = NULL;
	refModel.bd = NULL;
	alignment.modelDataBase = NULL;
	sampledUnitSphere.Element = NULL;
	clusterMap = NULL;
	SClusters.Element = NULL;
	SO3Samples.Element = NULL;
	convexTemplateLUT.LUT.Element = NULL;
	CTIDescriptorMapingData.iD = NULL;
	CTIDescriptorMapingData.tree = NULL;
	CTIDescriptorMapingData.R = NULL;
	iSurfelArray.Element = NULL;
	iVertexArray.Element = NULL;
	iConcavitySurfelArray.Element = NULL;
	iConcavityVertexArray.Element = NULL;
	bSurfelInArray = NULL;
	bVertexInArray = NULL;
	convexHullVertices.Element = NULL;
	concavityVertices.Element = NULL;
	hypotheses.Element = NULL;
	interpretation.Element = NULL;
	sortedHypotheses.Element = NULL;
	nSurfelCellHypotheses = NULL;
	fitData = NULL;
	D = NULL;
	bD = NULL;
	f = NULL;
	R = NULL;
	pObjects = NULL;
	primitiveLayerOutput.q = NULL;
	primitiveLayerOutput.y = NULL;
	primitiveLayerOutput.R = NULL;
	maxnSCClusters = 4;
	maxnSUClusters = 2;
	maxnSTClusters = 2;
	connectedComponentMaxDist = 0.050f;
	connectedComponentMinSize = 100;
	maxDistFromNHull = 0.3f;
	maxedmax = 0.005f;
	kMaxMatchCost = 0.06f;
	kTangentDistanceTolerance = 0.01f;
	tangentOrientationTolerance = 0.01f;
	maxBoundingSphereRadius = 0.2f;
	kZMin2 = 5.0f;
	alignmentSupportThr = 0.5f;
	maxOutlierPerc = 10;	// %
	maxz = 2.0f;	// m
	collisionPercThr = 50;
	convexityErrorCoeff = 0.01f;
	invisibleComponentPenal = 0.0001f;
	unassignedPointCost = 0.0001f;
	multiplePointAssignmentCost = 0.0001f;
	outlierCost = 0.001f;
	sigmaRdeg = 10.0f;	// deg
	simulatedAnnealingnIterationsPerHypothesis = 50;
	simulatedAnnealingnSamples = 500;
	fitParams.alpha = 20.0f;
	fitParams.beta = 0.02f;
	fitParams.lambda0 = 0.0f;
	fitParams.lambda1 = 0.01f;
	fitParams.regularization = 3;
	fitParams.maxe = 0.010f;
	fitParams.bInit = false;
	fitParams.bGnd = true;
	fitParams.bGnd2 = true;
	fitParams.maxnIterations = 20;
	fitParams.nFitRotLSIterations = 10;
	b3DNetTestSet = false;
	b3DNetVOI = false;
	bGndConstraint = true;
	bVisualization = false;
	bVisualizeBestInstance = false;
	bLoadCTIDataBase = true;
	bSingleClass = false;
	bGenerateModelCTIs = false;
	bCreateAlignmentTree = false;
	bPrimitiveLayer = false;
	bCTIAlignment = false;
	bTrainingStructure = false;
	bDepthImage = true;
	bOwnsObjectGraph = false;
	bAlignWithConvexTemplate = false;
	bHorizontalBottomAssumption = false;
	bShapeSpace = false;

	hypClass = new int[10];
	memset(hypClass, 0, 10 * sizeof(int));

	descriptorMem.Create(1000000000);

	QList<VN> *pVNList = &VNList;

	RVLQLIST_INIT(pVNList);

	timeInterpret = 0;

	debug1 = debug2 = -1;
}


VNClassifier::~VNClassifier()
{
	Clear();
}

void VNClassifier::Create(char *cfgFileName)
{
	// sampledUnitSphere <- set of unit vectors obtained by uniform sampling of the unit sphere

	alignment.TemplateMatrix(sampledUnitSphere);

	// Sample SO3.

	SampleSO3();

	// Create convex clustering tool.

	convexClustering.pMem = pMem;

	convexClustering.pSurfels = pSurfels;

	convexClustering.pSurfelDetector = pSurfelDetector;

	convexClustering.bDetectGroundPlane = false;
	convexClustering.bOverlappingClusters = true;
	convexClustering.bEdgeClusters = true;
	convexClustering.bObjectDetection = false;
	
	// Create concave clustering tool.

	concaveClustering.pMem = pMem;

	concaveClustering.pSurfels = pSurfels;

	concaveClustering.pSurfelDetector = pSurfelDetector;

	concaveClustering.clusterType = -1.0f;

	concaveClustering.bDetectGroundPlane = false;
	concaveClustering.bOverlappingClusters = true;
	concaveClustering.bEdgeClusters = true;
	concaveClustering.bObjectDetection = false;

	/// Create meta models.

	char metaModelFileName[] = "metamodel0.dat";

	FILE *fp;

	VN *pModel;

	// Create meta model RVLVN_METAMODEL_CONVEX.

	pModel = new VN;

	VN_::CreateConvex(pModel, alignment.convexTemplate66, pMem0);

	//sprintf(metaModelFileName + strlen(metaModelFileName) - 5, "%1d.dat", RVLVN_METAMODEL_CONVEX);

	//fp = fopen(metaModelFileName, "w");

	//pModel->SaveFeatures(fp);

	//fclose(fp);

	//models.push_back(pModel);
	metaModels.push_back(pModel);

	// Create meta model RVLVN_METAMODEL_BANANA.

	pModel = new VN;

	VN_::CreateBanana(pModel, alignment.convexTemplate66, pMem0);

	//sprintf(metaModelFileName + strlen(metaModelFileName) - 5, "%1d.dat", RVLVN_METAMODEL_BANANA);

	//fp = fopen(metaModelFileName, "w");

	//pModel->SaveFeatures(fp);

	//fclose(fp);

	//models.push_back(pModel);
	metaModels.push_back(pModel);

	// Create meta model RVLVN_METAMODEL_BOTTLE.

	pModel = new VN;

	//VN_::CreateBottle(pModel, alignment.convexTemplate66, pMem0);
	VN_::CreateTwoConvex(pModel, alignment.convexTemplate66, pMem0);


	//sprintf(metaModelFileName + strlen(metaModelFileName) - 5, "%1d.dat", RVLVN_METAMODEL_BOTTLE);

	//fp = fopen(metaModelFileName, "w");

	//pModel->SaveFeatures(fp);

	//fclose(fp);

	//models.push_back(pModel);
	metaModels.push_back(pModel);

	// Create meta model RVLVN_METAMODEL_BOWL.

	pModel = new VN;

	VN_::CreateBowl2(pModel, alignment.convexTemplate66, pMem0);

	//sprintf(metaModelFileName + strlen(metaModelFileName) - 5, "%1d.dat", RVLVN_METAMODEL_BOWL);

	//fp = fopen(metaModelFileName, "w");

	//pModel->SaveFeatures(fp);

	//fclose(fp);

	//models.push_back(pModel);
	metaModels.push_back(pModel);

	// Create meta model RVLVN_METAMODEL_TORUS.

	pModel = new VN;

	VN_::CreateTorus(pModel, alignment.convexTemplate66, pMem0);

	//sprintf(metaModelFileName + strlen(metaModelFileName) - 5, "%1d.dat", RVLVN_METAMODEL_TORUS);

	//fp = fopen(metaModelFileName, "w");

	//pModel->SaveFeatures(fp);

	//fclose(fp);

	//models.push_back(pModel);
	metaModels.push_back(pModel);

	// Create meta model RVLVN_METAMODEL_HAMMER.

	pModel = new VN;

	VN_::CreateHammer(pModel, alignment.convexTemplate66, pMem0);
	

	//sprintf(metaModelFileName + strlen(metaModelFileName) - 5, "%1d.dat", RVLVN_METAMODEL_HAMMER);

	//fp = fopen(metaModelFileName, "w");

	//pModel->SaveFeatures(fp);

	//fclose(fp);

	//models.push_back(pModel);
	metaModels.push_back(pModel);

	// Create meta model RVLVN_METAMODEL_MUG.

	pModel = new VN;

	//VN_::CreateMug(pModel, pMem0);
	//VN_::CreateMug2(pModel, alignment.convexTemplate66, pMem0);
	//VN_::CreateMug3(pModel, alignment.convexTemplate66, pMem0);
	VN_::CreateMug4(pModel, alignment.convexTemplate66, pMem0);

	//sprintf(metaModelFileName + strlen(metaModelFileName) - 5, "%1d.dat", RVLVN_METAMODEL_MUG);

	//fp = fopen(metaModelFileName, "w");

	//pModel->SaveFeatures(fp);

	//fclose(fp);

	//models.push_back(pModel);
	metaModels.push_back(pModel);

	// Create meta model RVLVN_METAMODEL_PIPE.

	pModel = new VN;

	VN_::CreatePipe(pModel, alignment.convexTemplate66, pMem0);
	//VN_::CreateBowl2(pModel, alignment.convexTemplate66, pMem0);

	//sprintf(metaModelFileName + strlen(metaModelFileName) - 5, "%1d.dat", RVLVN_METAMODEL_PIPE);

	//fp = fopen(metaModelFileName, "w");

	//pModel->SaveFeatures(fp);

	//fclose(fp);

	//models.push_back(pModel);
	metaModels.push_back(pModel);

	// Create meta model RVLVN_METAMODEL_CAR.

	//pModel = new VN;

	//VN_::CreateCar(pModel, alignment.convexTemplate66, pMem0);

	//sprintf(metaModelFileName + strlen(metaModelFileName) - 5, "%1d.dat", RVLVN_METAMODEL_CAR);

	//fp = fopen(metaModelFileName, "w");

	//pModel->SaveFeatures(fp);

	//fclose(fp);

	//models.push_back(pModel);

	// Create meta model RVLVN_METAMODEL_APPLE.

	//pModel = new VN;

	//VN_::CreateApple(pModel, alignment.convexTemplate66, pMem0);

	//sprintf(metaModelFileName + strlen(metaModelFileName) - 5, "%1d.dat", RVLVN_METAMODEL_APPLE);

	//fp = fopen(metaModelFileName, "w");

	//pModel->SaveFeatures(fp);

	//fclose(fp);

	//models.push_back(pModel);

	// Create meta model RVLVN_METAMODEL_MUG5.

	pModel = new VN;

	VN_::CreateMug5(pModel, alignment.convexTemplate66, pMem0);
	metaModels.push_back(pModel);

	// Set parameters.

	CreateParamList();

	paramList.LoadParams(cfgFileName);

	sigmaR = sigmaRdeg * DEG2RAD;

	clusteringTolerance = 3.0f * convexClustering.kNoise * 2.0f / pSurfelDetector->kPlane;

	if (b3DNetTestSet)
		b3DNetVOI = true;

	// Configure 

	if (problem == RVLRECOGNITION_PROBLEM_PART_SEGMENTATION)
		convexClustering.bLabelConstrainedClustering = concaveClustering.bLabelConstrainedClustering = true;

	// Create object classes.

	if (modelDB == RVLVN_MODELDB_3DNET)
		VN_::_3DNetDatabaseClasses(this);
	else if (modelDB == RVLVN_MODELDB_OSD)
		VN_::OSDDatabaseClasses(this);

	// Load shape space.

	if (VNShapeSpaceFileName)
		LoadShapeSpace(VNShapeSpaceFileName);

	// Initialize alignment tool.

	if (mode == RVLRECOGNITION_MODE_TRAINING)
		//if (trainingMethod == 0)
			bCTIAlignment = true;

	if (bCTIAlignment || b3DNetTestSet)
	{
		alignment.pMem0 = pMem0;
		alignment.pMem = pMem;
		alignment.pSurfels = pSurfels;
		alignment.pSurfelDetector = pSurfelDetector;

		alignment.problem = RVLRECOGNITION_PROBLEM_CLASSIFICATION;

		alignment.Init(cfgFileName);

		pObjects = alignment.pObjects;

		if (bGenerateModelCTIs)
			bLoadCTIDataBase = false;

		if (modelDB != RVLVN_MODELDB_NONE)
		{
			alignment.modelDataBase = RVLCreateString(modelDataBase);

			if (bLoadCTIDataBase)
			{
				printf("Loading CTI database %s...", modelDataBase);

				alignment.LoadModelDataBase();

				printf("completed.\n");
			}
		}

		alignment.bGroundPlaneRFDescriptors = true;
	}
	else if (pObjects == NULL)
	{
		pObjects = new SURFEL::ObjectGraph;

		pObjects->CreateParamList(pMem0);

		pObjects->ParamList.LoadParams(cfgFileName);

		pObjects->pMem = pMem;
		pObjects->pSurfels = pSurfels;

		bOwnsObjectGraph = true;
	}

	// Create convex template lookup table.	

	printf("Creating convex template lookup table...");

	PSGM_::CreateTemplateLookUpTable(sampledUnitSphere, 200, convexTemplateLUT);

	printf("completed.\n");

	// Load CTI descriptor mapper data.

	PSGM_::LoadCTIDescriptorMapping("..\\CTIDescriptorMapper.dat", CTIDescriptorMapingData);

	// Allocate memory for classification.

	int iClass;
	RECOG::ClassData *pClass;

	if (bShapeSpace)
	{
	fitData = new RECOG::VN_::FitData[classArray.n];

	for (iClass = 0; iClass < classArray.n; iClass++)
	{
		pClass = classArray.Element + iClass;

			RECOG::VN_::Init(fitData[iClass], pClass->M, fitParams.beta, fitParams.lambda0, fitParams.lambda1, false, 0);
	}
	}

	int nZ = CTIDescriptorMapingData.nPan * (CTIDescriptorMapingData.nTilt - 1) + 1;

	int nDescriptors = nZ * 6 * 4 * CTIDescriptorMapingData.nRoll;

	D = new float[sampledUnitSphere.h * nDescriptors];

	bD = new uchar[sampledUnitSphere.h * nDescriptors];

	f = new float[sampledUnitSphere.h * nDescriptors];

	R = new float[9 * nZ * CTIDescriptorMapingData.nRoll];

	// Initialize primitive layer output.

	int mTotal = 0;

	for (iClass = 0; iClass < classArray.n; iClass++)
	{
		pClass = classArray.Element + iClass;

		mTotal += pClass->M.h;
	}

	primitiveLayerOutput.R = new float[9 * alignment.CTIDescMap.nCTI];
	primitiveLayerOutput.q = new float[mTotal * alignment.CTIDescMap.nCTI];
	primitiveLayerOutput.y = new float[classArray.n * alignment.CTIDescMap.nCTI];
}

void VNClassifier::CreateParamList()
{
	paramList.m_pMem = pMem0;

	RVLPARAM_DATA *pParamData;

	paramList.Init();

	pParamData = paramList.AddParam("VN.trainingMethod", RVLPARAM_TYPE_INT, &trainingMethod);
	pParamData = paramList.AddParam("VN.3DNetTestSet", RVLPARAM_TYPE_BOOL, &b3DNetTestSet);
	pParamData = paramList.AddParam("VN.3DNetVOI", RVLPARAM_TYPE_BOOL, &b3DNetVOI);
	pParamData = paramList.AddParam("VN.ModelDB", RVLPARAM_TYPE_ID, &modelDB);
	paramList.AddID(pParamData, "3DNET", RVLVN_MODELDB_3DNET);
	paramList.AddID(pParamData, "OSD", RVLVN_MODELDB_OSD);
	paramList.AddID(pParamData, "SHAPENET", RVLVN_MODELDB_SHAPENET);
	pParamData = paramList.AddParam("VN.GenerateModelCTIs", RVLPARAM_TYPE_BOOL, &bGenerateModelCTIs);
	pParamData = paramList.AddParam("VN.bCreateAlignmentTree", RVLPARAM_TYPE_BOOL, &bCreateAlignmentTree);
	pParamData = paramList.AddParam("VN.primitiveLayer", RVLPARAM_TYPE_BOOL, &bPrimitiveLayer);
	pParamData = paramList.AddParam("VN.kMaxMatchCost", RVLPARAM_TYPE_FLOAT, &kMaxMatchCost);
	pParamData = paramList.AddParam("VN.tangentOrientationTolerance", RVLPARAM_TYPE_FLOAT, &tangentOrientationTolerance);
	pParamData = paramList.AddParam("VN.kTangentDistanceTolerance", RVLPARAM_TYPE_FLOAT, &kTangentDistanceTolerance);
	pParamData = paramList.AddParam("VN.maxnSCClusters", RVLPARAM_TYPE_INT, &maxnSCClusters);
	pParamData = paramList.AddParam("VN.maxnSUClusters", RVLPARAM_TYPE_INT, &maxnSUClusters);
	pParamData = paramList.AddParam("VN.maxnSTClusters", RVLPARAM_TYPE_INT, &maxnSTClusters);
	pParamData = paramList.AddParam("Recognition.problem", RVLPARAM_TYPE_ID, &problem);
	paramList.AddID(pParamData, "CLASSIFICATION", RVLRECOGNITION_PROBLEM_CLASSIFICATION);
	paramList.AddID(pParamData, "PART_SEGMENTATION", RVLRECOGNITION_PROBLEM_PART_SEGMENTATION);
	pParamData = paramList.AddParam("Recognition.mode", RVLPARAM_TYPE_ID, &mode);
	paramList.AddID(pParamData, "TRAINING", RVLRECOGNITION_MODE_TRAINING);
	paramList.AddID(pParamData, "RECOGNITION", RVLRECOGNITION_MODE_RECOGNITION);
	pParamData = paramList.AddParam("ModelDataBase", RVLPARAM_TYPE_STRING, &modelDataBase); //Vidovic
	pParamData = paramList.AddParam("ModelsInDataBase", RVLPARAM_TYPE_STRING, &modelsInDataBase); //Vidovic
	pParamData = paramList.AddParam("VN.PrimitiveCTIDataBase", RVLPARAM_TYPE_STRING, &primitiveDataBase);
	pParamData = paramList.AddParam("VN.voxelSize", RVLPARAM_TYPE_FLOAT, &voxelSize);
	pParamData = paramList.AddParam("VN.connectedComponentMaxDist", RVLPARAM_TYPE_FLOAT, &connectedComponentMaxDist);
	pParamData = paramList.AddParam("VN.connectedComponentMinSize", RVLPARAM_TYPE_INT, &connectedComponentMinSize);
	pParamData = paramList.AddParam("VN.fit.alpha", RVLPARAM_TYPE_FLOAT, &(fitParams.alpha));
	pParamData = paramList.AddParam("VN.fit.beta", RVLPARAM_TYPE_FLOAT, &(fitParams.beta));
	pParamData = paramList.AddParam("VN.fit.lambda0", RVLPARAM_TYPE_FLOAT, &(fitParams.lambda0));
	pParamData = paramList.AddParam("VN.fit.lambda1", RVLPARAM_TYPE_FLOAT, &(fitParams.lambda1));
	pParamData = paramList.AddParam("VN.fit.regularization", RVLPARAM_TYPE_INT, &(fitParams.regularization));
	pParamData = paramList.AddParam("VN.fit.maxe", RVLPARAM_TYPE_FLOAT, &(fitParams.maxe));
	pParamData = paramList.AddParam("VN.fit.kSceneSupport", RVLPARAM_TYPE_FLOAT, &(fitParams.kSceneSupport));
	pParamData = paramList.AddParam("VN.fit.kOutliers", RVLPARAM_TYPE_FLOAT, &(fitParams.kOutliers));
	pParamData = paramList.AddParam("VN.fit.kHull", RVLPARAM_TYPE_FLOAT, &(fitParams.kHull));
	pParamData = paramList.AddParam("VN.fit.kZMin", RVLPARAM_TYPE_FLOAT, &(fitParams.kZMin));
	pParamData = paramList.AddParam("VN.fit.cosSurfaceRayAngleThr", RVLPARAM_TYPE_FLOAT, &(fitParams.cosSurfaceRayAngleThr));
	pParamData = paramList.AddParam("VN.fit.maxnIterations", RVLPARAM_TYPE_INT, &(fitParams.maxnIterations));
	pParamData = paramList.AddParam("VN.fit.nFitRotLSIterations", RVLPARAM_TYPE_INT, &(fitParams.nFitRotLSIterations));
	pParamData = paramList.AddParam("VN.fit.bGndConstraint2", RVLPARAM_TYPE_BOOL, &(fitParams.bGnd2));
	pParamData = paramList.AddParam("VN.alignWithConvexTemplate", RVLPARAM_TYPE_BOOL, &bAlignWithConvexTemplate);
	pParamData = paramList.AddParam("VN.horizontalBottomAssumption", RVLPARAM_TYPE_BOOL, &bHorizontalBottomAssumption);
	pParamData = paramList.AddParam("VN.kZMin2", RVLPARAM_TYPE_FLOAT, &kZMin2);
	pParamData = paramList.AddParam("VN.alignmentSupportThr", RVLPARAM_TYPE_FLOAT, &alignmentSupportThr);
	pParamData = paramList.AddParam("VN.maxOutlierPerc", RVLPARAM_TYPE_INT, &maxOutlierPerc);
	pParamData = paramList.AddParam("VN.maxz", RVLPARAM_TYPE_FLOAT, &maxz);
	pParamData = paramList.AddParam("VN.collisionPercThr", RVLPARAM_TYPE_INT, &collisionPercThr);
	pParamData = paramList.AddParam("VN.convexityErrorCoeff", RVLPARAM_TYPE_FLOAT, &convexityErrorCoeff);
	pParamData = paramList.AddParam("VN.invisibleComponentPenal", RVLPARAM_TYPE_FLOAT, &invisibleComponentPenal);
	pParamData = paramList.AddParam("VN.unassignedPointCost", RVLPARAM_TYPE_FLOAT, &unassignedPointCost);
	pParamData = paramList.AddParam("VN.multiplePointAssignmentCost", RVLPARAM_TYPE_FLOAT, &multiplePointAssignmentCost);
	pParamData = paramList.AddParam("VN.outlierCost", RVLPARAM_TYPE_FLOAT, &outlierCost);
	pParamData = paramList.AddParam("VN.sigmaRdeg", RVLPARAM_TYPE_FLOAT, &sigmaRdeg);
	pParamData = paramList.AddParam("VN.simulatedAnnealing.nIterationsPerHypothesis", RVLPARAM_TYPE_INT, &simulatedAnnealingnIterationsPerHypothesis);
	pParamData = paramList.AddParam("VN.simulatedAnnealing.nSamples", RVLPARAM_TYPE_INT, &simulatedAnnealingnSamples);
	pParamData = paramList.AddParam("VN.visualization", RVLPARAM_TYPE_BOOL, &bVisualization);
	pParamData = paramList.AddParam("VN.visualizeBestInstance", RVLPARAM_TYPE_BOOL, &bVisualizeBestInstance);
	pParamData = paramList.AddParam("VN.bGndConstraint", RVLPARAM_TYPE_BOOL, &bGndConstraint);	
	pParamData = paramList.AddParam("VN.singleClass", RVLPARAM_TYPE_BOOL, &bSingleClass);
	pParamData = paramList.AddParam("VN.visualizeConvexClusters", RVLPARAM_TYPE_BOOL, &(visualizationData.bVisualizeConvexClusters));
	pParamData = paramList.AddParam("VN.visualizeConcaveClusters", RVLPARAM_TYPE_BOOL, &(visualizationData.bVisualizeConcaveClusters));
	pParamData = paramList.AddParam("VN.visualizeSurfels", RVLPARAM_TYPE_BOOL, &(visualizationData.bVisualizeSurfels));
	pParamData = paramList.AddParam("VN.visualizeAllClusters", RVLPARAM_TYPE_BOOL, &(visualizationData.bVisualizeAllClusters));
	pParamData = paramList.AddParam("VN.visualizeCTIs", RVLPARAM_TYPE_BOOL, &(visualizationData.bVisualizeCTIs));
	pParamData = paramList.AddParam("VN.visualizeSceneFitting", RVLPARAM_TYPE_BOOL, &(visualizationData.bVisualizeSceneFitting));
	pParamData = paramList.AddParam("VN.ShapeSpaceFileName", RVLPARAM_TYPE_STRING, &VNShapeSpaceFileName);
	pParamData = paramList.AddParam("VN.ValidFeaturesFileName", RVLPARAM_TYPE_STRING, &validFeaturesFileName);
	pParamData = paramList.AddParam("VN.trainingStructure", RVLPARAM_TYPE_BOOL, &bTrainingStructure);
	pParamData = paramList.AddParam("VN.debug1", RVLPARAM_TYPE_INT, &debug1);
	pParamData = paramList.AddParam("VN.debug2", RVLPARAM_TYPE_INT, &debug2);
	pParamData = paramList.AddParam("VN.LearnMetamodels", RVLPARAM_TYPE_BOOL, &bLearnMetamodels);
}

void VNClassifier::Init(Mesh *pMesh)
{
	FreeTmpMem();

	iSurfelArray.Element = new int[pSurfels->NodeArray.n];
	iVertexArray.Element = new int[pSurfels->vertexArray.n];
	iConcavitySurfelArray.Element = new int[pSurfels->NodeArray.n];
	iConcavityVertexArray.Element = new int[pSurfels->vertexArray.n];
	bSurfelInArray = new bool[pSurfels->NodeArray.n];
	memset(bSurfelInArray, 0, pSurfels->NodeArray.n * sizeof(bool));
	bVertexInArray = new bool[pSurfels->vertexArray.n];
	memset(bVertexInArray, 0, pSurfels->vertexArray.n * sizeof(bool));

	convexHullVertices.w = 3;

	convexHullVertices.Element = new float[3 * pSurfels->vertexArray.n];

	concavityVertices.w = 3;

	concavityVertices.Element = new float[3 * pSurfels->vertexArray.n];

	memConvexHull.Create(10000000);

	CTIMeshVertexMemSize = 10 * sampledUnitSphere.h;

	CTIMesh.NodeArray.Element = new Point[CTIMeshVertexMemSize];

	memCTIMesh.Create(10000000);

	convexTemplate = convexClustering.ConvexTemplatenT();

	QList<VN> *pVNList = &VNList;

	RVLQLIST_INIT(pVNList);
}

void VNClassifier::FreeTmpMem()
{
	delete[] iSurfelArray.Element;
	delete[] iVertexArray.Element;
	delete[] iConcavitySurfelArray.Element;
	delete[] iConcavityVertexArray.Element;
	delete[] bSurfelInArray;
	delete[] bVertexInArray;
	delete[] convexHullVertices.Element;
	delete[] concavityVertices.Element;

	VN *pVN = VNList.pFirst;

	VN *pNextVN;

	while (pVN)
	{
		pNextVN = pVN->pNext;

		delete pVN;

		pVN = pNextVN;
	}
}

void VNClassifier::Clear()
{
	int iModel;

	for (iModel = 0; iModel < models.size(); iModel++)
		delete[] models[iModel];

	models.clear();

	int iClass;
	ClassData *pClass;

	for (iClass = 0; iClass < classArray.n; iClass++)
	{
		pClass = classArray.Element + iClass;

		RVL_DELETE_ARRAY(pClass->M.Element);
		RVL_DELETE_ARRAY(pClass->qMin);
		RVL_DELETE_ARRAY(pClass->qMax);
		RVL_DELETE_ARRAY(pClass->A);
		RVL_DELETE_ARRAY(pClass->b);
		RVL_DELETE_ARRAY(pClass->q0);
		RVL_DELETE_ARRAY(pClass->valid);
		RVL_DELETE_ARRAY(pClass->VNGraspNormals_.Element);
		RVL_DELETE_ARRAY(pClass->faces.Element);
	}

	RVL_DELETE_ARRAY(classArray.Element);
	RVL_DELETE_ARRAY(sceneObject.sampleArray.Element);
	RVL_DELETE_ARRAY(refModel.d);
	RVL_DELETE_ARRAY(refModel.bd);
	RVL_DELETE_ARRAY(modelDataBase);
	RVL_DELETE_ARRAY(primitiveDataBase);
	RVL_DELETE_ARRAY(alignment.modelDataBase);
	RVL_DELETE_ARRAY(sceneFileName);
	RVL_DELETE_ARRAY(validFeaturesFileName);
	RVL_DELETE_ARRAY(sampledUnitSphere.Element);
	RVL_DELETE_ARRAY(clusterMap);
	RVL_DELETE_ARRAY(SClusters.Element);
	RVL_DELETE_ARRAY(SO3Samples.Element);
	RVL_DELETE_ARRAY(VNShapeSpaceFileName);
	RVL_DELETE_ARRAY(convexTemplateLUT.LUT.Element);
	PSGM_::DeleteCTIDescriptorMapping<float>(CTIDescriptorMapingData);
	RVL_DELETE_ARRAY(interpretation.Element);
	RVL_DELETE_ARRAY(sortedHypotheses.Element);
	RVL_DELETE_ARRAY(hypotheses.Element);
	RVL_DELETE_ARRAY(nSurfelCellHypotheses);
	if (fitData)
	{
		int iClass;
		for (iClass = 0; iClass < classArray.n; iClass++)
			RECOG::VN_::Delete(fitData[iClass]);
		delete[] fitData;
	}
	RVL_DELETE_ARRAY(D);
	RVL_DELETE_ARRAY(bD);
	RVL_DELETE_ARRAY(f);
	RVL_DELETE_ARRAY(R);

	if (bOwnsObjectGraph)
		delete pObjects;

	RVL_DELETE_ARRAY(primitiveLayerOutput.q);
	RVL_DELETE_ARRAY(primitiveLayerOutput.y);
	RVL_DELETE_ARRAY(primitiveLayerOutput.R);
}

void VNClassifier::ComputeDescriptor(
	Mesh *pMesh,
	float *RIn,
	float *tIn,
	float *&dS,
	bool *&bdS,
	Box<float> &SBoundingBox,
	int iModel)
{
	// Detect surfels.

	pSurfels->Init(pMesh);

	pSurfelDetector->Init(pMesh, pSurfels, pMem);

	printf("Segmentation to surfels...");

	pSurfelDetector->Segment(pMesh, pSurfels);

	printf("completed.\n");

	int nSurfels = pSurfels->NodeArray.n;

	printf("No. of surfels = %d\n", nSurfels);

	pSurfels->DetectVertices(pMesh);

	// Create scene object.	

	float *R = sceneObject.R;

	if (RIn)
	{
		RVLCOPYMX3X3(RIn, R);
	}
	else
	{
		RVLUNITMX3(R);
	}

	float *t = sceneObject.t;

	if (tIn)
	{
		RVLCOPY3VECTOR(tIn, t);
	}
	else
	{
		RVLNULL3VECTOR(t);
	}

	sceneObject.vertexArray = new float[3 * pSurfels->vertexArray.n];

	float *P = sceneObject.vertexArray;

	int iVertex;
	SURFEL::Vertex *pVertex;

	for (iVertex = 0; iVertex < pSurfels->vertexArray.n; iVertex++, P += 3)
	{
		pVertex = pSurfels->vertexArray.Element[iVertex];

		RVLTRANSF3(pVertex->P, R, t, P);
	}

	sceneObject.NArray = new float[3 * pSurfels->NodeArray.n];

	float *N = sceneObject.NArray;

	int iSurfel;
	Surfel *pSurfel;

	for (iSurfel = 0; iSurfel < pSurfels->NodeArray.n; iSurfel++, N += 3)
	{
		pSurfel = pSurfels->NodeArray.Element + iSurfel;

		RVLMULMX3X3VECT(R, pSurfel->N, N);
	}

	int nSamplePts = 300;

	pSurfels->SampleMeshDistanceFunction(pMesh, R, t, nSamplePts, voxelSize, sampleVoxelDistance, sceneObject.sampleArray);

	// Cluster surfels into convex and concave surfaces.

	Array<RECOG::SceneCluster> SClusters;

	bool bConcavity, bTorus;

	VN *pModel = (iModel >= 0 ? models[iModel] : NULL);

	if (trainingMethod == 0)
		pModel->ClusterTypes(bConcavity, bTorus);
	else if (trainingMethod == 1)
	{
		bConcavity = true;
		bTorus = false;
	}

	int nSCClusters, nSUClusters;

	PSGM_::ConvexAndConcaveClusters(pMesh, pSurfels, pSurfelDetector, &convexClustering, &concaveClustering,
		SClusters, nSCClusters, nSUClusters, 0, NULL, maxnSCClusters, maxnSUClusters, bConcavity, (bTorus ? maxnSTClusters : 0),
		visualizationData.bVisualizeConvexClusters, visualizationData.bVisualizeConcaveClusters, visualizationData.bVisualizeSurfels);

	//fpDebug = fopen("debug.txt", "w");

	//fprintf(fpDebug, "nFeatures=%d\n", pModel->featureArray.n);

	//for (int i = 0; i < pModel->featureArray.n; i++)
	//	fprintf(fpDebug, "%f\t%f\t%f\n", pModel->featureArray.Element[i].N[0], pModel->featureArray.Element[i].N[1], pModel->featureArray.Element[i].N[2]);

	//for (int i = pModel->featureArray.n; i < pModel->NodeArray.n; i++)
	//	fprintf(fpDebug, "O%d\n", pModel->NodeArray.Element[i].operation);

	//RECOG::VN_::Edge *pEdge = pModel->EdgeList.pFirst;

	//while (pEdge)
	//{
	//	fprintf(fpDebug, "E%d-%d\n", pEdge->data.a, pEdge->data.b);

	//	pEdge = pEdge->pNext;
	//}

	//fprintf(fpDebug, "#conv.clusters %d, #concave clusters: %d\n", convexClustering.clusters.n, concaveClustering.clusters.n);

	//RECOG::PSGM_::Cluster *pSCluster_;

	//for (int i = 0; i < SClusters.n; i++)
	//{
	//	RECOG::SceneCluster *pSCluster = SClusters.Element + i;

	//	if (pSCluster->type == RVLVN_CLUSTER_TYPE_CONVEX || pSCluster->type == RVLVN_CLUSTER_TYPE_CONCAVE)
	//	{
	//		pSCluster_ = (RECOG::PSGM_::Cluster *)(pSCluster->vpCluster);

	//		fprintf(fpDebug, "%d\n", pSCluster_->size);
	//	}
	//}

	// Cluster surfels into convex surfaces.

	//convexClustering.pMesh = pMesh;

	//convexClustering.Clusters();

	// Cluster surfels into concave surfaces.

	//concaveClustering.pMesh = pMesh;

	//concaveClustering.Clusters();

	// Selection color for visualization.

	//uchar SelectionColor[] = {0, 255, 0};

	// Visualization

	//Visualizer visualizer;

	//visualizer.Create();

	//pSurfels->NodeColors(SelectionColor);

	// Visualization of convex clusters.

	//RVL_DELETE_ARRAY(convexClustering.clusterColor);

	//RandomColors(SelectionColor, convexClustering.clusterColor, convexClustering.clusters.n);

	//convexClustering.InitDisplay(&visualizer, pMesh, SelectionColor);

	//convexClustering.Display();

	//visualizer.Run();

	//pSurfels->DisplayData.bCallbackFunctionsDefined = false;

	// Visualization of concave clusters.

	//RVL_DELETE_ARRAY(concaveClustering.clusterColor);

	//RandomColors(SelectionColor, concaveClustering.clusterColor, concaveClustering.clusters.n);

	//concaveClustering.InitDisplay(&visualizer, pMesh, SelectionColor);

	//concaveClustering.Display();

	//visualizer.Run();

	//pSurfels->DisplayData.bCallbackFunctionsDefined = false;

	// Surfel visualization.

	//surfels.NodeColors(SelectionColor);

	//surfels.InitDisplay(&visualizer, &mesh, &surfelDetector);		

	//surfels.Display(&visualizer, &mesh);

	//visualizer.Run();

	//pSurfels->DisplayData.bCallbackFunctionsDefined = false;

	if (trainingMethod == 0)
	{
		printf("Matching VN model to scene...");

		P = sceneObject.vertexArray;

		InitBoundingBox<float>(&SBoundingBox, P);

		P += 3;

		for (iVertex = 1; iVertex < pSurfels->vertexArray.n; iVertex++, P += 3)
			UpdateBoundingBox<float>(&SBoundingBox, P);

		dS = new float[pModel->featureArray.n];

		bdS = new bool[pModel->featureArray.n];

		//pModel->fpDebug = fpDebug;

		pModel->Match4(pMesh, sceneObject, this, SClusters, SBoundingBox, dS, bdS, bTrainingStructure);

		//fclose(fpDebug);

		delete[] SClusters.Element;
		delete[] sceneObject.vertexArray;
		delete[] sceneObject.NArray;

		printf("completed.\n");
	}
	else if (trainingMethod == 1)
	{
		// Vidovi
		FILE *fpDescriptors = fopen("C:\\RVL\\ExpRez\\descriptors.txt", "a");
		FILE *fpVertices = fopen("C:\\RVL\\ExpRez\\vertices.txt", "a");
		int iMetaModel = 0;
		//int iModel = 0;
		//VN *pModel_ = models[iModel];

		RECOG::SceneCluster *pSCluster;
		RECOG::PSGM_::Cluster *pSCluster_;

		float R[9];
		RVLUNITMX3(R);

		//NHull is not used
		Array<SURFEL::NormalHullElement> NHull; //not used
		NHull.Element = NULL;
		NHull.n = 0;
		
		//float *dS_ = new float[pModel_->featureArray.n]; //change 66 to variable
		//uchar *bdS_ = new uchar[pModel_->featureArray.n]; //change 66 to variable
		float *dS_ = new float[sampledUnitSphere.h];
		uchar *bdS_ = new uchar[sampledUnitSphere.h];

		//3.10.2018. - added because ComputeDescriptor function was changed
		float *NS = new float[sampledUnitSphere.h * sampledUnitSphere.w];
		int *iV = new int[sampledUnitSphere.h];

		for (int i = 0; i < SClusters.n; i++)
		{
			pSCluster = SClusters.Element + i;

			if (pSCluster->type == RVLVN_CLUSTER_TYPE_CONVEX || pSCluster->type == RVLVN_CLUSTER_TYPE_CONCAVE)
			{
				pSCluster_ = (RECOG::PSGM_::Cluster *)(pSCluster->vpCluster);

				if (pSCluster->type == RVLVN_CLUSTER_TYPE_CONVEX)
				{
					//ComputeDescriptor(pSCluster_->iVertexArray, NHull, R, dS_, bdS_, 1.0F, true, false);
					//SaveDescriptor(fpDescriptors, dS_, bdS_, 0, iMetaModel);
					ComputeDescriptor(pSCluster_->iVertexArray, NHull, R, dS_, bdS_, NS, iV, 1.0F, true, false);
					//SaveDescriptor(fpDescriptors, dS_, (bool*)bdS_, 0, iMetaModel);
					SaveDescriptor(fpDescriptors, dS_, bdS_, R, i, 0);
				}
				else
				{
					//ComputeDescriptor(pSCluster_->iVertexArray, NHull, R, dS_, bdS_, -1.0F, true, false);
					//SaveDescriptor(fpDescriptors, dS_, bdS_, 1, iMetaModel);
					ComputeDescriptor(pSCluster_->iVertexArray, NHull, R, dS_, bdS_, NS, iV, -1.0F, true, false);
					//SaveDescriptor(fpDescriptors, dS_, (bool*)bdS_, 1, iMetaModel);
					SaveDescriptor(fpDescriptors, dS_, bdS_, R, i, 1);
				}
			}

			SaveVertices(fpVertices, pSCluster_);
		}

		fclose(fpDescriptors);
		fclose(fpVertices);

		delete[] dS_;
		delete[] bdS_;
		RVL_DELETE_ARRAY(NHull.Element);
		delete[] NS;
		delete[] iV;
 
		// END Vidovi
	}
}

float VNClassifier::ComputeDescriptor(
	Mesh *pMesh,
	float *RIn,
	float *tIn,
	float *&dS,
	bool *&bdS,
	Box<float> &SBoundingBox,
	int iModel,
	float score)
{
	// Detect surfels.

	pSurfels->Init(pMesh);

	pSurfelDetector->Init(pMesh, pSurfels, pMem);

	printf("Segmentation to surfels...");

	pSurfelDetector->Segment(pMesh, pSurfels);

	printf("completed.\n");

	int nSurfels = pSurfels->NodeArray.n;

	printf("No. of surfels = %d\n", nSurfels);

	pSurfels->DetectVertices(pMesh);

	// Create scene object.	

	float *R = sceneObject.R;

	if (RIn)
	{
		RVLCOPYMX3X3(RIn, R);
	}
	else
	{
		RVLUNITMX3(R);
	}

	float *t = sceneObject.t;

	if (tIn)
	{
		RVLCOPY3VECTOR(tIn, t);
	}
	else
	{
		RVLNULL3VECTOR(t);
	}

	sceneObject.vertexArray = new float[3 * pSurfels->vertexArray.n];

	float *P = sceneObject.vertexArray;

	int iVertex;
	SURFEL::Vertex *pVertex;

	for (iVertex = 0; iVertex < pSurfels->vertexArray.n; iVertex++, P += 3)
	{
		pVertex = pSurfels->vertexArray.Element[iVertex];

		RVLTRANSF3(pVertex->P, R, t, P);
	}

	sceneObject.NArray = new float[3 * pSurfels->NodeArray.n];

	float *N = sceneObject.NArray;

	int iSurfel;
	Surfel *pSurfel;

	for (iSurfel = 0; iSurfel < pSurfels->NodeArray.n; iSurfel++, N += 3)
	{
		pSurfel = pSurfels->NodeArray.Element + iSurfel;

		RVLMULMX3X3VECT(R, pSurfel->N, N);
	}

	int nSamplePts = 300;

	pSurfels->SampleMeshDistanceFunction(pMesh, R, t, nSamplePts, voxelSize, sampleVoxelDistance, sceneObject.sampleArray);

	// Cluster surfels into convex and concave surfaces.

	Array<RECOG::SceneCluster> SClusters;

	bool bConcavity, bTorus;

	VN *pModel = (iModel >= 0 ? models[iModel] : NULL);

	if (trainingMethod == 0)
		pModel->ClusterTypes(bConcavity, bTorus);
	else if (trainingMethod == 1)
	{
		bConcavity = true;
		bTorus = false;
	}

	int nSCClusters, nSUClusters;

	PSGM_::ConvexAndConcaveClusters(pMesh, pSurfels, pSurfelDetector, &convexClustering, &concaveClustering,
		SClusters, nSCClusters, nSUClusters, 0, NULL, maxnSCClusters, maxnSUClusters, bConcavity, (bTorus ? maxnSTClusters : 0),
		visualizationData.bVisualizeConvexClusters, visualizationData.bVisualizeConcaveClusters, visualizationData.bVisualizeSurfels);

	//fpDebug = fopen("debug.txt", "w");

	//fprintf(fpDebug, "nFeatures=%d\n", pModel->featureArray.n);

	//for (int i = 0; i < pModel->featureArray.n; i++)
	//	fprintf(fpDebug, "%f\t%f\t%f\n", pModel->featureArray.Element[i].N[0], pModel->featureArray.Element[i].N[1], pModel->featureArray.Element[i].N[2]);

	//for (int i = pModel->featureArray.n; i < pModel->NodeArray.n; i++)
	//	fprintf(fpDebug, "O%d\n", pModel->NodeArray.Element[i].operation);

	//RECOG::VN_::Edge *pEdge = pModel->EdgeList.pFirst;

	//while (pEdge)
	//{
	//	fprintf(fpDebug, "E%d-%d\n", pEdge->data.a, pEdge->data.b);

	//	pEdge = pEdge->pNext;
	//}

	//fprintf(fpDebug, "#conv.clusters %d, #concave clusters: %d\n", convexClustering.clusters.n, concaveClustering.clusters.n);

	//RECOG::PSGM_::Cluster *pSCluster_;

	//for (int i = 0; i < SClusters.n; i++)
	//{
	//	RECOG::SceneCluster *pSCluster = SClusters.Element + i;

	//	if (pSCluster->type == RVLVN_CLUSTER_TYPE_CONVEX || pSCluster->type == RVLVN_CLUSTER_TYPE_CONCAVE)
	//	{
	//		pSCluster_ = (RECOG::PSGM_::Cluster *)(pSCluster->vpCluster);

	//		fprintf(fpDebug, "%d\n", pSCluster_->size);
	//	}
	//}

	// Cluster surfels into convex surfaces.

	//convexClustering.pMesh = pMesh;

	//convexClustering.Clusters();

	// Cluster surfels into concave surfaces.

	//concaveClustering.pMesh = pMesh;

	//concaveClustering.Clusters();

	// Selection color for visualization.

	//uchar SelectionColor[] = {0, 255, 0};

	// Visualization

	//Visualizer visualizer;

	//visualizer.Create();

	//pSurfels->NodeColors(SelectionColor);

	// Visualization of convex clusters.

	//RVL_DELETE_ARRAY(convexClustering.clusterColor);

	//RandomColors(SelectionColor, convexClustering.clusterColor, convexClustering.clusters.n);

	//convexClustering.InitDisplay(&visualizer, pMesh, SelectionColor);

	//convexClustering.Display();

	//visualizer.Run();

	//pSurfels->DisplayData.bCallbackFunctionsDefined = false;

	// Visualization of concave clusters.

	//RVL_DELETE_ARRAY(concaveClustering.clusterColor);

	//RandomColors(SelectionColor, concaveClustering.clusterColor, concaveClustering.clusters.n);

	//concaveClustering.InitDisplay(&visualizer, pMesh, SelectionColor);

	//concaveClustering.Display();

	//visualizer.Run();

	//pSurfels->DisplayData.bCallbackFunctionsDefined = false;

	// Surfel visualization.

	//surfels.NodeColors(SelectionColor);

	//surfels.InitDisplay(&visualizer, &mesh, &surfelDetector);		

	//surfels.Display(&visualizer, &mesh);

	//visualizer.Run();

	//pSurfels->DisplayData.bCallbackFunctionsDefined = false;

	if (trainingMethod == 0)
	{
		printf("Matching VN model to scene...");

		P = sceneObject.vertexArray;

		InitBoundingBox<float>(&SBoundingBox, P);

		P += 3;

		for (iVertex = 1; iVertex < pSurfels->vertexArray.n; iVertex++, P += 3)
			UpdateBoundingBox<float>(&SBoundingBox, P);

		dS = new float[pModel->featureArray.n];

		bdS = new bool[pModel->featureArray.n];

		//pModel->fpDebug = fpDebug;

		pModel->Match4(pMesh, sceneObject, this, SClusters, SBoundingBox, dS, bdS, bTrainingStructure);
		score = pModel->score;

		//fclose(fpDebug);

		delete[] SClusters.Element;
		delete[] sceneObject.vertexArray;
		delete[] sceneObject.NArray;

		printf("completed.\n");
	}
	else if (trainingMethod == 1)
	{
		// Vidovi
		FILE *fpDescriptors = fopen("C:\\RVL\\ExpRez\\descriptors.txt", "a");
		FILE *fpVertices = fopen("C:\\RVL\\ExpRez\\vertices.txt", "a");
		int iMetaModel = 0;
		//int iModel = 0;
		//VN *pModel_ = models[iModel];

		RECOG::SceneCluster *pSCluster;
		RECOG::PSGM_::Cluster *pSCluster_;

		float R[9];
		RVLUNITMX3(R);

		//NHull is not used
		Array<SURFEL::NormalHullElement> NHull; //not used
		NHull.Element = NULL;
		NHull.n = 0;

		//float *dS_ = new float[pModel_->featureArray.n]; //change 66 to variable
		//uchar *bdS_ = new uchar[pModel_->featureArray.n]; //change 66 to variable
		float *dS_ = new float[sampledUnitSphere.h];
		uchar *bdS_ = new uchar[sampledUnitSphere.h];

		//3.10.2018. - added because ComputeDescriptor function was changed
		float *NS = new float[sampledUnitSphere.h * sampledUnitSphere.w];
		int *iV = new int[sampledUnitSphere.h];

		for (int i = 0; i < SClusters.n; i++)
		{
			pSCluster = SClusters.Element + i;

			if (pSCluster->type == RVLVN_CLUSTER_TYPE_CONVEX || pSCluster->type == RVLVN_CLUSTER_TYPE_CONCAVE)
			{
				pSCluster_ = (RECOG::PSGM_::Cluster *)(pSCluster->vpCluster);

				if (pSCluster->type == RVLVN_CLUSTER_TYPE_CONVEX)
				{
					//ComputeDescriptor(pSCluster_->iVertexArray, NHull, R, dS_, bdS_, 1.0F, true, false);
					//SaveDescriptor(fpDescriptors, dS_, bdS_, 0, iMetaModel);
					ComputeDescriptor(pSCluster_->iVertexArray, NHull, R, dS_, bdS_, NS, iV, 1.0F, true, false);
					//SaveDescriptor(fpDescriptors, dS_, (bool*)bdS_, 0, iMetaModel);
					SaveDescriptor(fpDescriptors, dS_, bdS_, R, i, 0);
				}
				else
				{
					//ComputeDescriptor(pSCluster_->iVertexArray, NHull, R, dS_, bdS_, -1.0F, true, false);
					//SaveDescriptor(fpDescriptors, dS_, bdS_, 1, iMetaModel);
					ComputeDescriptor(pSCluster_->iVertexArray, NHull, R, dS_, bdS_, NS, iV, -1.0F, true, false);
					//SaveDescriptor(fpDescriptors, dS_, (bool*)bdS_, 1, iMetaModel);
					SaveDescriptor(fpDescriptors, dS_, bdS_, R, i, 1);
				}
			}

			SaveVertices(fpVertices, pSCluster_);
		}

		fclose(fpDescriptors);
		fclose(fpVertices);

		delete[] dS_;
		delete[] bdS_;
		RVL_DELETE_ARRAY(NHull.Element);
		delete[] NS;
		delete[] iV;

		// END Vidovi
	}
	return score;
}

void VNClassifier::ComputeDescriptors(
	Array<int> iSurfelArray,
	Array<int> iVertexArray,
	float *d,
	uchar *bd,
	float *dU,
	uchar *bdU)
{
	// NHull <- convex hull of all surfel normals from iSurfelArray (Tried, but it was not successful.)

	Array<SURFEL::NormalHullElement> NHull;

	NHull.Element = new SURFEL::NormalHullElement[iSurfelArray.n];
	NHull.n = 0;

	int i;
	int iSurfel;
	Surfel *pSurfel;

	for (i = 0; i < iSurfelArray.n; i++)
	{
		iSurfel = iSurfelArray.Element[i];

		pSurfel = pSurfels->NodeArray.Element + iSurfel;

		pSurfels->UpdateNormalHull(NHull, pSurfel->N);
	}

	// Only for debugging purpose!

	//float maxDistFromNHull = -2.0;

	//for (i = 0; i < iSurfelArray.n; i++)
	//{
	//	iSurfel = iSurfelArray.Element[i];

	//	pSurfel = pSurfels->NodeArray.Element + iSurfel;

	//	float distFromNHull = pSurfels->DistanceFromNormalHull(NHull, pSurfel->N);

	//	if (distFromNHull > maxDistFromNHull)
	//		maxDistFromNHull = distFromNHull;
	//}

	float *NS = new float[sampledUnitSphere.h * sampledUnitSphere.w];

	int *iV = new int[sampledUnitSphere.h];

	//int debug = 0;

	///

	// Compute a CTI descriptor for every rotation in SO3Samples.

	int iR;
	float *d_;
	uchar *bd_;
	float *R;
	
	for (iR = 0; iR < SO3Samples.c; iR++)
	{
		R = SO3Samples.Element + 9 * iR;

		d_ = d + sampledUnitSphere.h * iR;

		bd_ = bd + sampledUnitSphere.h * iR;

		ComputeDescriptor(iVertexArray, NHull, R, d_, bd_, NS, iV);

		if (dU)
		{
			d_ = dU + sampledUnitSphere.h * iR;

			bd_ = bdU + sampledUnitSphere.h * iR;

			ComputeDescriptor(iVertexArray, NHull, R, d_, bd_, NS, iV, -1.0f);
		}
	}

	// Free memory.

	delete[] NHull.Element;
	delete[] NS;
	delete[] iV;
}

void VNClassifier::ComputeDescriptors(
	Array<int> iSurfelArray,
	Array<int> iVertexArray,
	float **pd,
	uchar **pbd)
{
	CRVLMem *pMem = &descriptorMem;

	int descriptorDataSize = sampledUnitSphere.h * SO3Samples.c;

	float *d;

	RVLMEM_ALLOC_STRUCT_ARRAY(pMem, float, descriptorDataSize, d);

	uchar *bd;

	RVLMEM_ALLOC_STRUCT_ARRAY(pMem, uchar, descriptorDataSize, bd);

	ComputeDescriptors(iSurfelArray, iVertexArray, d, bd);

	*pd = d;

	*pbd = bd;
}

void VNClassifier::ComputeDescriptors(
	float *dSrc,
	float *dTgt,
	uchar *bd)
{
	// Create CTI form dSrc.

	float tCTIc_CTI[3];

	vtkSmartPointer<vtkPolyData> polyData = GenerateCTIPrimitivePolydata_RW(sampledUnitSphere.Element, dSrc, sampledUnitSphere.h, false, 
		NULL, tCTIc_CTI);

	// Copy CTI vertices to ptArray.

	Array<Point> ptArray;

	ptArray.n = polyData->GetNumberOfPoints();

	ptArray.Element = new Point[ptArray.n];

	vtkSmartPointer<vtkFloatArray> pointData = pointData->SafeDownCast(polyData->GetPoints()->GetData());

	Point *pPt = ptArray.Element;

	int iPt;
	QList<MeshEdgePtr> *pEdgeList;

	for (iPt = 0; iPt < ptArray.n; iPt++, pPt++)
	{
		pointData->GetTupleValue(iPt, pPt->P);

		RVLSUM3VECTORS(pPt->P, tCTIc_CTI, pPt->P);

		pEdgeList = &(pPt->EdgeList);

		RVLQLIST_INIT(pEdgeList);
	}

	// Copy CTI faces to faces.

	vtkSmartPointer<vtkCellArray> faceData = polyData->GetPolys();

	int nFaces = polyData->GetNumberOfPolys();

	int nFacePtsTotal = 0;

	int iFace;
	vtkIdType *ptIdx;
	vtkIdType nFacePts;
	faceData->InitTraversal();
	
	for (iFace = 0; iFace < nFaces; iFace++)
	{
		faceData->GetNextCell(nFacePts, ptIdx);

		nFacePtsTotal += (int)nFacePts;
	}

	Array<Array<int>> faces;

	faces.Element = new Array<int>[nFaces];

	int *ptIdxMem = new int[nFacePtsTotal];

	int *iPtIdx = ptIdxMem;

	faceData->InitTraversal();

	int i;

	for (iFace = 0; iFace < nFaces; iFace++)
	{
		faceData->GetNextCell(nFacePts, ptIdx);

		faces.Element[iFace].n = (int)nFacePts;
		faces.Element[iFace].Element = iPtIdx;

		for (i = 0; i < nFacePts; i++)
			*(iPtIdx++) = (int)(ptIdx[i]);
	}

	// Create edges.

	MeshEdge *edgeMem = new MeshEdge[nFacePtsTotal];

	MeshEdge *pEdge = edgeMem;

	MeshEdgePtr *edgePtrMem = new MeshEdgePtr[2 * nFacePtsTotal];

	MeshEdgePtr *pEdgePtr = edgePtrMem;

	bool *bEdge = new bool[ptArray.n * ptArray.n];

	memset(bEdge, 0, ptArray.n * ptArray.n * sizeof(bool));

	// Only for debugging purpose!

	//int nDebug = 0;

	//int *edgeCnt = new int[ptArray.n * ptArray.n];

	//memset(edgeCnt, 0, ptArray.n * ptArray.n * sizeof(int));

	///

	Array<int> *pFace;
	int iPt1, iPt2;
	int iEdge;

	for (iFace = 0; iFace < nFaces; iFace++)
	{
		pFace = faces.Element + iFace;

		iPt1 = pFace->Element[pFace->n - 1];

		for (i = 0; i < pFace->n; i++)
		{
			//nDebug++;

			iPt2 = pFace->Element[i];

			//edgeCnt[iPt1 * ptArray.n + iPt2]++;

			iEdge = iPt1 * ptArray.n + iPt2;

			//if (iPt1 == 0 || iPt2 == 0)
			//	int debug = 0;

			if (!bEdge[iEdge])
			{
				//if (bEdge[iPt1 * ptArray.n + iPt2])
				//	int debug = 0;

				bEdge[iEdge] = bEdge[iPt2 * ptArray.n + iPt1] = true;
				
				pEdge->iVertex[0] = iPt1;
				pEdge->iVertex[1] = iPt2;
				pEdge->pVertexEdgePtr[0] = pEdgePtr;
				pEdge->pVertexEdgePtr[1] = pEdgePtr + 1;

				pPt = ptArray.Element + iPt1;

				pEdgeList = &(pPt->EdgeList);

				RVLQLIST_ADD_ENTRY(pEdgeList, pEdgePtr);

				pEdgePtr->pEdge = pEdge;

				pEdgePtr++;
				
				pPt = ptArray.Element + iPt2;

				pEdgeList = &(pPt->EdgeList);

				RVLQLIST_ADD_ENTRY(pEdgeList, pEdgePtr);

				pEdgePtr->pEdge = pEdge;

				pEdgePtr++;

				pEdge++;
			}

			iPt1 = iPt2;
		}
	}

	// Only for debugging purpose!

	//for (i = 0; i < ptArray.n * ptArray.n; i++)
	//	if (edgeCnt[i] > 1)
	//		int debug = 0;

	//for (i = 0; i < ptArray.n; i++)
	//	for (int j = i + 1; j < ptArray.n; j++)
	//		if (edgeCnt[i * ptArray.n + j] > 0)
	//			if (edgeCnt[j * ptArray.n + i] == 0)
	//				int debug = 0;

	//delete[] edgeCnt;

	///

	delete[] bEdge;

	// Compute a CTI descriptor for every rotation in SO3Samples.

	int iR, j;
	float *d;
	float *R;
	float *N;
	float N_[3];
	float dmax, d_;
	Point *pPtNext, *pPt_;
	uchar *bd_;

	for (iR = 0; iR < SO3Samples.c; iR++)
	{
		R = SO3Samples.Element + 9 * iR;

		d = dTgt + sampledUnitSphere.h * iR;

		bd_ = bd + sampledUnitSphere.h * iR;

		for (i = 0; i < sampledUnitSphere.h; i++)
		{
			//if (i == 33)
			//	int debug = 0;

			N = sampledUnitSphere.Element + 3 * i;

			RVLMULMX3X3VECT(R, N, N_);

			pPt = ptArray.Element;

			dmax = RVLDOTPRODUCT3(N_, pPt->P);

			for (j = 1; j < ptArray.n; j++)
			{
				pPt_ = ptArray.Element + j;

				d_ = RVLDOTPRODUCT3(N_, pPt_->P);

				if (d_ > dmax)
					dmax = d_;
			}

			//while (pPt)
			//{
			//	pPtNext = NULL;

			//	pEdgePtr = pPt->EdgeList.pFirst;

			//	while (pEdgePtr)
			//	{
			//		iPt2 = RVLPCSEGMENT_GRAPH_GET_OPPOSITE_NODE(pEdgePtr);

			//		pPt_ = ptArray.Element + iPt2;

			//		d_ = RVLDOTPRODUCT3(N_, pPt_->P);

			//		if (d_ > dmax)
			//		{
			//			dmax = d_;

			//			pPtNext = pPt_;

			//			break;
			//		}

			//		pEdgePtr = pEdgePtr->pNext;
			//	}

			//	pPt = pPtNext;
			//}

			//if (RVLABS(dmax_ - dmax) > 0.005f)
			//	int debug = 0;

			d[i] = dmax;

			bd_[i] = (dmax < 0.0f ? 1 : 0);
		}		
	}

	// Deallocate memory.

	delete[] ptArray.Element;
	delete[] ptIdxMem;
	delete[] faces.Element;
	delete[] edgeMem;
	delete[] edgePtrMem;
}

void VNClassifier::ComputeDescriptors(
	Mesh *pConvexHull,
	float *R,
	float *d,
	uchar *bd,
	Point **pVertex)
{
	Point *pPt0 = pConvexHull->NodeArray.Element + pConvexHull->iValidVertices.Element[0];

	float d0 = -pPt0->P[2];

	int j;
	Point *pPt;
	float *P;

	for (j = 1; j < pConvexHull->iValidVertices.n; j++)
	{
		pPt = pConvexHull->NodeArray.Element + pConvexHull->iValidVertices.Element[j];

		P = pPt->P;

		if (-P[2] > d0)
		{
			d0 = -P[2];

			pPt0 = pPt;
		}
	}

	Pair<int, int> *tree = CTIDescriptorMapingData.tree;

	int rollBlockSize = 4 * CTIDescriptorMapingData.nRoll;

	pVertex[tree[0].a] = pPt0;

	int iChild, id, iRoll, iPt2, iD;
	float *NM;
	float *R__;
	float NS[3];
	float dmax, d_;
	Point *pPtNext, *pPt_;
	MeshEdgePtr *pEdgePtr;

	for (j = 0; j < rollBlockSize; j++)
	{
		pPt = pVertex[tree[j].a];

		iChild = tree[j].b;

		id = iChild % sampledUnitSphere.h;

		iRoll = iChild / sampledUnitSphere.h;

		NM = sampledUnitSphere.Element + 3 * id;

		R__ = R + 9 * iRoll;

		RVLMULMX3X3VECT(R__, NM, NS);

		dmax = RVLDOTPRODUCT3(NS, pPt->P);

		pPtNext = pPt;

		while (pPtNext)
		{
			pPtNext = NULL;

			pEdgePtr = pPt->EdgeList.pFirst;

			while (pEdgePtr)
			{
				iPt2 = RVLPCSEGMENT_GRAPH_GET_OPPOSITE_NODE(pEdgePtr);

				pPt_ = pConvexHull->NodeArray.Element + iPt2;

				if (pPt_->bValid)
				{
					d_ = RVLDOTPRODUCT3(NS, pPt_->P);

					if (d_ > dmax)
					{
						dmax = d_;

						pPtNext = pPt_;

						break;
					}
				}

				pEdgePtr = pEdgePtr->pNext;
			}

			if (pPtNext)
				pPt = pPtNext;
		}

		// Only for debugging purpose!

		//float dmax_ = -1e9;

		//for (i = 0; i < pConvexHull->iValidVertices.n; i++)
		//{
		//	P = pConvexHull->NodeArray.Element[pConvexHull->iValidVertices.Element[i]].P;

		//	d_ = RVLDOTPRODUCT3(NS, P);

		//	if (d_ > dmax_)
		//		dmax_ = d_;
		//}

		//if (dmax_ != dmax)
		//	int debug = 0;

		///

		iD = iRoll * sampledUnitSphere.h + id;

		d[iD] = dmax;

		bd[iD] = (dmax < 0.0f ? 1 : 0);

		pVertex[iChild] = pPt;
	}	// for (j = 0; j < nRollBlock; j++)
}

void VNClassifier::ComputeDescriptor(
	Array<int> iVertexArray,
	Array<SURFEL::NormalHullElement> NHull,
	float *R,
	float *d,
	uchar *bd,
	float *NS,
	int *iV,
	float o,
	bool bNormalTest,
	bool bVisibilityTest)
{
	float dmax = 0.0f;
	float dmaxValid = 0.0f;

	int i;
	int iN;
	int iVertex;
	float *N_, *P;
	float *N;
	SURFEL::Vertex *pVertex;
	//SURFEL::Vertex *pTangentVertex;
	//float *d_;
	//uchar *bd_;
	float d__, distFromNHull;
	//uchar bd__;
	bool bdmax, bdmaxValid;

	for (iN = 0; iN < sampledUnitSphere.h; iN++)
	{
		//if (iN == 33)
		//	int debug = 0;

		N_ = sampledUnitSphere.Element + sampledUnitSphere.w * iN;

		N = NS + sampledUnitSphere.w * iN;

		RVLMULMX3X3VECT(R, N_, N);

		bdmax = bdmaxValid = false;

		for (i = 0; i < iVertexArray.n; i++)
		{
			iVertex = iVertexArray.Element[i];

			pVertex = pSurfels->vertexArray.Element[iVertex];

			P = pVertex->P;

			d__ = RVLDOTPRODUCT3(N, P);

			if (bdmax)
			{
				if (o * d__ > o * dmax)
				{
					dmax = d__;

					iV[iN] = iVertex;
				}
			}
			else
			{
				dmax = d__;

				iV[iN] = iVertex;

				bdmax = true;
			}

			if (bNormalTest)
			{
				if (d__ < 0.0f || !bVisibilityTest)
				{
					distFromNHull = pSurfels->DistanceFromNormalHull(pVertex->normalHull, N);

					if (distFromNHull <= maxDistFromNHull)
					{
						if (bdmaxValid)
						{
							if (o * d__ > o * dmaxValid)
								dmaxValid = d__;
						}
						else
						{
							dmaxValid = d__;

							bdmaxValid = true;
						}
					}
				}
			}
		}	// for (i = 0; i < iVertexArray.n; i++)

		if (bdmaxValid)
			d[iN] = ((bd[iN] = (o * (dmax - dmaxValid) <= maxedmax)) ? dmaxValid : dmax);
		else
		{
			d[iN] = dmax;

			bd[iN] = 0x00;
		}

		//bd__ = 0x00;

		//if (pTangentVertex->normalHull.n >= 3)
		//{
		//	distFromNHull = pSurfels->DistanceFromNormalHull(pTangentVertex->normalHull, N);

		//	if (distFromNHull <= 0.0f)
		//		bd__ = 0x01;
		//}

		//if (bd__ == 0x00)
		//{
		//	distFromNHull = pSurfels->DistanceFromNormalHull(NHull, N);

		//	bd__ = (distFromNHull <= maxDistFromNHull);
		//}

		//bd[iN] = bd__;
	}
}

void VNClassifier::ComputeDescriptor(
	Mesh *pConvexHull,
	float *R,
	float *d,
	uchar *bd,
	Array<int> *piVertexArray,
	float o,
	int *iV)
{
	int i, iN;
	float *NM, *P;
	float NS[3];
	float dmax, d_;
	bool bFirst;

	if (piVertexArray)
	{
		dmax = 0.0f;
		float dmaxValid = 0.0f;

		int iVertex;
		SURFEL::Vertex *pVertex;
		float d__;
		bool bdmax, bdmaxValid;

		for (iN = 0; iN < sampledUnitSphere.h; iN++)
		{
			//if (iN == 22)
			//	int debug = 0;

			NM = sampledUnitSphere.Element + sampledUnitSphere.w * iN;

			RVLMULMX3X3VECT(R, NM, NS);

			bdmax = bdmaxValid = false;

			for (i = 0; i < pConvexHull->iValidVertices.n; i++)
			{
				iVertex = piVertexArray->Element[pConvexHull->iValidVertices.Element[i]];

			//for (i = 0; i < piVertexArray->n; i++)
			//{
			//	iVertex = piVertexArray->Element[i];

				pVertex = pSurfels->vertexArray.Element[iVertex];

				P = pVertex->P;

				d__ = RVLDOTPRODUCT3(NS, P);

				if (bdmax)
				{
					if (o * d__ > o * dmax)
					{
						dmax = d__;

						if (iV)
							iV[iN] = iVertex;
					}
				}
				else
				{
					dmax = d__;

					if (iV)
						iV[iN] = iVertex;

					bdmax = true;
				}

				if (piVertexArray)
				{
					if (d__ < 0.0f)
					{
						if (!(pVertex->type & RVLSURFELVERTEX_TYPE_OCCLUSION))
						{
							if (bdmaxValid)
							{
								if (o * d__ > o * dmaxValid)
									dmaxValid = d__;
							}
							else
							{
								dmaxValid = d__;

								bdmaxValid = true;
							}
						}
					}
				}
			}	// for every vertex

			if (bdmaxValid)
				d[iN] = ((bd[iN] = (o * (dmax - dmaxValid) <= maxedmax)) ? dmaxValid : dmax);
			else
			{
				d[iN] = dmax;

				bd[iN] = 0x00;
			}
		}	// for every component of the convex template
	}
	else
	{
		for (iN = 0; iN < sampledUnitSphere.h; iN++)
		{
			NM = sampledUnitSphere.Element + sampledUnitSphere.w * iN;

			RVLMULMX3X3VECT(R, NM, NS);

			bFirst = true;

			for (i = 0; i < pConvexHull->iValidVertices.n; i++)
			{
				P = pConvexHull->NodeArray.Element[pConvexHull->iValidVertices.Element[i]].P;

				d_ = RVLDOTPRODUCT3(NS, P);

				if (bFirst)
				{
					dmax = d_;
					bFirst = false;
				}
				else if (d_ > dmax)
					dmax = d_;
			}

			d[iN] = dmax;
			bd[iN] = (dmax < 0.0f ? 1 : 0);
		}
	}
}

void VNClassifier::MergeDescriptors(
	float *dSrc1,
	uchar *bdSrc1,
	float *dSrc2,
	uchar *bdSrc2,
	float o,
	float *dTgt,
	uchar *bdTgt)
{
	float *dSrc1_ = dSrc1;
	float *dSrc2_ = dSrc2;
	float *dTgt_ = dTgt;
	
	uchar *bdSrc1_ = bdSrc1;
	uchar *bdSrc2_ = bdSrc2;
	uchar *bdTgt_ = bdTgt;

	int i, j;

	for (i = 0; i < SO3Samples.c; i++)
	{
		for (j = 0; j < sampledUnitSphere.h; j++)
		{
			if (o * dSrc1_[j] >= o * dSrc2_[j])
			{
				dTgt_[j] = dSrc1_[j];
				bdTgt_[j] = bdSrc1_[j];
			}
			else
			{
				dTgt_[j] = dSrc2_[j];
				bdTgt_[j] = bdSrc2_[j];
			}
		}

		dSrc1_ += sampledUnitSphere.h;
		dSrc2_ += sampledUnitSphere.h;
		dTgt_  += sampledUnitSphere.h;

		bdSrc1_ += sampledUnitSphere.h;
		bdSrc2_ += sampledUnitSphere.h;
		bdTgt_  += sampledUnitSphere.h;
	}
}

void VNClassifier::Learn(
	char *modelSequenceFileName,
	int iClass,
	Visualizer *pVisualizer)
{
	float resolution = 0.01f;

	Eigen::MatrixXf A(3, 66);

	A = alignment.ConvexTemplatenT();

	int iMetaModel = (iClass < classArray.n ? classArray.Element[iClass].iMetaModel : -1);

	Mesh mesh;

	FileSequenceLoader modelsLoader;
	FileSequenceLoader dbLoader;

	char modelFilePath[200];
	char modelFileName[200];

	if (!modelDataBase)
		modelDataBase = "modelDB.dat";

	if (!modelsInDataBase)
		modelsInDataBase = "DBModels.txt";

	modelsLoader.Init(modelSequenceFileName);
	dbLoader.Init(modelsInDataBase);

	FILE *fp = fopen(modelDataBase, "a");

	char *VNModelDataBaseFileName = RVLCreateFileName(modelDataBase, ".dat", -1, ".vn.dat");

	FILE *fpVNDB = fopen(VNModelDataBaseFileName, "w");

	bool saveDBSequenceFile = false;

	printf("Model DB creation started...\n");

	unsigned char color[] = { 0, 128, 255 };

	float *dS = NULL;
	bool *bdS = NULL;

	int currentModelID;
	Box<float> SBoundingBox;
	VN *pModel;
	float R[9];
	float t[3];
	char modelFilePath_[200];
	char modelFileName_[200];
	char *modelDescriptorFileName;
	int modelID, modelID_;
	FILE *fpDescriptor;

	int nMetamodels;
	FILE *fpMetamodelCost = fopen("MetamodelCost.txt", "w");


	nMetamodels = 5;
	int MetaModelIDsArray[] = { 0, 2, 3, 6, 8 };
	float MetaModelScore[] = { 0, 0, 0, 0, 0 };

	while (modelsLoader.GetNext(modelFilePath, modelFileName))
	{
		

		printf("\nProcessing model %s!\n", modelFileName);

		
		if (LoadMesh(vpMeshBuilder, modelFilePath, &mesh, false))
		{
			FILE *fp = fopen("C:\\RVL\\Debug\\P.txt", "w");

			int i;
			float *P;

			for (i = 0; i < mesh.NodeArray.n; i++)
			{
				P = mesh.NodeArray.Element[i].P;

				fprintf(fp, "%f\t%f\t%f\n", P[0], P[1], P[2]);
			}

			fclose(fp);

			// Determine the model ID of the model defined in file modelFileName.

			while (dbLoader.GetNext(modelFilePath_, modelFileName_, &modelID_))
				if (strcmp(modelFileName, modelFileName_) == 0)
				{
					modelID = modelID_;

					break;
				}

			// If the model is not found, then R <- unit matrix and t <- null vector, otherwise align model with the reference model of this class.

			if (modelID < 0)
			{
				printf("CTI of the considered model is not available!\n");

				RVLUNITMX3(R);
				RVLNULL3VECTOR(t);
			}
			else
			{
				printf("CTI alignment.\n");

				alignment.ObjectAlignment(alignment.MCTISet.SegmentCTIs.Element[modelID],
					alignment.MCTISet.pCTI.Element,
					alignment.MCTISet.SegmentCTIs.Element[classArray.Element[iClass].iRefInstance],
					alignment.MCTISet.pCTI.Element, A, R, t, true);
			}

			if (trainingMethod == 0)
			{


				if (bLearnMetamodels)
				{


					std::vector<std::vector<float>> metaModelScorePerModelArr;
					float score=-1;



					for (int iMM = 0; iMM < nMetamodels; iMM++)
					{
						//metaModelScorePerModelArr[iMM].push_back(ComputeDescriptor(&mesh, R, t, dS, bdS, SBoundingBox, MetaModelIDsArray[i], score));
						MetaModelScore[iMM] += ComputeDescriptor(&mesh, R, t, dS, bdS, SBoundingBox, MetaModelIDsArray[iMM], score);

						if (bVisualization)
						{
							if (pVisualizer)
							{
								pModel = models[MetaModelIDsArray[iMM]];

								pVisualizer->renderer->RemoveAllViewProps();

								pModel->Display(pVisualizer, 0.01f, dS, bdS, visualizationData.SDFSurfaceValue);

								pVisualizer->Run();
							}
						}
					}



				}
			}
		}
	}

	fprintf(fpMetamodelCost, "%f\t%f\t%f\t%f\t%f\n", MetaModelScore[0], MetaModelScore[1], MetaModelScore[2], MetaModelScore[3], MetaModelScore[4]);

	float min = MetaModelScore[0];
	int idMin = MetaModelIDsArray[0];
	for (int iMM = 1; iMM < nMetamodels; iMM++)
	{
		if (MetaModelScore[iMM] < min)
		{
			min = MetaModelScore[iMM];
			idMin = MetaModelIDsArray[iMM];
		}
			
	}
	fclose(fpMetamodelCost);
	iMetaModel = idMin;


	modelsLoader.ResetID();
	

	while (modelsLoader.GetNext(modelFilePath, modelFileName))
	{
		// Process mesh.

		//mesh.LoadPolyDataFromPLY(modelFilePath);

		//vtkSmartPointer<vtkPolyData> polyData = MESH::CreateVisibleSurfaceMesh(mesh.pPolygonData, 0.01f, 5);

		//// Create a mapper and actor.
		//vtkSmartPointer<vtkPolyDataMapper> mapper = vtkSmartPointer<vtkPolyDataMapper>::New();
		//mapper->SetInputData(polyData);
		//vtkSmartPointer<vtkActor> actor = vtkSmartPointer<vtkActor>::New();
		//actor->SetMapper(mapper);

		//pVisualizer->renderer->AddActor(actor);

		//pVisualizer->Run();

		///

		//if (ModelExistInDB(modelFileName, dbLoader))
		//	continue;

		printf("\nProcessing model %s!\n", modelFileName);

		//saveDBSequenceFile = true;

		if (LoadMesh(vpMeshBuilder, modelFilePath, &mesh, false))
		{
			FILE *fp = fopen("C:\\RVL\\Debug\\P.txt", "w");

			int i;
			float *P;

			for (i = 0; i < mesh.NodeArray.n; i++)
			{
				P = mesh.NodeArray.Element[i].P;

				fprintf(fp, "%f\t%f\t%f\n", P[0], P[1], P[2]);
			}

			fclose(fp);

			// Check if there exists a label file.

			char *labelFileName = RVLCreateFileName(modelFilePath, ".ply", -1, "_.seg");

			mesh.LoadLabels(labelFileName);

			delete[] labelFileName;

			currentModelID = dbLoader.GetLastModelID() + 1;

			pMem->Clear();

			dbLoader.ResetID();

			modelID = -1;

			// Determine the model ID of the model defined in file modelFileName.

			while (dbLoader.GetNext(modelFilePath_, modelFileName_, &modelID_))
				if (strcmp(modelFileName, modelFileName_) == 0)
				{
					modelID = modelID_;

					break;
				}

			// If the model is not found, then R <- unit matrix and t <- null vector, otherwise align model with the reference model of this class.

			if (modelID < 0)
			{
				printf("CTI of the considered model is not available!\n");

				RVLUNITMX3(R);
				RVLNULL3VECTOR(t);
			}
			else
			{
				printf("CTI alignment.\n");

				alignment.ObjectAlignment(alignment.MCTISet.SegmentCTIs.Element[modelID],
					alignment.MCTISet.pCTI.Element,
					alignment.MCTISet.SegmentCTIs.Element[classArray.Element[iClass].iRefInstance],
					alignment.MCTISet.pCTI.Element, A, R, t, true);
			}

			if (trainingMethod == 0)
			{

				ComputeDescriptor(&mesh, R, t, dS, bdS, SBoundingBox, iMetaModel);


				

				dbLoader.AddModel(currentModelID, modelFilePath, modelFileName);

				modelDescriptorFileName = RVLCreateFileName(modelFilePath, ".ply", -1, ".vnd");

				fpDescriptor = fopen(modelDescriptorFileName, "w");

				delete[] modelDescriptorFileName;

				SaveDescriptor(fpDescriptor, dS, bdS, modelID_, iMetaModel);

				SaveDescriptor(fpVNDB, dS, bdS, modelID_, iMetaModel);

				fclose(fpDescriptor);

				if (bVisualization)
				{
					if (pVisualizer)
					{
						pModel = models[iMetaModel];

						//ExpandBox<float>(&SBoundingBox, 10.0f * resolution);

						pVisualizer->renderer->RemoveAllViewProps();

						//pModel->Display(pVisualizer, SBoundingBox, visualizationData.resolution, dS, bdS, visualizationData.SDFSurfaceValue);
						pModel->Display(pVisualizer, 0.01f, dS, bdS, visualizationData.SDFSurfaceValue);

						//pVisualizer->DisplayPointSet<float, MESH::Sample>(sceneObject.sampleArray, color, 6.0f);

						pVisualizer->Run();
					}
				}

				RVL_DELETE_ARRAY(dS);
				RVL_DELETE_ARRAY(bdS);
			}
			else if (trainingMethod == 1)
			{
				ComputeDescriptor(&mesh, R, t, dS, bdS, SBoundingBox, iMetaModel);
			}
			else if (trainingMethod == 2)
			{
				SegmentToParts(&mesh);
			}
		}
		else
			printf("VN descriptor cannot be computed for this mesh!\n");
	}

	fclose(fpVNDB);

	printf("Model DB creation completed!\n");

	if (saveDBSequenceFile)
		SaveModelID(dbLoader, modelsInDataBase);

	fclose(fp);
}

void VNClassifier::LearnPrimitives(char *primitiveFileName)
{
	Eigen::MatrixXf A(3, 66);

	A = alignment.ConvexTemplatenT();

	// Load primitives from file and create convex hulls.

	FILE *fp = fopen(primitiveFileName, "r");

	if (fp == NULL)
	{
		printf("ERROR: Can't open %s\n", primitiveFileName);

		return;
	}

	printf("Loading primitives from %s...", primitiveFileName);

	Array<Mesh> convexHulls;

	fscanf(fp, "%d\n", &(convexHulls.n));

	if (convexHulls.n <= 0)
	{
		fclose(fp);

		return;
	}

	memConvexHull.Create(100000000);

	convexHullVertices.w = 3;

	convexHulls.Element = new Mesh[convexHulls.n];

	int iPrimitive, iPt, nPts;
	char tempString[100];
	float *P;
	Mesh *pConvexHull;

	for (iPrimitive = 0; iPrimitive < convexHulls.n; iPrimitive++)
	{
		fscanf(fp, "%d\t%s\t%d\t%d\n", &convexHulls.Element[iPrimitive].idPrimitive, tempString, &convexHulls.Element[iPrimitive].idCluster, &nPts);

		RVLCopyString(tempString, &convexHulls.Element[iPrimitive].name);

		convexHullVertices.Element = new float[3 * nPts];

		convexHullVertices.h = nPts;

		if (nPts > 0)
		{
			for (iPt = 0; iPt < nPts; iPt++)
			{
				P = convexHullVertices.Element + 3 * iPt;

				fscanf(fp, "%f\t%f\t%f\n", P, P + 1, P + 2);
			}
		}

		pConvexHull = convexHulls.Element + iPrimitive;

		pConvexHull->ConvexHull(convexHullVertices, &memConvexHull, false);

		delete[] convexHullVertices.Element;
	}

	fclose(fp);

	printf("completed.\n");

	int nCT = alignment.convexTemplate.n;

	int i, iCTI;

	if (bGenerateModelCTIs)
	{
		// Generate primitive CTIs.

		FILE *fp = fopen(modelDataBase, "w");

		Array<int> iCTIArray;

		for (iPrimitive = 0; iPrimitive < convexHulls.n; iPrimitive++)
		{
			printf("%d\n", iPrimitive);

			alignment.ReferenceFrames(convexHulls.Element + iPrimitive);

			alignment.CTISet.pCTI.Element = new PSGM_::ModelInstance *[alignment.CTISet.pCTI.n];

			QLIST::CreatePtrArray<PSGM_::ModelInstance>(&(alignment.CTISet.CTI), &(alignment.CTISet.pCTI));

			iCTIArray.Element = new int[alignment.CTISet.pCTI.n];

			for (iCTI = 0; iCTI < alignment.CTISet.pCTI.n; iCTI++)
				iCTIArray.Element[iCTI] = iCTI;

			iCTIArray.n = alignment.CTISet.pCTI.n;

			//if (iPrimitive == 19)
			//	int debug = 0;

			iCTI = alignment.ReferenceFrame(convexHulls.Element + iPrimitive, iCTIArray, alignment.CTISet.pCTI.Element);

			alignment.SaveCTI(fp, alignment.CTISet.pCTI.Element[iCTI], iPrimitive);

			delete[] alignment.CTISet.pCTI.Element;
			delete[] iCTIArray.Element;
		}

		fclose(fp);
	}
	else
	{
		// Create translational subspace.

		float alpha;

		float *MT = alignment.CreateTranslationalSubspace(alpha);
		
		// D <- primitive CTIs in the cannonical form.

		int nPrimitives = alignment.MCTISet.pCTI.n;

		float *d = new float[nCT];

		float *D = new float[nPrimitives * nCT];

		PSGM_::ModelInstanceElement *CTIDesc;
		float beta;
		float q[4];

		for (iPrimitive = 0; iPrimitive < nPrimitives; iPrimitive++)
		{
			CTIDesc = alignment.MCTISet.pCTI.Element[iPrimitive]->modelInstance.Element;

			for (i = 0; i < nCT; i++)
				d[i] = CTIDesc[i].d;

			alignment.CreateSubpace(d, MT, alpha, beta, q);

			memcpy(D + iPrimitive * nCT, MT + 3 * nCT, nCT * sizeof(float));
		}

		// Create alignment tree.

		MSTree MST;

		MST.Init();

		float **RArray = new float *[nPrimitives];
		float **tArray = new float *[nPrimitives];
		int **iRArray = new int *[nPrimitives];
		float **sArray = new float *[nPrimitives];

		int iPrimitive_;
		float *E;
		float *R;
		int *RP, *iD;
		int *iR;
		float *t;
		float *s;
		int iR_;
		float *d_, *d__;
		float cs, e;

		if (bCreateAlignmentTree)
		{
			// Align primitives.
			//alignment.LoadModelDataBase();

			for (iPrimitive = 0; iPrimitive < nPrimitives; iPrimitive++)
			{
				if (iPrimitive > 0)
				{
					E = new float[iPrimitive];

					R = new float[9 * iPrimitive];

					t = new float[3 * iPrimitive];

					iR = new int[iPrimitive];

					s = new float[iPrimitive];

					d__ = D + iPrimitive * nCT;

					for (iPrimitive_ = 0; iPrimitive_ < iPrimitive; iPrimitive_++)
					{
						//if (iPrimitive == 990 && iPrimitive_ == 103)
						//	int debug = 0;

						// Approach 0:

						//E[iPrimitive_] = -Align(convexHulls.Element[iPrimitive_].faces, convexHulls.Element[iPrimitive].faces, R + 9 * iPrimitive_);

						// Approach 1:

							//alignment.ObjectAlignment(alignment.MCTISet.SegmentCTIs.Element[iPrimitive],
							//alignment.MCTISet.pCTI.Element,
							//alignment.MCTISet.SegmentCTIs.Element[iPrimitive_],
							//alignment.MCTISet.pCTI.Element, A, R + 9 * iPrimitive_, t + 3 * iPrimitive_, false, E + iPrimitive_);

						// Approach 2:

						if (iPrimitive >= 798 && iPrimitive_ >= 798)
						{
							E[iPrimitive_] = 0.0f;
							iR[iPrimitive_] = 0;
						}
						else
						{
							d_ = D + iPrimitive_ * nCT;

							for (iR_ = 0; iR_ < alignment.CTIDescMap.nCTI; iR_++)
							{
								iD = alignment.CTIDescMap.iD + nCT * iR_;

								for (i = 0; i < nCT; i++)
									d[i] = d_[iD[i]];

								RVLDOTPRODUCT(d, d__, nCT, cs, i);

								e = 1.0f - cs * cs;

								if (iR_ == 0 || e < E[iPrimitive_])
								{
									E[iPrimitive_] = e;

									iR[iPrimitive_] = iR_;
								}
							}
						}
					}

					RArray[iPrimitive] = R;
					tArray[iPrimitive] = t;
					iRArray[iPrimitive] = iR;
					sArray[iPrimitive] = s;
				}
				else
					E = NULL;

				MST.Update(E);
				printf("Adding primitive: %d\n", iPrimitive);
			}
		}
		else
		{
			// Load alignment tree.


		}
	
		// Align primitives.

		int *RAligned = new int[9 * nPrimitives];

		int iRefPrimitive = 0;

		MST.T.OrientTreeEdges(iRefPrimitive);

		int *queue = new int[nPrimitives];

		int *pPut = queue;

		*(pPut++) = iRefPrimitive;

		RP = RAligned + 9 * iRefPrimitive;

		RVLUNITMX3(RP);

		int *pFetch = queue;

		GRAPH::EdgePtr<GRAPH::Edge> *pEdgePtr;
		int *RP_, *RPP;

		while (pFetch < pPut)
		{
			iPrimitive = *(pFetch++);

			RP = RAligned + 9 * iPrimitive;

			pEdgePtr = MST.T.NodeArray.Element[iPrimitive].EdgeList.pFirst;

			while (pEdgePtr)
			{
				if (pEdgePtr->pEdge->iVertex[0] == iPrimitive)
				{
					iPrimitive_ = pEdgePtr->pEdge->iVertex[1];

					*(pPut++) = iPrimitive_;

					RP_ = RAligned + 9 * iPrimitive_;

					if (iPrimitive < iPrimitive_)
					{
						iR_ = iRArray[iPrimitive_][iPrimitive];

						RPP = alignment.CTIDescMap.R + 9 * iR_;

						RVLMXMUL3X3(RP, RPP, RP_);
					}
					else
					{
						iR_ = iRArray[iPrimitive][iPrimitive_];

						RPP = alignment.CTIDescMap.R + 9 * iR_;

						RVLMXMUL3X3T2(RP, RPP, RP_);
					}
				}

				pEdgePtr = pEdgePtr->pNext;
			}
		}

		delete[] queue;
		
		// Create aligned CTIs.

		float *DAligned = new float[nCT * nPrimitives];

		int R_[9];

		for (iPrimitive = 0; iPrimitive < nPrimitives; iPrimitive++)
		{
			RP = RAligned + 9 * iPrimitive;

			d = D + nCT * iPrimitive;

			d_ = DAligned + nCT * iPrimitive;

			for (iR_ = 0; iR_ < alignment.CTIDescMap.nCTI; iR_++)
			{
				RP_ = alignment.CTIDescMap.R + 9 * iR_;

				for (i = 0; i < 9; i++)
					if (RP[i] != RP_[i])
						break;

				if (i >= 9)
					break;
			}

			//if (iR_ >= alignment.CTIDescMap.nCTI)
			//	int debug = 0;

			iD = alignment.CTIDescMap.iD + nCT * iR_;

			for (i = 0; i < nCT; i++)
				d_[iD[i]] = d[i];
		}

		// Save aligned CTI database.

		uchar *bAllValid = new uchar[nCT];

		for (i = 0; i < nCT; i++)
			bAllValid[i] = true;

		FILE *fpAlignedCTIDB = fopen(primitiveDataBase, "w");

		float fR[9];

		RVLUNITMX3(fR);

		for (iPrimitive = 0; iPrimitive < nPrimitives; iPrimitive++)
		{
			RP = RAligned + 9 * iPrimitive;
			
			SaveDescriptor(fpAlignedCTIDB, DAligned + nCT * iPrimitive, bAllValid, fR, iPrimitive, 0);
		}
		
		fclose(fpAlignedCTIDB);

		// Visualization.

		float t_[3];

		RVLNULL3VECTOR(t_);

		//float R_[9], R__[9], t_[3], t__[3];
		int R__[9];
		int iPrimitive__;
		float s_;
		float M3x3Tmp[9], V3Tmp[3];
		if (bVisualization)
		{
			Visualizer visualizer;

			visualizer.Create();

			double color[3] = { 1, 1, 1 };

			bool bContinue = true;
			while (bContinue)
			{
				printf("Enter primitive1 ID: ");

				scanf("%d", &iRefPrimitive);

				if (iRefPrimitive < 0)
					break;

				printf("Enter primitive2 ID: ");

				scanf("%d", &iPrimitive);

				if (iPrimitive < 0 || iPrimitive == iRefPrimitive)
					break;

				RVLUNITMX3(R_);

				RVLCOPYMX3X3(R_, fR);

				alignment.DisplayCTI(&visualizer, sampledUnitSphere.Element, sampledUnitSphere.h, DAligned + iRefPrimitive * nCT, fR, t_, NULL, color);

				//RP = RAligned + 9 * iRefPrimitive;

				//RP_ = RAligned + 9 * iPrimitive;

				//RVLMXMUL3X3T1(RP, RP_, R_);

				//MST.T.OrientTreeEdges(iRefPrimitive);

				//iPrimitive_ = iPrimitive;

				//while (iPrimitive_ != iRefPrimitive)
				//{
				//	pEdgePtr = MST.T.NodeArray.Element[iPrimitive_].EdgeList.pFirst;

				//	while (pEdgePtr)
				//	{
				//		if (pEdgePtr->pEdge->iVertex[0] != iPrimitive_)
				//			break;

				//		pEdgePtr = pEdgePtr->pNext;
				//	}

				//	iPrimitive__ = pEdgePtr->pEdge->iVertex[0];

				//	if (iPrimitive__ < iPrimitive_)
				//	{
				//		iR_ = iRArray[iPrimitive_][iPrimitive__];

				//		RP = alignment.CTIDescMap.R + 9 * iR_;

				//		RVLMXMUL3X3(RP, R_, R__);

				//		//R = RArray[iPrimitive_] + 9 * iPrimitive__;
				//		//t = tArray[iPrimitive_] + 3 * iPrimitive__;

				//		//RVLMXMUL3X3T1(R, R, M3x3Tmp);

				//		//s = (M3x3Tmp[0] + M3x3Tmp[4] + M3x3Tmp[8]) / 3.0f;

				//		//RVLINVTRANSF3D(R, t, R__, t__);

				//		//RVLSCALEMX3X32(R__, s, M3x3Tmp);
				//		//RVLSCALE3VECTOR2(t__, s, V3Tmp);

				//		//RVLCOMPTRANSF3D(M3x3Tmp, V3Tmp, R_, t_, R__, t__);
				//	}
				//	else
				//	{
				//		iR_ = iRArray[iPrimitive__][iPrimitive_];

				//		RP = alignment.CTIDescMap.R + 9 * iR_;

				//		RVLMXMUL3X3T1(RP, R_, R__);

				//		//R = RArray[iPrimitive__] + 9 * iPrimitive_;
				//		//t = tArray[iPrimitive__] + 3 * iPrimitive_;

				//		//RVLCOMPTRANSF3D(R, t, R_, t_, R__, t__);
				//	}

				//	RVLCOPYMX3X3(R__, R_);

				//	iPrimitive_ = iPrimitive__;
				//}

				//int iCTI;

				//int iCTI = alignment.ObjectAlignment(convexHulls.Element + iPrimitive, alignment.MCTISet.SegmentCTIs.Element[iPrimitive],
				//	alignment.MCTISet.pCTI.Element);		

				//alignment.OptimalTranslationAndScale(alignment.MCTISet.pCTI.Element[alignment.MCTISet.SegmentCTIs.Element[0].Element[0]], 
				//	alignment.MCTISet.pCTI.Element[alignment.MCTISet.SegmentCTIs.Element[iPrimitive].Element[iCTI]], M, alpha, t_, s_);

				//float *RBox = alignment.MCTISet.pCTI.Element[alignment.MCTISet.SegmentCTIs.Element[iPrimitive].Element[iCTI]]->R;

				//RVLSCALEMX3X3(RBox, s_, R__);

				//RVLTRASPOSE3X3(R__, R_)

				RVLCOPYMX3X3(R_, fR);

				alignment.DisplayCTI(&visualizer, sampledUnitSphere.Element, sampledUnitSphere.h, DAligned + iPrimitive * nCT, fR, t_);

				visualizer.Run();

				visualizer.renderer->RemoveAllViewProps();
			}
		}

		//

		delete[] RAligned;
		delete[] DAligned;
		
		//Save edges and costs to file.

		FILE *fpEdges = fopen((std::string(resultsFolder) + "\\VNEdges.txt").data(), "w");
		FILE *fpRot = fopen((std::string(resultsFolder) + "\\VNRot.txt").data(), "w");
		FILE *fpCost = fopen((std::string(resultsFolder) + "\\VNCost.txt").data(), "w");
		RVLUNITMX3(R_);
		int temp;
		for (iPrimitive = 0; iPrimitive < nPrimitives; iPrimitive++)
		{			
			for (iPrimitive_ = 0; iPrimitive_ < iPrimitive; iPrimitive_++)
			{
				fprintf(fpCost, "%f\n", MST.E[iPrimitive][iPrimitive_]);
			}

			pEdgePtr = MST.T.NodeArray.Element[iPrimitive].EdgeList.pFirst;
			while (pEdgePtr)
			{
				if (pEdgePtr->pEdge->iVertex[0] == iPrimitive)
				{
					iPrimitive__ = pEdgePtr->pEdge->iVertex[0];
					iPrimitive_ = pEdgePtr->pEdge->iVertex[1];

					if (iPrimitive__ < iPrimitive_)
					{
						temp = iPrimitive__;
						iPrimitive__ = iPrimitive_;
						iPrimitive_ = temp;
					}

					R = RArray[iPrimitive__] + 9 * iPrimitive_;

					//RVLMXMUL3X3(R, R_, R__);

					fprintf(fpRot, "%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\n", R[0], R[1], R[2], R[3], R[4], R[5], R[6], R[7], R[8]);

					fprintf(fpEdges, "%d %d\n", iPrimitive__, iPrimitive_ );
				}

				pEdgePtr = pEdgePtr->pNext;
			}

		}
		fclose(fpEdges);
		fclose(fpCost);
		fclose(fpRot);


		//Read edges and costs from file:

		/*FILE *fpEdges = fopen((std::string(resultsFolder) + "\\VNEdges.txt").data(), "r");
		FILE *fpCost = fopen((std::string(resultsFolder) + "\\VNCost.txt").data(), "r");
		MST.Init();
		for (iPrimitive = 0; iPrimitive < convexHulls.n; iPrimitive++)
		{
		if (iPrimitive > 0)
		{
		E = new float[iPrimitive];

		R = new float[9 * iPrimitive];

		for (iPrimitive_ = 0; iPrimitive_ < iPrimitive; iPrimitive_++)
		fscanf(fpCost, "%f\t", &E[iPrimitive_]);

		}
		else
		E = NULL;

		MST.Update(E);
		}

		int iRefPrimitive = 0;

		MST.T.OrientTreeEdges(iRefPrimitive);
		*/

		// Free memory.

		for (iPrimitive = 1; iPrimitive < nPrimitives; iPrimitive++)
		{
			delete[] RArray[iPrimitive];
			delete[] tArray[iPrimitive];
			delete[] iRArray[iPrimitive];
			delete[] sArray[iPrimitive];
		}
		delete[] RArray;
		delete[] tArray;
		delete[] iRArray;
		delete[] sArray;
		delete[] MT;
		delete[] D;
		delete[] d;
	}	// if (!bGenerateModelCTIs)

	// Free memory.

	delete[] convexHulls.Element;
}

void VNClassifier::Interpret(
	Mesh *pMesh,
	int iClassIn)
{


#ifdef RVLVN_TIME_MESUREMENT
	LARGE_INTEGER CNTRSegmentationSTART, CNTRSegmentationEND;
	LARGE_INTEGER CNTRDescriptorGenerationSTART, CNTRDescriptorGenerationEND;

	LARGE_INTEGER frequency;
	QueryPerformanceFrequency((LARGE_INTEGER *)&frequency);
#endif

#ifdef RVLVN_TIME_MESUREMENT
	//Timer START
	QueryPerformanceCounter((LARGE_INTEGER *)&CNTRSegmentationSTART);
#endif
	// Determine the greatest number of latent variables per class.

	int maxnLatentVariables = 0;

	ClassData *pVNClass;
	int iClass;

	for (iClass = 0; iClass < classArray.n; iClass++)
	{
		pVNClass = classArray.Element + iClass;

		if (pVNClass->M.Element == NULL)
			continue;

		if (pVNClass->M.h > maxnLatentVariables)
			maxnLatentVariables = pVNClass->M.h;
	}

	if (maxnLatentVariables == 0)
		return;

	// Detect ground plane.

	bGnd = false;

	Array<int> groundPlaneSurfelArray;

	groundPlaneSurfelArray.Element = NULL;
	groundPlaneSurfelArray.n = 0;

	if (b3DNetTestSet)
	{
		_3DNetGround(alignment.sceneFileName, NGnd, dGnd);

		bGnd = true;

		pSurfelDetector->bGnd = true;
		RVLCOPY3VECTOR(NGnd, pSurfelDetector->NGnd);
		pSurfelDetector->dGnd = dGnd;
	}

	// Detect surfels.

	pSurfels->Init(pMesh);

	pSurfelDetector->Init(pMesh, pSurfels, pMem);

	printf("Segmentation to surfels...");

	pSurfelDetector->Segment(pMesh, pSurfels);

	printf("completed.\n");

	int nSurfels = pSurfels->NodeArray.n;

	printf("No. of surfels = %d\n", nSurfels);

	// Relations between adjacent surfels.

#ifdef RVLSURFEL_IMAGE_ADJACENCY
	pSurfels->SurfelRelations(pMesh);
#endif

	// Detect ground plane.

	if (pSurfels->bGroundContactVertices && !bGnd)
	{
		groundPlaneSurfelArray.Element = new int[pSurfels->NodeArray.n];

		pSurfels->DetectDominantPlane(pMesh, groundPlaneSurfelArray, NGnd, dGnd);

		bGnd = true;
	}

	// Detect vertices.

	printf("Detect vertices.\n");

	pSurfels->DetectVertices(pMesh);

	/// Detect objects.

	float RSG[9];
	float tSG[3];
	float rVOI;

	if (b3DNetVOI)
		PSGM_::_3DNetVOI(RSG, tSG, rVOI);

	int iObject;

#ifdef RVLSURFEL_IMAGE_ADJACENCY
	if (b3DNetTestSet)
	{
		// Detect objects as connected surfel sets.

		pObjects->pMesh = pMesh;

		pObjects->CreateObjectsAsConnectedComponents(groundPlaneSurfelArray, connectedComponentMaxDist, connectedComponentMinSize);

		RVL_DELETE_ARRAY(groundPlaneSurfelArray.Element);

		// Sort objects.

		pObjects->nValidObjects = -1;
		pObjects->sortedObjectArray.n = -1;

		pObjects->SortObjects();

		// Assign vertices to objects.

		pObjects->GetVertices();

		// Detect objects in VOI

		uchar mask = RVLPCSEGMENT_OBJECT_FLAG_GND;
		uchar flags = 0x00;

		if (pObjects->b3DNetVOI)
		{
			pObjects->ObjectsInVOI(RSG, tSG, rVOI);

			mask |= RVLPCSEGMENT_OBJECT_FLAG_IN_VOI;
			flags |= RVLPCSEGMENT_OBJECT_FLAG_IN_VOI;
		}

		RECOG::GroupObjectsAccordingToCTIProximity(pMesh, pObjects, &alignment, 0.100f, flags, flags);

		// Get foreground object.

		iObject = pObjects->GetForegroundObject();

		if (iObject < 0)
			return;
	}	// if (b3DNetTestSet)	
	else
#endif
	{
		// Identify surfels in VOI.

		if (b3DNetVOI)
		{
			pSurfels->SurfelsInVOI(RSG, tSG, rVOI);

			convexClustering.clusteringSurfelFlags = convexClustering.clusteringSurfelFlagMask = RVLSURFEL_FLAG_VOI;
			concaveClustering.clusteringSurfelFlags = concaveClustering.clusteringSurfelFlagMask = RVLSURFEL_FLAG_VOI;
		}

		// Cluster surfels into convex and concave surfaces.

		RVL_DELETE_ARRAY(SClusters.Element);		

		RVL_DELETE_ARRAY(clusterMap);

		clusterMap = new int[pSurfels->NodeArray.n];

		int nCClusters, nUClusters;

		PSGM_::ConvexAndConcaveClusters(pMesh, pSurfels, pSurfelDetector, &convexClustering, &concaveClustering,
			SClusters, nCClusters, nUClusters, 0, clusterMap, -1, -1, true, 0,
			visualizationData.bVisualizeConvexClusters, visualizationData.bVisualizeConcaveClusters, visualizationData.bVisualizeSurfels);

		// Compute descriptors.

		int descriptorBlockSize = SO3Samples.c * sampledUnitSphere.h;

		int descriptorMemSize = descriptorBlockSize * (SClusters.n + nUClusters);

		float *d = new float[descriptorMemSize];

		float *d_ = d;

		uchar *bd = new uchar[descriptorMemSize];

		uchar *bd_ = bd;

		int iCluster;
		SceneCluster *pCluster;
		PSGM_::Cluster *pCluster_;
		float *dU_;
		uchar *bdU_;

		for (iCluster = 0; iCluster < SClusters.n; iCluster++)
		{
			pCluster = SClusters.Element + iCluster;

			pCluster_ = (PSGM_::Cluster *)(pCluster->vpCluster);

			if (pCluster->type == RVLVN_CLUSTER_TYPE_CONVEX)
			{
				ComputeDescriptors(pCluster_->iSurfelArray, pCluster_->iVertexArray, d_, bd_);

				pCluster_->d = d_;
				pCluster_->bd = bd_;
				pCluster_->dU = NULL;
				pCluster_->bdU = NULL;

				d_ += descriptorBlockSize;

				bd_ += descriptorBlockSize;
			}
			else if (pCluster->type == RVLVN_CLUSTER_TYPE_CONCAVE)
			{
				dU_ = d_ + descriptorBlockSize;

				bdU_ = bd_ + descriptorBlockSize;

				ComputeDescriptors(pCluster_->iSurfelArray, pCluster_->iVertexArray, d_, bd_, dU_, bdU_);

				pCluster_->d = d_;
				pCluster_->bd = bd_;
				pCluster_->dU = dU_;
				pCluster_->bdU = bdU_;

				d_ = dU_ + descriptorBlockSize;

				bd_ = bdU_ + descriptorBlockSize;
			}
		}

		//

		RVLMEM_ALLOC_STRUCT_ARRAY(pMem, float, descriptorBlockSize, d_);
		RVLMEM_ALLOC_STRUCT_ARRAY(pMem, uchar, descriptorBlockSize, bd_);

		PSGM_::Cluster *pCluster__[2];

		pCluster = SClusters.Element + 2;

		pCluster__[0] = (PSGM_::Cluster *)(pCluster->vpCluster);

		pCluster = SClusters.Element + 13;

		pCluster__[1] = (PSGM_::Cluster *)(pCluster->vpCluster);

		MergeDescriptors(pCluster__[0]->d, pCluster__[0]->bd, pCluster__[1]->d, pCluster__[1]->bd, 1.0f, d_, bd_);

		dU_ = pCluster__[1]->dU;
		bdU_ = pCluster__[1]->bdU;

#ifdef RVLVN_RECORD_RESULTS

		FILE *fp = fopen((std::string(resultsFolder) + "\\scenedescriptors.txt").data(), "w");

		SaveDescriptors(fp, pCluster__[0]->d, pCluster__[0]->bd, 2, RVLVN_CLUSTER_TYPE_CONVEX);

		SaveDescriptors(fp, pCluster__[1]->dU, pCluster__[1]->bdU, 13, RVLVN_CLUSTER_TYPE_CONCAVE);

		SaveDescriptors(fp, d_, bd_, SClusters.n, RVLVN_CLUSTER_TYPE_CONVEX);

		fclose(fp);
#endif
		//

		ClassData *pClass = classArray.Element + 3;

		float *qOpt = new float[pClass->M.h];

		int iR;

		float E = ProjectToLatentSubspace(d_, bd_, dU_, bdU_, pClass, qOpt, iR);

		delete[] d;
		delete[] bd;

		float *R_ = SO3Samples.Element + 9 * iR;

		Array<int> iSurfelArray;

		iSurfelArray.Element = new int[pSurfels->NodeArray.n];

		iSurfelArray.n = 0;

		Array<int> iVertexArray;

		iVertexArray.Element = new int[pSurfels->vertexArray.n];

		iVertexArray.n = 0;

		int i, j;
		
		for (i = 0; i < 2; i++)
		{
			for (j = 0; j < pCluster__[i]->iVertexArray.n; j++)
				iVertexArray.Element[iVertexArray.n++] = pCluster__[i]->iVertexArray.Element[j];

			for (j = 0; j < pCluster__[i]->iSurfelArray.n; j++)
				iSurfelArray.Element[iSurfelArray.n++] = pCluster__[i]->iSurfelArray.Element[j];
		}

		Camera camera;

		camera.fu = 525;
		camera.fv = 525;
		camera.uc = 320;
		camera.vc = 240;
		camera.w = pMesh->width;
		camera.h = pMesh->height;

		SURFEL::SceneSamples sceneSamples;

		pSurfels->SampleSurfelSet(pMesh, iSurfelArray, iVertexArray, camera, sceneSamples);

		Array<OrientedPoint> PtArray;

		PtArray.n = sceneSamples.imagePtArray.h;

		PtArray.Element = new OrientedPoint[PtArray.n];

		float *c;

		//memset(qOpt + 4, 0, 6 * sizeof(float));

		float R[9];
		float EHull;

		bool bGnd = fitParams.bGnd;

		fitParams.bInit = false;
		fitParams.bGnd = fitParams.bGnd2;

		VN *pModel = models[pClass->iMetaModel];

		pModel->FitRotLMCC(pSurfels, iVertexArray, sceneSamples, pClass, NULL, NULL, R_, fitParams, qOpt, R, EHull, false, &camera);

		float *dM = new float[pClass->M.w];

		fitParams.bInit = true;
		fitParams.bGnd = bGnd;

		RVLMULMXTVECT(pClass->M.Element, qOpt, pClass->M.h, pClass->M.w, dM, i, j, c);

		float t[3];

		RVLNULL3VECTOR(t);

		pModel->Project(dM, R, t, camera, sceneSamples.imagePtArray, PtArray);

		// Visualize all clusters.

		if (visualizationData.bVisualizeAllClusters)
		{
			uchar SelectionColor[] = { 0, 255, 0 };

			RandomColors(SelectionColor, visualizationData.clusterColor, SClusters.n);

			InitDisplay(visualizationData.pVisualizer, pMesh, SelectionColor);

			pSurfels->DisplayEdgeFeatures();

			DisplayClusters();	

			pSurfels->DisplayVertices();

			unsigned char color[] = { 0, 128, 255 };

			Array<Point> PtArray_;

			PtArray_.Element = new Point[PtArray.n];

			PtArray_.n = 0;

			int iPt;
			float *P_;
			OrientedPoint *pP;

			for (iPt = 0; iPt < PtArray.n; iPt++)
			{
				pP = PtArray.Element + iPt;

				if (pP->P[2] <= 2.0f)
				{
					P_ = PtArray_.Element[PtArray_.n++].P;

					RVLCOPY3VECTOR(pP->P, P_);
				}
			}

			visualizationData.pVisualizer->DisplayPointSet<float, Point>(PtArray_, color, 6.0f);

			visualizationData.pVisualizer->Run();

			//visualizationData.pVisualizer->renderer->RemoveAllViewProps();

			delete[] visualizationData.clusterColor;
			delete[] PtArray_.Element;
		}

		delete[] qOpt;
		delete[] iVertexArray.Element;
		delete[] PtArray.Element;
		delete[] dM;
	}	// if (!b3DNetTestSet)

	// Create CTIs.

	alignment.CTISet.Init();

	SURFEL::Object *pObject = pObjects->objectArray.Element + iObject;

	//RVLCOPY3VECTOR(NGnd, alignment.NGnd);
	//alignment.dGnd = dGnd;
	//alignment.bGnd = true;

	//alignment.CTIs(pObject->surfelList, pObject->iVertexArray, 0, iObject, &(alignment.CTISet), pMem, true);

	//alignment.CTISet.CopyCTIsToArray();

	//// Save model instances to a file.

	//printf("Save model instances to a file.\n");

	//char *PSGModelInstanceFileName = RVLCreateString(alignment.sceneFileName);

	//sprintf(PSGModelInstanceFileName + strlen(PSGModelInstanceFileName) - 3, "cti");

	//FILE *fp = fopen(PSGModelInstanceFileName, "w");

	//delete[] PSGModelInstanceFileName;

	//alignment.SaveModelInstances(fp); //Vidovic

	//fclose(fp);

	//// Align the scene object with a model.

	//alignment.pMesh = pMesh;

	//alignment.CreateClusterFromObject(pObjects, iObject);

	//int iModel;
	float R[9], t[3];

	//alignment.Classify(pMesh, pVNClass->iRefInstance, pVNClass->iRefInstance + 1, iModel, R, t);

	//FILE *fpR = fopen("R.txt", "w");

	//PrintMatrix<float>(fpR, R, 3, 3);

	//fclose(fpR);

	///

	//float s;
	//float R_[9];
	//float t_[3];
			
	//RVLEXTRACTSCALEFROMROT(R, s, R_);

	//float sSM = 1.0f / s;

	//RVLSCALE3VECTOR(t, sSM, t_);

	//float RSM[9];
	//float tSM[3];

	//RVLINVTRANSF3D(R_, t, RSM, tSM);

	//Array<TangentVertexCorrespondence> correspondences;

	//correspondences.Element = new TangentVertexCorrespondence[alignment.convexTemplate.n];

	//float score;
	//float RSM_[9];
	//float tSM_[3];

	//TangentAlignment(pSurfels, pObject->iVertexArray, sSM, alignment.MTGSet.A, refModel.d, R_, t_, 50.0f * sSM,
	//	score, correspondences, R_, t_);

	//RVLINVTRANSF3D(RSM_, tSM_, R_, t);
	//RVLSCALE3VECTOR(t, s, t);
	//RVLSCALE3VECTOR(t_, s, t);
	//RVLSCALEMX3X3(R_, s, R);

	//float *q0 = new float[pVNClass->M.h];

	//RVLCOPY3VECTOR(t, q0);

	//q0[3] = sqrt(RVLDOTPRODUCT3(R, R));

	//memset(q0 + 4, 0, (pVNClass->M.h - 4) * sizeof(float));

	Camera camera;

	camera.fu = 525;
	camera.fv = 525;
	camera.uc = 320;
	camera.vc = 240;
	camera.w = pMesh->width;
	camera.h = pMesh->height;

	Array<int> iSurfelArray;

	iSurfelArray.Element = new int[pSurfels->NodeArray.n];

	QLIST::CopyToArray(&(pObject->surfelList), &iSurfelArray);

	SURFEL::SceneSamples sceneSamples;

	pSurfels->SampleSurfelSet(pMesh, iSurfelArray, pObject->iVertexArray, camera, sceneSamples);

	delete[] iSurfelArray.Element;

	if (bGndConstraint)
	{
		sceneSamples.PGnd = new float[3 * pObject->iVertexArray.n];

		pSurfels->ProjectVerticesOntoGroundPlane(pObject->iVertexArray, NGnd, dGnd, sceneSamples.PGnd);
	}
	else
		sceneSamples.PGnd = NULL;

	//Rect<float> ROI;

	//pSurfels->GetDepthImageROI(pObject->iVertexArray, camera, ROI);

	//Array2D<float> imagePtArray;

	//SampleRect<float>(&ROI, 10.0f, 32, imagePtArray);

	fitParams.bInit = true;

	RVLNULL3VECTOR(t);

	float *q = new float[maxnLatentVariables];

	int iMostProbableClass = -1;

	float *qOpt = new float[maxnLatentVariables];

	bool bFirst = true;

	float bestScore = 0.0f; 

	FILE *fpInstances = fopen((std::string(resultsFolder) + "\\instances.txt").data(), "w");

	int iFirstClass, iLastClass;

	if (bSingleClass)
		iFirstClass = iLastClass = iClassIn;
	else
	{
		iFirstClass = 0;
		iLastClass = classArray.n - 1;
	}

	std::string resultsFolderName = std::string(alignment.resultsFolder);
	FILE *fpScore = fopen((resultsFolderName + "\\VNScore.txt").data(), "a");

	fitParams.fpScore = fpScore;

	int iMetaModel;
	VN *pModel;
	float score;
	float ROpt[9];

	for (iClass = iFirstClass; iClass <= iLastClass; iClass++)
	{
		pVNClass = classArray.Element + iClass;

		if (pVNClass->M.Element == NULL)
			continue;

#ifdef RVLVN_RECORD_RESULTS

		printf("Class %d\n", iClass);
#endif
		iMetaModel = pVNClass->iMetaModel;

		pModel = models[iMetaModel];
		
		//pModel->Transform(refModel.d, R_, t, d);

		//pModel->FitLM(pMesh, pSurfels, pObject->surfelList, pObject->iVertexArray, camera, pVNClass->M, R_, d, q, true);
		//pModel->FitRotLM(pMesh, pSurfels, pObject->surfelList, pObject->iVertexArray, camera, pVNClass, d, R_, fitParams, q, R__, true);
		score = pModel->FitRotLM(this, pMesh, pSurfels, pObject->iVertexArray, camera, sceneSamples, pVNClass, NULL, 
			fitParams, q, R, bVisualization, bVisualizeBestInstance);

		if (bFirst || score > bestScore)
		{
			bestScore = score;

			RVLCOPYMX3X3(R, ROpt);

			memcpy(qOpt, q, pVNClass->M.h * sizeof(float));

			iMostProbableClass = iClass;

			bFirst = false;
		}
#ifdef RVLVN_RECORD_RESULTS

		printf("\n");
#endif
		//float *d = new float[pVNClass->M.w];

		//int i, j;
		//float *a;

		//RVLMULMXTVECT(pVNClass->M.Element, q, pVNClass->M.h, pVNClass->M.w, d, i, j, a);

		SaveLatentVector(fpInstances, q, iClass);

		//delete[] d;
	}

	fclose(fpInstances);

	fprintf(fpScore, "\n");

	fclose(fpScore);
	
#ifdef RVLVN_RECORD_RESULTS

	//printf("Most probable class: %d\n\n\n", iMostProbableClass);
#endif
	hypClass[iMostProbableClass]++;

#ifdef RVLVN_RECORD_RESULTS

	//confusion matrix:
	printf("confusion matrix\n");
	for (int rb = 0; rb < 10; rb++)
	{
		printf("\n%d: %d", rb, hypClass[rb]);
	}

	FILE *fpHyp = fopen((resultsFolderName + "\\VNClassify.txt").data(), "a");	
	int command = 0;

	if (iMostProbableClass != iClassIn)
	{
		fprintf(fpHyp, "%s\n", alignment.sceneFileName);
	}
	fclose(fpHyp);
#endif
	Array<OrientedPoint> PtArray;

	PtArray.n = sceneSamples.imagePtArray.h;

	PtArray.Element = new OrientedPoint[PtArray.n];

	pVNClass = classArray.Element + iMostProbableClass;

	pModel = models[pVNClass->iMetaModel];

	float *d = new float[pModel->featureArray.n];

	int i, j;
	float *a;

	RVLMULMXTVECT(pVNClass->M.Element, qOpt, pVNClass->M.h, pVNClass->M.w, d, i, j, a);

	pModel->Project(d, ROpt, t, camera, sceneSamples.imagePtArray, PtArray);
	//pModel->Project(d, R, t, camera, imagePtArray, PtArray);
	//pModel->Project(refModel.d, R, t, camera, imagePtArray, PtArray);

	SURFEL::DeleteSceneSamples(sceneSamples);

	

	//FILE *fp = fopen("P.txt", "w");

	//PrintMatrix<float>(fp, PtArray.Element, PtArray.h, PtArray.w);

	//fclose(fp);

	// Visualization

	if (bVisualization)
	{
		unsigned char SelectionColor[] = { 0, 255, 0 };

		pSurfels->NodeColors(SelectionColor);

		Visualizer visualizer;

		visualizer.Create();

		//pObjects->InitDisplay(&visualizer, pMesh, SelectionColor);
		//pObjects->Display();

		alignment.InitDisplay(&visualizer, pMesh, SelectionColor);
		alignment.Display();


		unsigned char color[] = { 0, 128, 255 };

		Array<Point> PtArray_;

		MESH::CreatePointArrayFromOrientedPointArray(PtArray, PtArray_, 2.0f);

		visualizer.DisplayPointSet<float, Point>(PtArray_, color, 6.0f);
		
		
		Gripper grip;
		
		
		vtkSmartPointer<vtkTransform> TIn = vtkSmartPointer<vtkTransform>::New();

		double TGS[16], RGS[9], RGM[9];
		
		int n1, n2, n3, n4, n5;
		n1 = pVNClass->VNGraspNormals_.Element[0].n1;
		n2 = pVNClass->VNGraspNormals_.Element[0].n2;
		n3 = pVNClass->VNGraspNormals_.Element[0].n3;
		n4 = pVNClass->VNGraspNormals_.Element[0].n4;
		n5 = pVNClass->VNGraspNormals_.Element[0].n5;
		
	//	//mug:
	//	n1 = 11;
	//	n2 = 92;
	//	n3 = 15;
	//	n4 = 18;
	//	n5 = 0;

	//	//hammer:
	//	n1 = 99;
	//	n2 = 85;
	//	n3 = 89;
	//	n4 = 26;
	//	n5 = 92;
	//	
	//	//banana:
	///*	n1 = 57;
	//	n2 = 73;
	//	n3 = 55;
	//	n4 = 54;
	//	n5 = 0;*/
	//	n1 = 55;
	//	n2 = 72;
	//	n3 = 50;
	//	n4 = 57;
	//	n5 = 0;

	//	//bottle:
	//	//parallel planes for gripping
	//	n1 = 87;
	//	n2 = 72;
	//	//adjacent planes
	//	n3 = 89;
	//	n4 = 91;
	//	//top plane
	//	n5 = 61;

	//	//car:
	//	n1 = 55;
	//	n2 = 22;
	//	n3 = 59;
	//	n4 = 62;
	//	n5 = 44;

	//	n1 = 55;
	//	n2 = 22;
	//	n3 = 58;
	//	n4 = 61;
	//	n5 = 0;

		float width = abs(d[n1] + d[n2]);
		float tx[3], tz[3], ty[3], ts1[3], ts2[3];
		float a;
		ts1[0] = pModel->featureArray.Element[n3].N[0];
		ts1[1] = pModel->featureArray.Element[n3].N[1];
		ts1[2] = pModel->featureArray.Element[n3].N[2];

		ts2[0] = pModel->featureArray.Element[n4].N[0];
		ts2[1] = pModel->featureArray.Element[n4].N[1];
		ts2[2] = pModel->featureArray.Element[n4].N[2];


		RGM[1] = ty[0] = pModel->featureArray.Element[n1].N[0];
		RGM[4] = ty[1] = pModel->featureArray.Element[n1].N[1];
		RGM[7] = ty[2] = pModel->featureArray.Element[n1].N[2];

		RGM[2] = tz[0] = pModel->featureArray.Element[n5].N[0];
		RGM[5] = tz[1] = pModel->featureArray.Element[n5].N[1];
		RGM[8] = tz[2] = pModel->featureArray.Element[n5].N[2];

		RVLCROSSPRODUCT3(ty, tz, tx);

		RGM[0] = tx[0];
		RGM[3] = tx[1];
		RGM[6] = tx[2];

		RVLMXMUL3X3(ROpt, RGM, RGS);


		a = (d[n3] * RVLDOTPRODUCT3(ts2, ty) - d[n4] * RVLDOTPRODUCT3(ts1, ty)) / (RVLDOTPRODUCT3(ts1, tx) * RVLDOTPRODUCT3(ts2, ty) - RVLDOTPRODUCT3(ts2, tx)*RVLDOTPRODUCT3(ts1, ty));

		float b = d[n1] - 0.5 * width;

		float t[3] = { a*RGS[0] + b * RGS[1] + d[n5] * RGS[2], a*RGS[3] + b * RGS[4] + d[n5] * RGS[5], a*RGS[6] + b * RGS[7] + d[n5] * RGS[8] };
		

		RVLHTRANSFMX(RGS, t, TGS);

		TIn->SetMatrix(TGS);


		
		//float translate[3] = { d[11] + (width / 2), 0, 0 };
		grip.CreateGripper(width, TIn, t, visualizer.renderer);

		//visualizer.window->AddRenderer(grip.renderer);
		//visualizer.window->AddRenderer(visualizer.renderer);
		visualizer.Run();

		delete[] PtArray_.Element;
	}

#ifdef RVLVN_TIME_MESUREMENT
	//Timer END
	QueryPerformanceCounter((LARGE_INTEGER *)&CNTRSegmentationEND);
	timeInterpret += (CNTRSegmentationEND.QuadPart - CNTRSegmentationSTART.QuadPart) * 1000.0 / frequency.QuadPart;
	printf("time: %.5f", timeInterpret);
#endif
	delete[] d;
	delete[] q;
	delete[] qOpt;
	//delete[] q0;
	//delete[] correspondences.Element;
}

bool VNClassifier::Classify(
	Mesh *pMesh,
	SURFEL::ObjectGraph *pObjects,
	Array<int> *piConvexObjects,
	Array<int> iSegmentArray,
	RECOG::VN_::FitData *fitData,
	QList<RECOG::VN_::Hypothesis> *pHypothesisList,
	float *d,
	uchar *bd,
	float *f,
	float *R,
	CRVLMem *pMem)
{
	RVLQLIST_INIT(pHypothesisList);	// (ARP3D.TR3.9: Algorithm 2, line 5)

	float maxcsNGnd = 0.99f;

	// Consider ground plane or not.

	bool bGnd_ = (bGnd && bHorizontalBottomAssumption);

	// Make union of surfels and vertices of the segments in iSegmentArray (ARP3D.TR3.9: Algorithm 2, line 4).

	pObjects->UnionOfObjects(iSegmentArray, iSurfelArray, iVertexArray, bSurfelInArray, bVertexInArray);

	int support = pSurfels->SupportSize(iSurfelArray);

	// maxConvexityError <- maximum size of the intersection between the support set of the hypothesis and a convex segment. 

	int maxConvexityError = 0;

	int i;

	if (convexityErrorCoeff > 1e-10)
	{
	if (piConvexObjects)
	{
		for (i = 0; i < iSurfelArray.n; i++)
			bSurfelInArray[iSurfelArray.Element[i]] = true;

		int intersection, convexityError;
		SURFEL::Object *pObject;
		QLIST::Index *piSurfel;

		for (i = 0; i < piConvexObjects->n; i++)
		{
			pObject = pObjects->objectArray.Element + piConvexObjects->Element[i];

			intersection = 0;

			piSurfel = pObject->surfelList.pFirst;

			while (piSurfel)
			{
				if (bSurfelInArray[piSurfel->Idx])
					intersection += pSurfels->NodeArray.Element[piSurfel->Idx].size;

				piSurfel = piSurfel->pNext;
			}

			convexityError = (intersection > pObject->size / 2 ? pObject->size - intersection : intersection);

			if (convexityError > maxConvexityError)
				maxConvexityError = convexityError;
		}

		for (i = 0; i < iSurfelArray.n; i++)
			bSurfelInArray[iSurfelArray.Element[i]] = false;
	}
	}

	// Copy vertices from pSurfels->vertexArray to convexHullVertices.

	float *P, *P_;

	for (i = 0; i < iVertexArray.n; i++)
	{
		P = pSurfels->vertexArray.Element[iVertexArray.Element[i]]->P;

		P_ = convexHullVertices.Element + 3 * i;

		RVLCOPY3VECTOR(P, P_);
	}

	convexHullVertices.h = iVertexArray.n;

	// Only for debugging purpose!

	//FILE *fpDebug = fopen("C:\\RVL\\Debug\\P.txt", "w");

	//PrintMatrix<float>(fpDebug, convexHullVertices.Element, convexHullVertices.h, 3);

	//fclose(fpDebug);

	//

	CRVLTimer timer;

	// Compute the approximate minimum bounding sphere of the input point set (ARP3D.TR3.9: Algorithm 2, a condition in line 1).

	timer.Start();

	float boundingSphereCenter[3];
	float boundingSphereRadius;

	BoundingSphere<float>(convexHullVertices, boundingSphereCenter, boundingSphereRadius);

	timer.Stop();

	double tBoundingSphere = timer.GetTime();

	// If the bounding sphere is to big, return false.

	if (boundingSphereRadius > maxBoundingSphereRadius)
		return false;

	// Create VN model and count outliers (ARP3D.TR3.9: Algorithm 2, lines 9, 10).

	VN *pVN = new VN;

	pVN->CreateEmpty();

	float I[9];

	RVLUNITMX3(I);

	float t[3];

	RVLNULL3VECTOR(t);

	Array2D<float> NCHArray;
	NCHArray.Element = NULL;
	float *dCH = NULL;

	Array2D<float> NConcavityArray;
	NConcavityArray.h = 0;
	float *dConcavity;

	//

	float *dS = NULL;

	Array<float> alphaArray;

	alphaArray.n = 0;

	Array<float> betaArray;

	betaArray.n = 0;

	bool bMetamodelConvex = true;

	int iFace;
	MESH::Face *pFace;
	int nOutliers;

	if (bDepthImage)
	{
		// Concavities.

		for (i = 0; i < iSegmentArray.n; i++)
			if (pObjects->objectArray.Element[iSegmentArray.Element[i]].flags & RVLPCSEGMENT_OBJECT_FLAG_CONCAVE)
				break;

		if (i < iSegmentArray.n)	// if there are concave segments in iSegmentArray
		{
			bMetamodelConvex = false;

			// Create convex hull of the union of concave segments (ARP3D.TR3.9: Algorithm 1, line 6)

			Array<int> iConcaveSegmentArray;

			iConcaveSegmentArray.Element = new int[iSegmentArray.n];

			iConcaveSegmentArray.n = 0;

			for (i = 0; i < iSegmentArray.n; i++)
				if (pObjects->objectArray.Element[iSegmentArray.Element[i]].flags & RVLPCSEGMENT_OBJECT_FLAG_CONCAVE)
					iConcaveSegmentArray.Element[iConcaveSegmentArray.n++] = iSegmentArray.Element[i];

			pObjects->UnionOfObjects(iConcaveSegmentArray, iConcavitySurfelArray, iConcavityVertexArray, bSurfelInArray, bVertexInArray);

			delete[] iConcaveSegmentArray.Element;

			for (i = 0; i < iConcavityVertexArray.n; i++)
			{
				P = pSurfels->vertexArray.Element[iConcavityVertexArray.Element[i]]->P;

				P_ = concavityVertices.Element + 3 * i;

				RVLCOPY3VECTOR(P, P_);
			}

			concavityVertices.h = iConcavityVertexArray.n;

			if (convexHull.ConvexHull(concavityVertices, &memConvexHull))
			{
				// Create complement of convexHull (ARP3D.TR3.9: Algorithm 1, line 7)

				convexHull.iVisibleFaces.n = 0;

				for (i = 0; i < convexHull.faces.n; i++)
				{
					pFace = convexHull.faces.Element[i];

					RVLNEGVECT3(pFace->N, pFace->N);

					pFace->d = -pFace->d;

					pFace->flags ^= RVLMESH_FACE_FLAG_VISIBLE;

					if (pFace->flags & RVLMESH_FACE_FLAG_VISIBLE)
						convexHull.iVisibleFaces.Element[convexHull.iVisibleFaces.n++] = i;
				}

				convexHull.VisibleSurface(NConcavityArray, dConcavity);

				pVN->AddModelCluster(0, RVLVN_CLUSTER_TYPE_CONCAVE, I, t, 0.5f, alphaArray, betaArray, NConcavityArray, pMem, 0.0f, NULL);
			}
		}		// if there are concave segments in iSegmentArray

		// Compute the convex hull of the input point set (ARP3D.TR3.9: Algorithm 1, line 6).

		timer.Start();

		int nSamples;

		if (convexHull.ConvexHull(convexHullVertices, &memConvexHull))
		{
			timer.Stop();

			double tConvexHull = timer.GetTime();

			// Combine the convex and concave metamodel components into a polyhedron (ARP3D.TR3.9: Algorithm 1, line 9).

			timer.Start();

			convexHull.VisibleSurface(NCHArray, dCH);

			if (NConcavityArray.h > 0)
			{
				pVN->AddModelCluster(1, RVLVN_CLUSTER_TYPE_CONVEX, I, t, 0.5f, alphaArray, betaArray, NCHArray, pMem, 0.0f, NULL);

				pVN->AddOperation(2, 1, 0, 1, &memConvexHull);

				pVN->SetOutput(2);

				dS = new float[NCHArray.h + NConcavityArray.h];

				memcpy(dS, dConcavity, NConcavityArray.h * sizeof(float));
				memcpy(dS + NConcavityArray.h, dCH, NCHArray.h * sizeof(float));
			}
			else
			{
				pVN->AddModelCluster(0, RVLVN_CLUSTER_TYPE_CONVEX, I, t, 0.5f, alphaArray, betaArray, NCHArray, pMem, 0.0f, NULL);

				pVN->SetOutput(0);

				dS = dCH;
			}

			pVN->Create(pMem);

			timer.Stop();

			double tCreateVNFromConvexMesh = timer.GetTime();

			// Check whether the segment union iSegmentArray covers its entire convex hull - count outliers (ARP3D.TR3.9: Algorithm 2, line 10). 

			float chamferDist;

			pVN->SceneFittingScore(pMesh, pSurfels, camera, iSurfelArray, iVertexArray, dS, fitParams.maxe, maxz,
				fitParams.cosSurfaceRayAngleThr, nSamples, chamferDist, nOutliers, visualizationData.bVisualizeSceneFitting);
		}
		else
			nOutliers = nSamples = 1;		

		if (NConcavityArray.h > 0)
		{
			delete[] NConcavityArray.Element;
			delete[] dConcavity;
		}

		// If the percentage of outliers is > maxOutlierPerc, then no hypothesis is generated (ARP3D.TR3.9: Algorithm 2, line 13).

		if (100 * nOutliers / nSamples > maxOutlierPerc)
		{
			delete pVN;
			RVL_DELETE_ARRAY(dCH);
			if (NConcavityArray.h > 0)
				RVL_DELETE_ARRAY(dS);

			return false;
		}

		// All convex hull faces with normal similar to the opposite ground plane normal are considered as visible.

		if (bGnd_)
		{
			for (i = 0; i < convexHull.faces.n; i++)
			{
				pFace = convexHull.faces.Element[i];

				if (pFace->flags & RVLMESH_FACE_FLAG_VISIBLE)
					continue;

				if (RVLDOTPRODUCT3(pFace->N, NGnd) > -maxcsNGnd)
					continue;

				convexHull.iVisibleFaces.Element[convexHull.iVisibleFaces.n++] = i;
			}
		}

		// Convex hull faces defined by occlusion vertices only are considered as invisible.

		Array<int> iVisibleFaces;

		iVisibleFaces.Element = new int[convexHull.iVisibleFaces.n];

		iVisibleFaces.n = 0;

		MeshEdgePtr *pEdgePtr, *pEdgePtr0;
		int iCHVertex;
		SURFEL::Vertex *pVertex_;

		for (i = 0; i < convexHull.iVisibleFaces.n; i++)
		{
			iFace = convexHull.iVisibleFaces.Element[i];

			pFace = convexHull.faces.Element[iFace];

			pEdgePtr0 = pEdgePtr = pFace->pFirstEdgePtr;

			do
			{
				iCHVertex = RVLPCSEGMENT_GRAPH_GET_NODE(pEdgePtr);

				pVertex_ = pSurfels->vertexArray.Element[iVertexArray.Element[iCHVertex]];

				if (!(pVertex_->type & RVLSURFELVERTEX_TYPE_OCCLUSION))
					break;

				pEdgePtr = pEdgePtr->pNext;

				if (pEdgePtr == NULL)
					pEdgePtr = convexHull.NodeArray.Element[iCHVertex].EdgeList.pFirst;

				pEdgePtr = RVLPCSEGMENT_GRAPH_GET_OPPOSITE_EDGE_PTR(pEdgePtr);
			} while (pEdgePtr != pEdgePtr0);

			if (!(pVertex_->type & RVLSURFELVERTEX_TYPE_OCCLUSION))
				iVisibleFaces.Element[iVisibleFaces.n++] = iFace;
		}

		if (iVisibleFaces.n == 0)
		{
			delete pVN;
			delete[] iVisibleFaces.Element;
			RVL_DELETE_ARRAY(dCH);
			if (NConcavityArray.h > 0)
				RVL_DELETE_ARRAY(dS);

			return false;
		}

		delete[] convexHull.iVisibleFaces.Element;

		convexHull.iVisibleFaces = iVisibleFaces;
	}	// if (bDepthImage)
	else
	{
		nOutliers = 0;

		// Compute the convex hull of the input point set.

		timer.Start();

		if (convexHull.ConvexHull(convexHullVertices, &memConvexHull))
		{
			timer.Stop();

			double tConvexHull = timer.GetTime();

			// Check whether the segment union iSegmentArray covers its entire convex hull.

			timer.Start();

			NCHArray.h = convexHull.iVisibleFaces.n = convexHull.faces.n;

			NCHArray.w = 3;

			NCHArray.Element = new float[NCHArray.w * NCHArray.h];

			dCH = new float[NCHArray.h];

			float *N;

			for (i = 0; i < convexHull.faces.n; i++)
			{
				convexHull.iVisibleFaces.Element[i] = i;

				convexHull.faces.Element[i]->flags |= RVLMESH_FACE_FLAG_VISIBLE;

				N = NCHArray.Element + 3 * i;

				pFace = convexHull.faces.Element[i];

				RVLCOPY3VECTOR(pFace->N, N);

				dCH[i] = pFace->d;
			}

			pVN->AddModelCluster(0, RVLVN_CLUSTER_TYPE_CONVEX, I, t, 0.5f, alphaArray, betaArray, NCHArray, pMem, 0.0f, NULL);

			pVN->SetOutput(0);

			dS = dCH;

			pVN->Create(pMem);

			timer.Stop();

			double tCreateVNFromConvexMesh = timer.GetTime();
		}
	}	// if (!bDepthImage)

	RVL_DELETE_ARRAY(NCHArray.Element);

	QList<VN> *pVNList = &VNList;

	RVLQLIST_ADD_ENTRY(pVNList, pVN);

	// Only for debugging purpose!

	//if (convexHull.iVisibleFaces.n >= 19)
	//{
	//	pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[6]]->Area = 1.0;
	//	pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[7]]->Area = 1.0;
	//	pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[10]]->Area = 1.0;
	//	pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[11]]->Area = 1.0;
	//	pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[13]]->Area = 1.0;
	//}

	///

	Array<int> iSegmentArray_;
	RVLMEM_ALLOC_STRUCT_ARRAY(pMem, int, iSegmentArray.n, iSegmentArray_.Element);
	memcpy(iSegmentArray_.Element, iSegmentArray.Element, iSegmentArray.n * sizeof(int));
	iSegmentArray_.n = iSegmentArray.n;
	Array<int> iSurfelArray_;
	RVLMEM_ALLOC_STRUCT_ARRAY(pMem, int, iSurfelArray.n, iSurfelArray_.Element);
	memcpy(iSurfelArray_.Element, iSurfelArray.Element, iSurfelArray.n * sizeof(int));
	iSurfelArray_.n = iSurfelArray.n;
	Array<int> iVertexArray_;
	RVLMEM_ALLOC_STRUCT_ARRAY(pMem, int, iVertexArray.n, iVertexArray_.Element);
	memcpy(iVertexArray_.Element, iVertexArray.Element, iVertexArray.n * sizeof(int));
	iVertexArray_.n = iVertexArray.n;
	float *dS_;
	RVLMEM_ALLOC_STRUCT_ARRAY(pMem, float, pVN->featureArray.n, dS_);
	memcpy(dS_, dS, pVN->featureArray.n * sizeof(float));

	RVL_DELETE_ARRAY(dCH);

	if (NConcavityArray.h > 0)
		RVL_DELETE_ARRAY(dS);
	
	// Align the faces of the convex hull with the convex template for different initial conditions.

	int *faceCorrespondence = new int[convexHull.iVisibleFaces.n];

	bool *faceMask = new bool[sampledUnitSphere.h];

	Array<int> iFaceArray;

	iFaceArray.Element = new int[convexHull.iVisibleFaces.n];

	iFaceArray.n = 0;

	Array<int> iModelNormalArray;
	iModelNormalArray.Element = NULL;

	float U[9];
	float *U_;
	Array<int> *piModelNormalArray;

	if (bAlignWithConvexTemplate)
	{
		for (i = 0; i < sampledUnitSphere.h; i++)
			faceMask[i] = true;

	InitNormalAlignment(convexHull.faces, convexHull.iVisibleFaces, U);

		U_ = U;

		piModelNormalArray = NULL;
	}
	else
	{
		iModelNormalArray.n = 6;

		iModelNormalArray.Element = new int[iModelNormalArray.n];

		memset(faceMask, 0, sampledUnitSphere.h * sizeof(bool));

		for (i = 0; i < iModelNormalArray.n; i++)
		{
			iModelNormalArray.Element[i] = i * 11;

			faceMask[i * 11] = true;
		}

		U_ = NULL;

		piModelNormalArray = &iModelNormalArray;
	}

	float RXZ[9];
	float *X = RXZ;
	float *Y = RXZ + 3;
	float *Z = RXZ + 6;

	timer.Start();

	float dPan = 2.0f * PI / (float)(CTIDescriptorMapingData.nPan);
	float dTilt = 0.25f * PI / (float)(CTIDescriptorMapingData.nTilt);
	float dRoll = 0.5f * PI / (float)(CTIDescriptorMapingData.nRoll);

	int nRollTotal = 4 * CTIDescriptorMapingData.nRoll;
	int nZ = CTIDescriptorMapingData.nPan * (CTIDescriptorMapingData.nTilt - 1) + 1;

	int rollBlockSize = CTIDescriptorMapingData.nRoll * sampledUnitSphere.h;

	Point **pVertex = new Point *[rollBlockSize];

	int nDescriptorsPerBlock = 6 * nRollTotal;

	int totalBlockDescriptorSize = nDescriptorsPerBlock * sampledUnitSphere.h;

	memset(f, 0, nZ * totalBlockDescriptorSize * sizeof(float));

	float Z_[3];

	Z_[2] = 1.0f;

	float X_[3];

	X_[2] = 0.0f;

	float fmaxiTilt = (float)(CTIDescriptorMapingData.nTilt - 1);

	int iZ = 0;

	// Only for debugging purpose!

	//float *ZArchive = new float[3 * nZ];

	///

	float maxScore = 0.0f;

	float *V3Tmp;
	int iPan, iTilt, iRoll, iRollBlockStart, nPan, iZRef, iRollRef;
	float pan, tilt, roll, maxTilt, cPan, sPan, cTilt, sTilt, cRoll, sRoll, d0, d_, dmax, abscPan, abssPan, x, y, fTmp;
	float RRoll[9], R0[9], RSM[9];
	int j;
	float *f_;
	float score;

	for (iTilt = 0; iTilt < CTIDescriptorMapingData.nTilt; iTilt++)
	{
		nPan = (iTilt > 0 ? CTIDescriptorMapingData.nPan : 1);

		for (iPan = 0; iPan < nPan; iPan++)
		{
			pan = (float)iPan * dPan;

			cPan = cos(pan);
			sPan = sin(pan);

			abscPan = RVLABS(cPan);
			abssPan = RVLABS(sPan);

			fTmp = (abscPan >= abssPan ? abscPan : abssPan);

			Z_[0] = cPan / fTmp;
			Z_[1] = sPan / fTmp;

			fTmp = sqrt(RVLDOTPRODUCT3(Z_, Z_));

			maxTilt = acos(1.0f / fTmp);

			tilt = maxTilt / fmaxiTilt * (float)iTilt;

			cTilt = cos(tilt);
			sTilt = sin(tilt);

			Z[0] = cPan * sTilt;
			Z[1] = sPan * sTilt;
			Z[2] = cTilt;

			// Only for debugging purpose!

			//float *ZArchive_ = ZArchive + 3 * iZ;

			//RVLCOPY3VECTOR(Z, ZArchive_);

			///

			X_[0] = -Z[1];
			X_[1] = Z[0];

			fTmp = sqrt(RVLDOTPRODUCT3(X_, X_));

			if (fTmp >= 1e-10)
			{
				RVLSCALE3VECTOR2(X_, fTmp, X);
			}
			else
				RVLSET3VECTOR(X, 1.0f, 0.0f, 0.0f);

			RVLCROSSPRODUCT3(Z, X, Y);

			iRollBlockStart = iZ * totalBlockDescriptorSize;

			//if (iZ == 8)
			//	int debug = 0;

			for (iRoll = 0; iRoll < CTIDescriptorMapingData.nRoll; iRoll++)
			{
				roll = (float)iRoll * dRoll;

				cRoll = cos(roll);
				sRoll = sin(roll);

				RVLROTZ(cRoll, sRoll, RRoll);

				RVLMXMUL3X3(RRoll, RXZ, RSM);

				RVLCOPYMX3X3T(RSM, R0);

				score = FitCTI(&convexHull, U_, R0, faceMask, 2, R + 9 * (iZ * CTIDescriptorMapingData.nRoll + iRoll), faceCorrespondence, piModelNormalArray);

#ifdef RVLVN_CLASSIFY_CTI_VERSION1
				f_ = f + iRollBlockStart + iRoll * sampledUnitSphere.h;

				for (j = 0; j < convexHull.iVisibleFaces.n; j++)
					f_[faceCorrespondence[j]] += convexHull.faces.Element[convexHull.iVisibleFaces.Element[j]]->Area;
#endif
				if (score > maxScore)
				{
					maxScore = score;

					iZRef = iZ;
					iRollRef = iRoll;
			}
			}

			iZ++;
		}	// for (iPan = 0; iPan < nPan; iPan++)
	}	// for (iTilt = 0; iTilt < CTIDescriptorMapingData.nTilt; iTilt++)

	RVL_DELETE_ARRAY(iModelNormalArray.Element);

	// Only for debugging purpose!

	//FILE *fpZArchive = fopen("C:\\RVL\\Debug\\Z.txt", "w");

	//PrintMatrix<float>(fpZArchive, ZArchive, nZ, 3);

	//fclose(fpZArchive);

	//delete[] ZArchive;

	///

	int k, iConvexTempleteElement, i_, j_, k_;
	float *R__, *RXZ_, *uS;
	float R_[9], RMS_[9], uM[3];
	float *N;
	float N_[3], absProj[3], RM0S[9];
	int imax;
	int *iD_, *intR;

	//maxScore = 0.0f;

	//for (iZ = 0; iZ < nZ; iZ++)
	//{
	//	for (iRoll = 0; iRoll < CTIDescriptorMapingData.nRoll; iRoll++)
	//	{
	//		R__ = R + 9 * (iZ * CTIDescriptorMapingData.nRoll + iRoll);

	//		score = 0.0f;

	//		for (i = 0; i < convexHull.iVisibleFaces.n; i++)
	//		{
	//			iFace = convexHull.iVisibleFaces.Element[i];

	//			pFace = convexHull.faces.Element[iFace];

	//			if (pFace->Area > 0.0f)
	//			{
	//				N = pFace->N;

	//				RVLMULMX3X3TVECT(R__, N, N_);

	//				absProj[0] = RVLABS(N_[0]);
	//				absProj[1] = RVLABS(N_[1]);
	//				absProj[2] = RVLABS(N_[2]);

	//				imax = (absProj[0] >= absProj[1] ? 0 : 1);
	//				if (absProj[2] > absProj[imax])
	//					imax = 2;

	//				score += (pFace->Area * absProj[imax]);
	//			}
	//		}

	//		if (score > maxScore)
	//		{
	//			maxScore = score;

	//			iZRef = iZ;
	//			iRollRef = iRoll;
	//		}
	//	}
	//}

	R__ = R + 9 * (iZRef * CTIDescriptorMapingData.nRoll + iRollRef);

	//roll = (float)iRollRef * dRoll;

	//cRoll = cos(roll);
	//sRoll = sin(roll);

	//RVLROTZ(cRoll, sRoll, RRoll);

	//RXZ_ = CTIDescriptorMapingData.R;

	//RVLMXMUL3X3T2(RRoll, RXZ_, R_);

	//RVLMXMUL3X3(R__, R_, RMS_);

	RVLCOPYMX3X3(R__, RMS_);

	if (bAlignWithConvexTemplate)
	{
	iFaceArray.n = 0;

	for (j = 0; j < convexHull.iVisibleFaces.n; j++)
	{
		iFace = convexHull.iVisibleFaces.Element[j];

		pFace = convexHull.faces.Element[iFace];

		uS = pFace->N;

		RVLVN_GET_CLOSEST_CONVEX_TEMPLATE_ELEMENT(convexTemplateLUT, uS, RMS_, uM, i_, j_, k_, iConvexTempleteElement);

		if (faceMask[iConvexTempleteElement])
			iFaceArray.Element[iFaceArray.n++] = iFace;
	}

		memset(faceMask, 0, sampledUnitSphere.h * sizeof(bool));

		for (k = 0; k < 6; k++)
			faceMask[k * 11] = true;

	InitNormalAlignment(convexHull.faces, iFaceArray, U);

	FitCTI(&convexHull, U, RMS_, faceMask, 1, RM0S, faceCorrespondence);
	}
	else
	{
		RVLCOPYMX3X3(RMS_, RM0S);
	}

	Visualizer *pVisualizer = NULL;

	float *dH = new float[sampledUnitSphere.h];
	float *dH_ = new float[sampledUnitSphere.h];
	uchar *bdH = new uchar[sampledUnitSphere.h];
	uchar *bdH_ = new uchar[sampledUnitSphere.h];

	ComputeDescriptor(&convexHull, RM0S, dH, bdH, &iVertexArray);

	if (visualizationData.bVisualizeCTIs)
	{
		pVisualizer = new Visualizer;

		pVisualizer->Create();

		pVisualizer->SetMesh(pMesh);

		int *mask = new int[sampledUnitSphere.h];

		for (i = 0; i < sampledUnitSphere.h; i++)
			mask[i] = (int)(bdH[i]);

		convexClustering.DisplayCTI(pVisualizer, sampledUnitSphere.Element, sampledUnitSphere.h, dH, RM0S, t, mask);

		//float tCTIc_CTI[3];

		//vtkSmartPointer<vtkPolyData> polyData = GenerateCTIPrimitivePolydata_RW(sampledUnitSphere.Element, d__,
		//	sampledUnitSphere.h, false, mask, tCTIc_CTI);

		delete[] mask;

		//pVisualizer->AddMesh(polyData);

		pVisualizer->Run();

		pVisualizer->renderer->RemoveAllViewProps();
	}

	// Map the remaining descriptors to the reference descriptors.	

	float *f__;

	for (j = 0; j < nZ; j++)
	{
		f__ = f + j * totalBlockDescriptorSize;

		for (i = 0; i < totalBlockDescriptorSize; i++)
			f__[i] = f__[CTIDescriptorMapingData.iD[i]];
	}

	timer.Stop();

	double tFitCTI = timer.GetTime();

	// Classification.

	timer.Start();

	///

	float totalArea = 0.0f;

	for (i = 0; i < convexHull.iVisibleFaces.n; i++)
		totalArea += convexHull.faces.Element[convexHull.iVisibleFaces.Element[i]]->Area;

	int nDescriptors = nZ * 6 * nRollTotal;

	float minE = -1.0f;

	bool bFirstClass = true;

	int iMostProbableClass = -1;

	Array<int> iAligned;

	iAligned.Element = new int[nDescriptors];

	Array<float *> secondAxisArray;

	secondAxisArray.Element = new float *[sampledUnitSphere.h];

	float *thirdAxis = new float[3 * sampledUnitSphere.h];

	int maxqSize = 0;

	int iClass;
	RECOG::ClassData *pClass;

	for (iClass = 0; iClass < classArray.n; iClass++)
	{
		pClass = classArray.Element + iClass;

		if (pClass->M.h > maxqSize)
			maxqSize = pClass->M.h;
	}

	float *q_ = new float[maxqSize];

	int id, idRef, iD, idEnd, iRoll_, nHypotheses, iHypothesis, nqs, iBlock, iR_;
	float correlation, maxCorrelation_, ELS, ELSV, E, w, bH, eqs, Eqs, ESP, z, zmin, wRelevant, maxf, f2;
	float *d__, *a, *a_, *AH, *q, *qs, *refAxis;
	uchar *bd__;
	float ROpt[9], RMS[9], W[3], PS[3], maxCorrelation[6], M3x3Tmp[9];
	bool bReal[3];
	int idx[3], iR[6];
	float C[9];
	VN_::Hypothesis *pHypothesis;
	//double eig[3];

	for (iClass = 1; iClass < classArray.n; iClass++)	// (ARP3D.TR3.9: Algorithm 2, line 16)
	{
		if (debug1 >= 0)
			if (iClass != debug1)
				continue;

		pClass = classArray.Element + iClass;

		if (pClass->bConvex && !bMetamodelConvex)
						continue;

		//nHypotheses = 0;

		//for (j = 0; j < nDescriptors; j++, f_ += sampledUnitSphere.h)
		//{
		//	correlation = 0.0f;

		//	for (k = 0; k < pClass->faces.n; k++)
		//		correlation += f_[pClass->faces.Element[k]];

		//	if (correlation > 0.5f * maxCorrelation)
		//		nHypotheses++;
		//}

		memset(faceMask, 0, sampledUnitSphere.h * sizeof(bool));

		for (k = 0; k < pClass->faces.n; k++)
			faceMask[pClass->faces.Element[k]] = true;

#ifdef RVLVN_CLASSIFY_HYPOTHESIS_CTI_LOG
		FILE *fpDesc = fopen("desc.dat", "w");
#endif

		for (iHypothesis = 0; iHypothesis < alignment.CTIDescMap.nCTI; iHypothesis++)
		{
			iD_ = alignment.CTIDescMap.iD + sampledUnitSphere.h * iHypothesis;

			for (i = 0; i < sampledUnitSphere.h; i++)
			{
				dH_[i] = dH[iD_[i]];
				bdH_[i] = bdH[iD_[i]];
			}

			d__ = dH_;
			bd__ = bdH_;

			intR = alignment.CTIDescMap.R + 9 * iHypothesis;

			RVLMXMUL3X3(RM0S, intR, RMS);

#ifdef RVLVN_CLASSIFY_HYPOTHESIS_CTI_LOG
			SaveDescriptor(fpDesc, dH_, bdH_, RMS, iHypothesis, 0);
#endif

			if (FitLS(d__, bd__, pClass, fitData + iClass, ELS))	// (ARP3D.TR3.9:	Algorithm 2, line 17; 
																	//					fitData[iClass].q_ <- Eq. (3); 
																	//					ELS <- the first termi in Eq. (4))
					{
				// Compute ELSV using Eq. (4) in ARP3D.TR3.9.

				ELSV = (bDepthImage ? ELS + invisibleComponentPenal * (float)(sampledUnitSphere.h / 2 - fitData[iClass].iV.n) : ELS);

				// Compute Eqs using Eq. (5) in ARP3D.TR3.9.

							nqs = pClass->M.h - 4;

							q = fitData[iClass].q_;

							qs = q + 4;

							Eqs = 0.0f;

							for (i = 0; i < pClass->nHull; i++)
							{
								AH = pClass->A + nqs * i;

								bH = q[3] * pClass->b[i];

								RVLDOTPRODUCT(AH, qs, nqs, eqs, j);

								eqs -= bH;

								if (eqs > Eqs)
									Eqs = eqs;
							}

							Eqs *= Eqs;

				// Compute E using Eq. (6) in ARP3D.TR3.9.

				E = (Eqs > ELSV ? Eqs : ELSV);

				// CTIMesh <- mesh representing CTI corresponding to the latent vector fitData[iClass].q_

						RVLMULMXTVECT(pClass->M.Element, fitData[iClass].q_, pClass->M.h, pClass->M.w, dH_, i, j, a);

#ifdef RVLVN_CLASSIFY_HYPOTHESIS_CTI_LOG
						//SaveDescriptor(fpDesc, dH_, bdH_, RMS, iHypothesis, 0);
#endif

						//printf("tCTIMesh=%lf\n", tCTIMesh);

						CRVLTimer timer2;

						timer2.Start();

						PSGM_::CreateCTIMesh(sampledUnitSphere, dH_, &CTIMesh, CTIMeshVertexMemSize, &memCTIMesh);

						timer2.Stop();

						double tCTIMesh = timer2.GetTime();

				// Compute the distance from the dominant planar surface and add it to the cost E.

				if (bGnd_)
						{
							zmin = dGnd;

							for (i = 0; i < CTIMesh.iValidVertices.n; i++)
							{
								P = CTIMesh.NodeArray.Element[CTIMesh.iValidVertices.Element[i]].P;

								RVLMULMX3X3VECT(RMS, P, PS);

								z = RVLDOTPRODUCT3(NGnd, PS);

								if (z < zmin)
									zmin = z;
							}

							zmin -= dGnd;

							ESP = (zmin < 0.010f ? zmin : 0.0f);

							ESP *= ESP;

							//printf("iClass=%d E=%f ESP=%f\n", iClass, E, ESP);

							E += (kZMin2 * ESP);
						}

				// minE <- minimum hypothesis cost
				// ROpt <- optimal orientation
				// q_ <- optimal latent vector 
				// (ARP3D.TR3.9: Algorithm 2, lines 20-23)

						if (bFirstClass || E < minE)
						{
							minE = E;

							iMostProbableClass = iClass;

							RVLCOPYMX3X3(RMS, ROpt);

							memcpy(q_, fitData[iClass].q_, pClass->M.h * sizeof(float));

							bFirstClass = false;
						}

						//FILE *fpDebug = fopen("C:\\RVL\\Debug\\P.txt", "w");

						//PrintMatrix<float>(fpDebug, pHypothesis->modelVertices.Element, pHypothesis->modelVertices.h, 3);

						//fclose(fpDebug);
					}
			else    // If LS equations can't be solved 
				minE = invisibleComponentPenal * (float)(sampledUnitSphere.h);

			if (bPrimitiveLayer)
			{

			}
		}	// for each hypothesis

#ifdef RVLVN_CLASSIFY_HYPOTHESIS_CTI_LOG
		fclose(fpDesc);
#endif
	}	// for every class

	// Create hypothesis (ARP3D.TR3.9: Algorithm 2, line 22)

	if (!bPrimitiveLayer)
	{
		RVLMEM_ALLOC_STRUCT(pMem, VN_::Hypothesis, pHypothesis);

		pHypothesis->pVN = pVN;
		pHypothesis->iSegmentArray = iSegmentArray_;
		pHypothesis->iSurfelArray = iSurfelArray_;
		pHypothesis->iVertexArray = iVertexArray_;
		pHypothesis->support = support;
		pHypothesis->d = dS_;
		pClass = classArray.Element + iMostProbableClass;
		if (bFirstClass)
		{
			pHypothesis->iClass = -1;
			pHypothesis->q = NULL;
			RVLUNITMX3(pHypothesis->R);
		}
		else
		{
			pHypothesis->iClass = iMostProbableClass;
			RVLMEM_ALLOC_STRUCT_ARRAY(pMem, float, pClass->M.h, pHypothesis->q);
			memcpy(pHypothesis->q, q_, pClass->M.h * sizeof(float));
			RVLCOPYMX3X3(ROpt, pHypothesis->R);
		}
		pHypothesis->shapeCost = minE;
		pHypothesis->convexityError = (float)maxConvexityError / (float)support;
		pHypothesis->nOutliers = nOutliers;
		pHypothesis->cost = pHypothesis->shapeCost + convexityErrorCoeff * pHypothesis->convexityError + outlierCost * (float)nOutliers;
		//RVLMEM_ALLOC_STRUCT_ARRAY(pMem, float, 3 * CTIMesh.iValidVertices.n, pHypothesis->modelVertices.Element);
		//pHypothesis->modelVertices.w = 3;
		//pHypothesis->modelVertices.h = CTIMesh.iValidVertices.n;
		//P_ = pHypothesis->modelVertices.Element;
		//for (i = 0; i < CTIMesh.iValidVertices.n; i++, P_ += 3)
		//{
		//	P = CTIMesh.NodeArray.Element[CTIMesh.iValidVertices.Element[i]].P;
		//	RVLCOPY3VECTOR(P, P_);
		//}

		RVLQLIST_ADD_ENTRY(pHypothesisList, pHypothesis);
	}

	timer.Stop();

	double tClassification = timer.GetTime();

	//printf("tClassification=%lf\n", tClassification);

	if (visualizationData.bVisualizeCTIs)
		delete pVisualizer;

	// Free memory.

	delete[] faceCorrespondence;
	delete[] faceMask;
	delete[] pVertex;
	delete[] dH;
	delete[] dH_;
	delete[] bdH;
	delete[] bdH_;
	delete[] iAligned.Element;
	delete[] secondAxisArray.Element;
	delete[] thirdAxis;
	delete[] iFaceArray.Element;
	delete[] q_;

	//

	return true;
}

bool VNClassifier::Classify2(
	Mesh *pMesh,
	SURFEL::ObjectGraph *pObjects,
	Array<int> *piConvexObjects,
	Array<int> iSegmentArray,
	RECOG::VN_::FitData *fitData,
	QList<RECOG::VN_::Hypothesis> *pHypothesisList,
	float *d,
	uchar *bd,
	float *f,
	float *R,
	CRVLMem *pMem)
{
	float maxcsNGnd = 0.99f;

	// Make union of surfels and vertices of the segments in iSegmentArray.

	pObjects->UnionOfObjects(iSegmentArray, iSurfelArray, iVertexArray, bSurfelInArray, bVertexInArray);

	int support = pSurfels->SupportSize(iSurfelArray);

	// maxConvexityError <- maximum size of the intersection between the support set of the hypothesis and a convex segment. 

	int maxConvexityError = 0;

	int i;

	if (piConvexObjects)
	{
		for (i = 0; i < iSurfelArray.n; i++)
			bSurfelInArray[iSurfelArray.Element[i]] = true;

		int intersection, convexityError;
		SURFEL::Object *pObject;
		QLIST::Index *piSurfel;

		for (i = 0; i < piConvexObjects->n; i++)
		{
			pObject = pObjects->objectArray.Element + piConvexObjects->Element[i];

			intersection = 0;

			piSurfel = pObject->surfelList.pFirst;

			while (piSurfel)
			{
				if (bSurfelInArray[piSurfel->Idx])
					intersection += pSurfels->NodeArray.Element[piSurfel->Idx].size;

				piSurfel = piSurfel->pNext;
			}

			convexityError = (intersection > pObject->size / 2 ? pObject->size - intersection : intersection);

			if (convexityError > maxConvexityError)
				maxConvexityError = convexityError;
		}

		for (i = 0; i < iSurfelArray.n; i++)
			bSurfelInArray[iSurfelArray.Element[i]] = false;
	}

	// Copy vertices from pSurfels->vertexArray to convexHullVertices.

	float *P, *P_;

	for (i = 0; i < iVertexArray.n; i++)
	{
		P = pSurfels->vertexArray.Element[iVertexArray.Element[i]]->P;

		P_ = convexHullVertices.Element + 3 * i;

		RVLCOPY3VECTOR(P, P_);
	}

	convexHullVertices.h = iVertexArray.n;

	// Only for debugging purpose!

	//FILE *fpDebug = fopen("C:\\RVL\\Debug\\P.txt", "w");

	//PrintMatrix<float>(fpDebug, convexHullVertices.Element, convexHullVertices.h, 3);

	//fclose(fpDebug);

	//

	CRVLTimer timer;

	// Compute the approximate minimum bounding sphere of the input point set.

	timer.Start();

	float boundingSphereCenter[3];
	float boundingSphereRadius;

	BoundingSphere<float>(convexHullVertices, boundingSphereCenter, boundingSphereRadius);

	timer.Stop();

	double tBoundingSphere = timer.GetTime();

	// If the bounding sphere is to big, return false.

	if (boundingSphereRadius > maxBoundingSphereRadius)
		return false;

	// Create VN model.

	VN *pVN = new VN;

	pVN->CreateEmpty();

	float I[9];

	RVLUNITMX3(I);

	float t[3];

	RVLNULL3VECTOR(t);

	Array2D<float> NCHArray;
	NCHArray.Element = NULL;
	float *dCH = NULL;
	
	Array2D<float> NConcavityArray;
	NConcavityArray.h = 0;
	float *dConcavity;

	//

	float *dS = NULL;

	Array<float> alphaArray;

	alphaArray.n = 0;

	Array<float> betaArray;

	betaArray.n = 0;

	int iFace;
	MESH::Face *pFace;

	if (bDepthImage)
	{
		// Concavities.

		for (i = 0; i < iSegmentArray.n; i++)
			if (pObjects->objectArray.Element[iSegmentArray.Element[i]].flags & RVLPCSEGMENT_OBJECT_FLAG_CONCAVE)
				break;

		if (i < iSegmentArray.n)	// if there are concave segments in iSegmentArray
		{
			Array<int> iConcaveSegmentArray;

			iConcaveSegmentArray.Element = new int[iSegmentArray.n];

			iConcaveSegmentArray.n = 0;

			for (i = 0; i < iSegmentArray.n; i++)
				if (pObjects->objectArray.Element[iSegmentArray.Element[i]].flags & RVLPCSEGMENT_OBJECT_FLAG_CONCAVE)
					iConcaveSegmentArray.Element[iConcaveSegmentArray.n++] = iSegmentArray.Element[i];

			pObjects->UnionOfObjects(iConcaveSegmentArray, iConcavitySurfelArray, iConcavityVertexArray, bSurfelInArray, bVertexInArray);

			delete[] iConcaveSegmentArray.Element;

			for (i = 0; i < iConcavityVertexArray.n; i++)
			{
				P = pSurfels->vertexArray.Element[iConcavityVertexArray.Element[i]]->P;

				P_ = concavityVertices.Element + 3 * i;

				RVLCOPY3VECTOR(P, P_);
			}

			concavityVertices.h = iConcavityVertexArray.n;

			if (convexHull.ConvexHull(concavityVertices, &memConvexHull))
			{
				convexHull.iVisibleFaces.n = 0;

				for (i = 0; i < convexHull.faces.n; i++)
				{
					pFace = convexHull.faces.Element[i];

					RVLNEGVECT3(pFace->N, pFace->N);

					pFace->d = -pFace->d;

					pFace->flags ^= RVLMESH_FACE_FLAG_VISIBLE;

					if (pFace->flags & RVLMESH_FACE_FLAG_VISIBLE)
						convexHull.iVisibleFaces.Element[convexHull.iVisibleFaces.n++] = i;
				}

				convexHull.VisibleSurface(NConcavityArray, dConcavity);

				pVN->AddModelCluster(0, RVLVN_CLUSTER_TYPE_CONCAVE, I, t, 0.5f, alphaArray, betaArray, NConcavityArray, pMem, 0.0f, NULL);
			}
		}

		// Compute the convex hull of the input point set.

		timer.Start();

		int nSamples, nOutliers;

		if (convexHull.ConvexHull(convexHullVertices, &memConvexHull))
		{
			timer.Stop();

			double tConvexHull = timer.GetTime();

			// Check whether the segment union iSegmentArray covers its entire convex hull.

			timer.Start();

			//model.CreateFromConvexMesh(&convexHull, dCH, &memConvexHull);

			convexHull.VisibleSurface(NCHArray, dCH);

			if (NConcavityArray.h > 0)
			{
				pVN->AddModelCluster(1, RVLVN_CLUSTER_TYPE_CONVEX, I, t, 0.5f, alphaArray, betaArray, NCHArray, pMem, 0.0f, NULL);

				pVN->AddOperation(2, 1, 0, 1, &memConvexHull);

				pVN->SetOutput(2);

				dS = new float[NCHArray.h + NConcavityArray.h];

				memcpy(dS, dConcavity, NConcavityArray.h * sizeof(float));
				memcpy(dS + NConcavityArray.h, dCH, NCHArray.h * sizeof(float));
			}
			else
			{
				pVN->AddModelCluster(0, RVLVN_CLUSTER_TYPE_CONVEX, I, t, 0.5f, alphaArray, betaArray, NCHArray, pMem, 0.0f, NULL);

				pVN->SetOutput(0);

				dS = dCH;
			}

			pVN->Create(pMem);

			timer.Stop();

			double tCreateVNFromConvexMesh = timer.GetTime();

			float chamferDist;

			pVN->SceneFittingScore(pMesh, pSurfels, camera, iSurfelArray, iVertexArray, dS, fitParams.maxe, maxz,
				fitParams.cosSurfaceRayAngleThr, nSamples, chamferDist, nOutliers, visualizationData.bVisualizeSceneFitting);
		}
		else
			nOutliers = nSamples = 1;		

		if (NConcavityArray.h > 0)
		{
			delete[] NConcavityArray.Element;
			delete[] dConcavity;
		}

		if (100 * nOutliers / nSamples > maxOutlierPerc)
			return false;

		// All convex hull faces with normal similar to the opposite ground plane normal are considered as visible.

		if (bGnd)
		{
			for (i = 0; i < convexHull.faces.n; i++)
			{
				pFace = convexHull.faces.Element[i];

				if (pFace->flags & RVLMESH_FACE_FLAG_VISIBLE)
					continue;

				if (RVLDOTPRODUCT3(pFace->N, NGnd) > -maxcsNGnd)
					continue;

				convexHull.iVisibleFaces.Element[convexHull.iVisibleFaces.n++] = i;
			}
		}

		// Convex hull faces defined by occlusion vertices only are considered as invisible.

		Array<int> iVisibleFaces;

		iVisibleFaces.Element = new int[convexHull.iVisibleFaces.n];

		iVisibleFaces.n = 0;

		MeshEdgePtr *pEdgePtr, *pEdgePtr0;
		int iCHVertex;
		SURFEL::Vertex *pVertex_;

		for (i = 0; i < convexHull.iVisibleFaces.n; i++)
		{
			iFace = convexHull.iVisibleFaces.Element[i];

			pFace = convexHull.faces.Element[iFace];

			pEdgePtr0 = pEdgePtr = pFace->pFirstEdgePtr;

			do
			{
				iCHVertex = RVLPCSEGMENT_GRAPH_GET_NODE(pEdgePtr);

				pVertex_ = pSurfels->vertexArray.Element[iVertexArray.Element[iCHVertex]];

				if (!(pVertex_->type & RVLSURFELVERTEX_TYPE_OCCLUSION))
					break;

				pEdgePtr = pEdgePtr->pNext;

				if (pEdgePtr == NULL)
					pEdgePtr = convexHull.NodeArray.Element[iCHVertex].EdgeList.pFirst;

				pEdgePtr = RVLPCSEGMENT_GRAPH_GET_OPPOSITE_EDGE_PTR(pEdgePtr);
			} while (pEdgePtr != pEdgePtr0);

			if (!(pVertex_->type & RVLSURFELVERTEX_TYPE_OCCLUSION))
				iVisibleFaces.Element[iVisibleFaces.n++] = iFace;
		}

		if (iVisibleFaces.n == 0)
		{
			delete pVN;
			delete[] iVisibleFaces.Element;
			RVL_DELETE_ARRAY(dCH);
			if (NConcavityArray.h > 0)
				RVL_DELETE_ARRAY(dS);

			return false;
		}

		delete[] convexHull.iVisibleFaces.Element;

		convexHull.iVisibleFaces = iVisibleFaces;
	}	// if (bDepthImage)
	else
	{
		// Compute the convex hull of the input point set.

		timer.Start();

		if (convexHull.ConvexHull(convexHullVertices, &memConvexHull))
		{
			timer.Stop();

			double tConvexHull = timer.GetTime();

			// Check whether the segment union iSegmentArray covers its entire convex hull.

			timer.Start();

			NCHArray.h = convexHull.iVisibleFaces.n = convexHull.faces.n;

			NCHArray.w = 3;

			NCHArray.Element = new float[NCHArray.w * NCHArray.h];

			dCH = new float[NCHArray.h];

			float *N;

			for (i = 0; i < convexHull.faces.n; i++)
			{
				convexHull.iVisibleFaces.Element[i] = i;

				convexHull.faces.Element[i]->flags |= RVLMESH_FACE_FLAG_VISIBLE;

				N = NCHArray.Element + 3 * i;

				pFace = convexHull.faces.Element[i];

				RVLCOPY3VECTOR(pFace->N, N);

				dCH[i] = pFace->d;
			}

			pVN->AddModelCluster(0, RVLVN_CLUSTER_TYPE_CONVEX, I, t, 0.5f, alphaArray, betaArray, NCHArray, pMem, 0.0f, NULL);

			pVN->SetOutput(0);

			dS = dCH;

			pVN->Create(pMem);

			timer.Stop();

			double tCreateVNFromConvexMesh = timer.GetTime();
		}
	}

	RVL_DELETE_ARRAY(NCHArray.Element);

	QList<VN> *pVNList = &VNList;

	RVLQLIST_ADD_ENTRY(pVNList, pVN);

	// Only for debugging purpose!

	//if (convexHull.iVisibleFaces.n >= 19)
	//{
	//	pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[6]]->Area = 1.0;
	//	pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[7]]->Area = 1.0;
	//	pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[10]]->Area = 1.0;
	//	pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[11]]->Area = 1.0;
	//	pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[13]]->Area = 1.0;
	//}

	///

	Array<int> iSegmentArray_;
	RVLMEM_ALLOC_STRUCT_ARRAY(pMem, int, iSegmentArray.n, iSegmentArray_.Element);
	memcpy(iSegmentArray_.Element, iSegmentArray.Element, iSegmentArray.n * sizeof(int));
	iSegmentArray_.n = iSegmentArray.n;
	Array<int> iSurfelArray_;
	RVLMEM_ALLOC_STRUCT_ARRAY(pMem, int, iSurfelArray.n, iSurfelArray_.Element);
	memcpy(iSurfelArray_.Element, iSurfelArray.Element, iSurfelArray.n * sizeof(int));
	iSurfelArray_.n = iSurfelArray.n;
	Array<int> iVertexArray_;
	RVLMEM_ALLOC_STRUCT_ARRAY(pMem, int, iVertexArray.n, iVertexArray_.Element);
	memcpy(iVertexArray_.Element, iVertexArray.Element, iVertexArray.n * sizeof(int));
	iVertexArray_.n = iVertexArray.n;
	float *dS_;
	RVLMEM_ALLOC_STRUCT_ARRAY(pMem, float, pVN->featureArray.n, dS_);
	memcpy(dS_, dS, pVN->featureArray.n * sizeof(float));

	RVL_DELETE_ARRAY(dCH);

	if (NConcavityArray.h > 0)
		RVL_DELETE_ARRAY(dS);

	int *faceCorrespondence = new int[convexHull.iVisibleFaces.n];

	bool *faceMask = new bool[sampledUnitSphere.h];

	for (i = 0; i < sampledUnitSphere.h; i++)
		faceMask[i] = true;

	float U[9];

	InitNormalAlignment(convexHull.faces, convexHull.iVisibleFaces, U);

	float RXZ[9];
	float *X = RXZ;
	float *Y = RXZ + 3;
	float *Z = RXZ + 6;

	timer.Start();

	float dPan = 2.0f * PI / (float)(CTIDescriptorMapingData.nPan);
	float dTilt = 0.25f * PI / (float)(CTIDescriptorMapingData.nTilt);
	float dRoll = 0.5f * PI / (float)(CTIDescriptorMapingData.nRoll);

	int nRollTotal = 4 * CTIDescriptorMapingData.nRoll;
	int nZ = CTIDescriptorMapingData.nPan * (CTIDescriptorMapingData.nTilt - 1) + 1;

	int rollBlockSize = CTIDescriptorMapingData.nRoll * sampledUnitSphere.h;

	Point **pVertex = new Point *[rollBlockSize];

	int nDescriptorsPerBlock = 6 * nRollTotal;

	int totalBlockDescriptorSize = nDescriptorsPerBlock * sampledUnitSphere.h;

	memset(f, 0, nZ * totalBlockDescriptorSize * sizeof(float));

	float Z_[3];

	Z_[2] = 1.0f;

	float X_[3];

	X_[2] = 0.0f;

	float fmaxiTilt = (float)(CTIDescriptorMapingData.nTilt - 1);

	int iZ = 0;

	// Only for debugging purpose!

	//float *ZArchive = new float[3 * nZ];

	///

	float *V3Tmp;
	int iPan, iTilt, iRoll, iRollBlockStart, nPan;
	float pan, tilt, roll, maxTilt, cPan, sPan, cTilt, sTilt, cRoll, sRoll, d0, d_, dmax, abscPan, abssPan, x, y, fTmp;
	float RRoll[9], R0[9], RSM[9];
	int j;
	float *f_;

	for (iTilt = 0; iTilt < CTIDescriptorMapingData.nTilt; iTilt++)
	{
		nPan = (iTilt > 0 ? CTIDescriptorMapingData.nPan : 1);

		for (iPan = 0; iPan < nPan; iPan++)
		{
			pan = (float)iPan * dPan;

			cPan = cos(pan);
			sPan = sin(pan);

			abscPan = RVLABS(cPan);
			abssPan = RVLABS(sPan);

			fTmp = (abscPan >= abssPan ? abscPan : abssPan);

			Z_[0] = cPan / fTmp;
			Z_[1] = sPan / fTmp;

			fTmp = sqrt(RVLDOTPRODUCT3(Z_, Z_));

			maxTilt = acos(1.0f / fTmp);

			tilt = maxTilt / fmaxiTilt * (float)iTilt;

			cTilt = cos(tilt);
			sTilt = sin(tilt);

			Z[0] = cPan * sTilt;
			Z[1] = sPan * sTilt;
			Z[2] = cTilt;

			// Only for debugging purpose!

			//float *ZArchive_ = ZArchive + 3 * iZ;

			//RVLCOPY3VECTOR(Z, ZArchive_);

			///

			X_[0] = -Z[1];
			X_[1] = Z[0];

			fTmp = sqrt(RVLDOTPRODUCT3(X_, X_));

			if (fTmp >= 1e-10)
			{
				RVLSCALE3VECTOR2(X_, fTmp, X);
			}
			else
				RVLSET3VECTOR(X, 1.0f, 0.0f, 0.0f);
			
			RVLCROSSPRODUCT3(Z, X, Y);

			iRollBlockStart = iZ * totalBlockDescriptorSize;

			//if (iZ == 11)
			//	int debug = 0;

			for (iRoll = 0; iRoll < CTIDescriptorMapingData.nRoll; iRoll++)
			{
				roll = (float)iRoll * dRoll;

				cRoll = cos(roll);
				sRoll = sin(roll);

				RVLROTZ(cRoll, sRoll, RRoll);

				RVLMXMUL3X3(RRoll, RXZ, RSM);

				RVLCOPYMX3X3T(RSM, R0);

				FitCTI(&convexHull, U, R0, faceMask, 2, R + 9 * (iZ * CTIDescriptorMapingData.nRoll + iRoll), faceCorrespondence);

#ifdef RVLVN_CLASSIFY_CTI_VERSION1
				f_ = f + iRollBlockStart + iRoll * sampledUnitSphere.h;

				for (j = 0; j < convexHull.iVisibleFaces.n; j++)
					f_[faceCorrespondence[j]] += convexHull.faces.Element[convexHull.iVisibleFaces.Element[j]]->Area;
#endif
			}

#ifdef RVLVN_CLASSIFY_CTI_VERSION0
			ComputeDescriptors(pConvexHull, R + 9 * iZ * CTIDescriptorMapingData.nRoll, d + iRollBlockStart, bd + iRollBlockStart, pVertex);
#endif

			iZ++;
		}	// for (iPan = 0; iPan < nPan; iPan++)
	}	// for (iTilt = 0; iTilt < CTIDescriptorMapingData.nTilt; iTilt++)

	// Only for debugging purpose!
	
	//FILE *fpZArchive = fopen("C:\\RVL\\Debug\\Z.txt", "w");

	//PrintMatrix<float>(fpZArchive, ZArchive, nZ, 3);

	//fclose(fpZArchive);

	//delete[] ZArchive;

	///

	// Map the remaining descriptors to the reference descriptors.	

#ifdef RVLVN_CLASSIFY_CTI_VERSION0
	float *d__;
	
	for (j = 0; j < nZ; j++)
	{
		d__ = d + j * totalBlockDescriptorSize;

		for (i = 0; i < totalBlockDescriptorSize; i++)
			d__[i] = d__[CTIDescriptorMapingData.iD[i]];
	}

	timer.Stop();

	double tFitCTI = timer.GetTime();
#endif

#ifdef RVLVN_CLASSIFY_CTI_VERSION1
	float *f__;

	for (j = 0; j < nZ; j++)
	{
		f__ = f + j * totalBlockDescriptorSize;

		for (i = 0; i < totalBlockDescriptorSize; i++)
			f__[i] = f__[CTIDescriptorMapingData.iD[i]];
	}

	timer.Stop();

	double tFitCTI = timer.GetTime();

	// Classification.

	timer.Start();

	Visualizer *pVisualizer = NULL;

	if (visualizationData.bVisualizeCTIs)
	{
		pVisualizer = new Visualizer;

		pVisualizer->Create();
	}

	///

	float totalArea = 0.0f;

	for (i = 0; i < convexHull.iVisibleFaces.n; i++)
		totalArea += convexHull.faces.Element[convexHull.iVisibleFaces.Element[i]]->Area;

	int nDescriptors = nZ * 6 * nRollTotal;

	float minE = -1.0f;

	bool bFirstClass = true;

	int iMostProbableClass = -1;

	RVLQLIST_INIT(pHypothesisList);

	float *dH = new float[sampledUnitSphere.h];

	Array<int> iAligned;

	iAligned.Element = new int[nDescriptors];

	Array<float *> secondAxisArray;

	secondAxisArray.Element = new float *[sampledUnitSphere.h];

	float *thirdAxis = new float[3 * sampledUnitSphere.h];

	Array<int> iFaceArray;

	iFaceArray.Element = new int[convexHull.iVisibleFaces.n];

	int iClass, k, id, idRef, iD, idEnd, iRoll_, nHypotheses, iHypothesis, nqs, iBlock, iR_, i_, j_, k_, iConvexTempleteElement;
	RECOG::ClassData *pClass;
	float correlation, maxCorrelation_, E, w, bH, eqs, Eqs, ESP, z, zmin, wRelevant, maxf, f2;
	float *R__, *d__, *RXZ_, *a, *a_, *AH, *q, *qs, *refAxis, *uS;
	uchar *bd__;
	float ROpt[9], R_[9], RMS[9], W[3], PS[3], maxCorrelation[6], M3x3Tmp[9], RMS_[9], uM[3];
	bool bReal[3];
	int idx[3], iR[6];
	float C[9];
	VN_::Hypothesis *pHypothesis;
	//double eig[3];

	for (iClass = 0; iClass < classArray.n; iClass++)
	{
		if (debug1 >= 0)
			if (iClass != debug1)
				continue;

		pClass = classArray.Element + iClass;

		memset(maxCorrelation, 0, 6 * sizeof(float));
		memset(iR, 0xff, 6 * sizeof(int));

		f_ = f;

		iAligned.n = 0;

		for (j = 0; j < nDescriptors; j++, f_ += sampledUnitSphere.h)
		{
			//if (j == 11 * 6 * 16)
			//	int debug = 0;

			//if (j == 10 * 6 * 16 + 2)
			//	int debug = 0;

			iBlock = (j / nRollTotal) % 6;

			if (pClass->bAxisAlignment)
			{
				correlation = 0.0f;

				for (k = 0; k < pClass->faces.n; k++)
					correlation += f_[pClass->faces.Element[k]];
			}
			else
			{
				RVLNULLMX3X3(C);

				wRelevant = 0.0f;

				for (k = 0; k < pClass->faces.n; k++)
				{
					i = pClass->faces.Element[k];

					w = f_[i] / totalArea;

					if (w < 1e-4)
						continue;

					wRelevant += w;

					a = sampledUnitSphere.Element + 3 * i;

					RVLSCALE3VECTOR(a, w, W);

					RVLVECTCOV3(W, M3x3Tmp);

					RVLSUMMX3X3(C, M3x3Tmp, C);
				}

				if (C[0] + C[4] + C[8] < 1e-8f)
					continue;

				if (wRelevant < alignmentSupportThr)
					continue;

				iAligned.Element[iAligned.n++] = j;

				//EigCov3<double>(C, eig, bReal);

				//if (!(bReal[0] && bReal[1] && bReal[2]))
				//	continue;

				//RVLSORT3ASCEND(eig, idx, fTmp);

				//correlation = eig[idx[1]] * eig[idx[2]];

				// Only for debugging purpose!

				//RVLCOMPLETESIMMX3(C);

				//FILE *fp = fopen("C:\\RVL\\Debug\\C.txt", "w");

				//PrintMatrix<float>(fp, C, 3, 3);

				//fclose(fp);

				//

				correlation = -C[2] * C[2] + C[0] * C[8] + C[4] * C[8] - C[1] * C[1] + C[0] * C[4] - C[5] * C[5];
			}

			if (correlation > maxCorrelation[iBlock])
			{
				maxCorrelation[iBlock] = correlation;

				iR[iBlock] = j;
			}
		}	// for (j = 0; j < nDescriptors; j++, f_ += sampledUnitSphere.h)

		for (iBlock = 0; iBlock < 6; iBlock++)
			if (iR[iBlock] >= 0)
				break;

		if (iBlock >= 6)
		{
			maxCorrelation_ = 0.0f;

			for (j = 0; j < iAligned.n; j++)
			{
				f_ = f + iAligned.Element[j] * sampledUnitSphere.h;

				maxf = 0.0;

				for (k = 0; k < pClass->faces.n; k++)
				{
					i = pClass->faces.Element[k];

					f2 = f_[i];

					if (f2 > maxf)
					{
						maxf = f2;

						refAxis = sampledUnitSphere.Element + 3 * i;

						w = f2 / totalArea;
					}
				}

				RVLSCALE3VECTOR(refAxis, w, W);

				RVLVECTCOV3(W, C);

				secondAxisArray.n = 0;

				for (k = 0; k < pClass->faces.n; k++)
				{
					i = pClass->faces.Element[k];

					a = sampledUnitSphere.Element + 3 * i;

					fTmp = RVLDOTPRODUCT3(refAxis, a);

					if (RVLABS(fTmp) < 0.5f)
					{
						secondAxisArray.Element[secondAxisArray.n] = a;

						a_ = thirdAxis + 3 * secondAxisArray.n;

						RVLCROSSPRODUCT3(refAxis, a, a_);

						secondAxisArray.n++;
					}
				}

				for (i = 0; i < sampledUnitSphere.h; i++)
				{
					w = f_[i] / totalArea;

					if (w < 1e-4)
						continue;

					a = sampledUnitSphere.Element + 3 * i;

					for (k = 0; k < secondAxisArray.n; k++)
					{
						a_ = secondAxisArray.Element[k];

						fTmp = RVLDOTPRODUCT3(a, a_);

						if (fTmp > 0.866f)
						{
							a_ = thirdAxis + 3 * k;

							fTmp = RVLDOTPRODUCT3(a, a_);

							if (RVLABS(fTmp) < 0.1)
							{
								RVLSCALE3VECTOR(a, w, W);

								RVLVECTCOV3(W, M3x3Tmp);

								RVLSUMMX3X3(C, M3x3Tmp, C);
							}
						}
					}
				}	// for (i = 0; i < sampledUnitSphere.h; i++)

				correlation = -C[2] * C[2] + C[0] * C[8] + C[4] * C[8] - C[1] * C[1] + C[0] * C[4] - C[5] * C[5];

				if (correlation > maxCorrelation_)
				{
					maxCorrelation_ = correlation;

					iR_ = iAligned.Element[j];

					iBlock = (iR_ / nRollTotal) % 6;

					iR[iBlock] = iR_;

					maxCorrelation[iBlock] = maxCorrelation_;
				}
			}	// for (j = 0; j < iAligned.n; j++)
		}	// if (iR < 0)

		//nHypotheses = 0;

		//for (j = 0; j < nDescriptors; j++, f_ += sampledUnitSphere.h)
		//{
		//	correlation = 0.0f;

		//	for (k = 0; k < pClass->faces.n; k++)
		//		correlation += f_[pClass->faces.Element[k]];

		//	if (correlation > 0.5f * maxCorrelation)
		//		nHypotheses++;
		//}

		memset(faceMask, 0, sampledUnitSphere.h * sizeof(bool));

		for (k = 0; k < pClass->faces.n; k++)
			faceMask[pClass->faces.Element[k]] = true;

		for (iHypothesis = 0; iHypothesis < pClass->iHypothesisBlocks.h; iHypothesis++)
		{
			maxCorrelation_ = 0.0f;

			iR_ = -1;

			for (j = 0; j < pClass->iHypothesisBlocks.w; j++)
			{
				iBlock = pClass->iHypothesisBlocks.Element[pClass->iHypothesisBlocks.w * iHypothesis + j];

				if (maxCorrelation[iBlock] > maxCorrelation_)
				{
					iR_ = iR[iBlock];

					maxCorrelation_ = maxCorrelation[iBlock];
				}
			}

			if (iR_ >= 0)	// if there are valid alignment proposals
			{
				iZ = iR_ / nDescriptorsPerBlock;

				iRoll = iR_ % CTIDescriptorMapingData.nRoll;

				R__ = R + 9 * (iZ * CTIDescriptorMapingData.nRoll + iRoll);

				//idRef = iZ * totalBlockDescriptorSize + iRoll * sampledUnitSphere.h;

				//d__ = d + idRef;

				//bd__ = bd + idRef;

				//ComputeDescriptor(pConvexHull, R__, d__, bd__);

				//iD = iZ * totalBlockDescriptorSize;

				//d__ = d + iD;

				//bd__ = bd + iD;

				//id = iR * sampledUnitSphere.h - iD;

				//idEnd = id + sampledUnitSphere.h;

				//for (i = id; i < idEnd; i++)
				//{
				//	d__[i] = d__[CTIDescriptorMapingData.iD[i]];
				//	bd__[i] = bd__[CTIDescriptorMapingData.iD[i]];
				//}

				//d__ += id;
				//bd__ += id;

				//nHypotheses = (iClass < 2 ? 2 : 1);

				//for (iHypothesis = 0; iHypothesis < nHypotheses; iHypothesis++)
				{
					iRoll_ = iR_ % nRollTotal;

					roll = (float)(iRoll_ - iRoll) * dRoll;

					cRoll = cos(roll);
					sRoll = sin(roll);

					RVLROTZ(cRoll, sRoll, RRoll);

					//iBlock = (iHypothesis == 0 ? (iR_ / nRollTotal) % 6 : (iBlock + 3) % 6);

					iBlock = (iR_ / nRollTotal) % 6;

					RXZ_ = CTIDescriptorMapingData.R + 9 * iBlock;

					RVLMXMUL3X3T2(RRoll, RXZ_, R_);

					RVLMXMUL3X3(R__, R_, RMS_);

					iFaceArray.n = 0;

					for (j = 0; j < convexHull.iVisibleFaces.n; j++)
					{
						iFace = convexHull.iVisibleFaces.Element[j];

						pFace = convexHull.faces.Element[iFace];

						uS = pFace->N;

						RVLVN_GET_CLOSEST_CONVEX_TEMPLATE_ELEMENT(convexTemplateLUT, uS, RMS_, uM, i_, j_, k_, iConvexTempleteElement);

						if (faceMask[iConvexTempleteElement])
							iFaceArray.Element[iFaceArray.n++] = iFace;
					}

					InitNormalAlignment(convexHull.faces, iFaceArray, U);

					FitCTI(&convexHull, U, RMS_, faceMask, 1, RMS, faceCorrespondence);

					id = iR_ * sampledUnitSphere.h;

					d__ = d + id;
					bd__ = bd + id;

					ComputeDescriptor(&convexHull, RMS, d__, bd__, &iVertexArray);

					// Only for debugging purpose!

					//float *d2 = new float[sampledUnitSphere.h];
					//uchar *bd2 = new uchar[sampledUnitSphere.h];

					//ComputeDescriptor(pConvexHull, RMS, d2, bd2);

					//delete[] d2;
					//delete[] bd2;

					//

					// Only for debugging purpose!

					//FILE *fp = fopen("C:\\RVL\\Debug\\scenedescriptors.txt", "w");

					//SaveDescriptor(fp, d__, bd__, R__, 0, 0);

					//fclose(fp);

					//

					if (visualizationData.bVisualizeCTIs)
					{
						pVisualizer->SetMesh(pMesh);

						int *mask = new int[sampledUnitSphere.h];

						for (i = 0; i < sampledUnitSphere.h; i++)
							mask[i] = (int)(bd__[i]);

						convexClustering.DisplayCTI(pVisualizer, sampledUnitSphere.Element, sampledUnitSphere.h, d__, RMS, t, mask);

						//float tCTIc_CTI[3];

						//vtkSmartPointer<vtkPolyData> polyData = GenerateCTIPrimitivePolydata_RW(sampledUnitSphere.Element, d__,
						//	sampledUnitSphere.h, false, mask, tCTIc_CTI);

						delete[] mask;
					
						//pVisualizer->AddMesh(polyData);

						pVisualizer->Run();

						pVisualizer->renderer->RemoveAllViewProps();
					}

					if (FitLS(d__, bd__, pClass, fitData + iClass, E))
					{
						if (bDepthImage)
							E += (0.0001 * (sampledUnitSphere.h / 2 - fitData[iClass].iV.n));

						if (iClass < 2)
						{
							nqs = pClass->M.h - 4;

							q = fitData[iClass].q_;

							qs = q + 4;

							Eqs = 0.0f;

							for (i = 0; i < pClass->nHull; i++)
							{
								AH = pClass->A + nqs * i;

								bH = q[3] * pClass->b[i];

								RVLDOTPRODUCT(AH, qs, nqs, eqs, j);

								eqs -= bH;

								if (eqs > Eqs)
									Eqs = eqs;
							}

							Eqs *= Eqs;

							if (Eqs > E)
								E = Eqs;
						}

						RVLMULMXTVECT(pClass->M.Element, fitData[iClass].q_, pClass->M.h, pClass->M.w, dH, i, j, a);

						//

						//printf("tCTIMesh=%lf\n", tCTIMesh);

						CRVLTimer timer2;

						timer2.Start();

						PSGM_::CreateCTIMesh(sampledUnitSphere, dH, &CTIMesh, CTIMeshVertexMemSize, &memCTIMesh);

						timer2.Stop();

						double tCTIMesh = timer2.GetTime();

						if (bGnd)
						{
							zmin = dGnd;

							for (i = 0; i < CTIMesh.iValidVertices.n; i++)
							{
								P = CTIMesh.NodeArray.Element[CTIMesh.iValidVertices.Element[i]].P;

								RVLMULMX3X3VECT(RMS, P, PS);

								z = RVLDOTPRODUCT3(NGnd, PS);

								if (z < zmin)
									zmin = z;
							}

							zmin -= dGnd;

							ESP = (zmin < 0.010f ? zmin : 0.0f);

							ESP *= ESP;

							//printf("iClass=%d E=%f ESP=%f\n", iClass, E, ESP);

							E += (kZMin2 * ESP);
						}						
					
						RVLMEM_ALLOC_STRUCT(pMem, VN_::Hypothesis, pHypothesis);

						pHypothesis->pVN = pVN;
						pHypothesis->iSegmentArray = iSegmentArray_;
						pHypothesis->iSurfelArray = iSurfelArray_;
						pHypothesis->iVertexArray = iVertexArray_;
						pHypothesis->support = support;
						pHypothesis->iClass = iClass;
						pHypothesis->d = dS_;
						RVLMEM_ALLOC_STRUCT_ARRAY(pMem, float, pClass->M.h, pHypothesis->q);
						memcpy(pHypothesis->q, fitData[iClass].q_, pClass->M.h * sizeof(float));
						RVLCOPYMX3X3(RMS, pHypothesis->R);
						pHypothesis->cost = E + convexityErrorCoeff * (float)maxConvexityError / (float)support;
						RVLMEM_ALLOC_STRUCT_ARRAY(pMem, float, 3 * CTIMesh.iValidVertices.n, pHypothesis->modelVertices.Element);
						pHypothesis->modelVertices.w = 3;
						pHypothesis->modelVertices.h = CTIMesh.iValidVertices.n;
						P_ = pHypothesis->modelVertices.Element;
						for (i = 0; i < CTIMesh.iValidVertices.n; i++, P_ += 3)
						{
							P = CTIMesh.NodeArray.Element[CTIMesh.iValidVertices.Element[i]].P;
							RVLCOPY3VECTOR(P, P_);
						}

						RVLQLIST_ADD_ENTRY(pHypothesisList, pHypothesis);

						//FILE *fpDebug = fopen("C:\\RVL\\Debug\\P.txt", "w");

						//PrintMatrix<float>(fpDebug, pHypothesis->modelVertices.Element, pHypothesis->modelVertices.h, 3);

						//fclose(fpDebug);

						//if (bFirstClass || E < minE)
						//{
						//	minE = E;

						//	iMostProbableClass = iClass;

						//	//RVLCOPYMX3X3(R__, ROpt);
						//	RVLCOPYMX3X3(RMS, ROpt);

						//	bFirstClass = false;
						//}
					}
				}	// for each hypothesis
			}	// if there are valid alignment proposals
		}	// for each hypothesis
	}	// for every class

	timer.Stop();

	double tClassification = timer.GetTime();

	//printf("tClassification=%lf\n", tClassification);
#endif	

	if (visualizationData.bVisualizeCTIs)
		delete pVisualizer;

	// Free memory.

	delete[] faceCorrespondence;
	delete[] faceMask;
	delete[] pVertex;
	delete[] dH;
	delete[] iAligned.Element;
	delete[] secondAxisArray.Element;
	delete[] thirdAxis;
	delete[] iFaceArray.Element;

	//

	return true;
}

// Function VNClassifier::InterpretationCost requires VNClassifier::InitInterpretationCostComputation to be called before the first call
// and VNClassifier::DeallocateInterpretationCostComputationMem after the last call.

float VNClassifier::InterpretationCost(bool *X)
{
	memset(nSurfelCellHypotheses, 0, pObjects->surfelCells.n * sizeof(int));

	float cost = 0.0f;

	int i, j, iHypothesis;
	VN_::Hypothesis *pHypothesis;
	SURFEL::Cell *pSurfelCell;
	Array<int> *pObjectCells;

	for (iHypothesis = 0; iHypothesis < hypotheses.n; iHypothesis++)
	{
		if (!X[iHypothesis])
			continue;

		pHypothesis = hypotheses.Element[iHypothesis];

		cost += pHypothesis->cost;

		for (i = 0; i < pHypothesis->iSegmentArray.n; i++)
		{
			pObjectCells = pObjects->objectCells + pHypothesis->iSegmentArray.Element[i];

			for (j = 0; j < pObjectCells->n; j++)
				nSurfelCellHypotheses[pObjectCells->Element[j]]++;
		}
	}

	int iCell;
	int nHypotheses;

	for (iCell = 0; iCell < pObjects->surfelCells.n; iCell++)
	{
		nHypotheses = nSurfelCellHypotheses[iCell];

		if (nHypotheses == 0)
			cost += (unassignedPointCost * (float)(pObjects->surfelCells.Element[iCell].size));
		else if (nHypotheses > 1)
			cost += (multiplePointAssignmentCost * (float)((nHypotheses - 1) * pObjects->surfelCells.Element[iCell].size));
	}

	return cost;
}

void VNClassifier::InitInterpretationCostComputation()
{
	nSurfelCellHypotheses = new int[pObjects->surfelCells.n];
}

void VNClassifier::DeallocateInterpretationCostComputationMem()
{
	RVL_DELETE_ARRAY(nSurfelCellHypotheses);
}

void VNClassifier::SegmentToParts(Mesh *pMesh)
{
	bool bHypothesisLog = true;

	// Segment mesh to surfels.				

	pSurfels->Init(pMesh);

	pSurfelDetector->Init(pMesh, pSurfels, pMem);

	printf("Segmentation to surfels... ");

	double StartTime, ExecTime;

	if (pSurfelDetector->pTimer)
		StartTime = pSurfelDetector->pTimer->GetTime();

	pSurfelDetector->Segment(pMesh, pSurfels);

	if (pSurfelDetector->pTimer)
		ExecTime = pSurfelDetector->pTimer->GetTime() - StartTime;

	printf("completed.\n");
	printf("No. of surfels = %d\n", pSurfels->NodeArray.n);

	if (pSurfelDetector->pTimer)
		printf("Total segmentation time = %lf s\n", ExecTime);

	pSurfels->DetectVertices(pMesh);

	// Cluster surfels into convex and concave surfaces.

	Array<RECOG::SceneCluster> SClusters;

	bool bConcavity, bTorus;

	bConcavity = true;
	bTorus = false;

	int nSCClusters, nSUClusters;

	PSGM_::ConvexAndConcaveClusters(pMesh, pSurfels, pSurfelDetector, &convexClustering, &concaveClustering,
		SClusters, nSCClusters, nSUClusters, 0, NULL, maxnSCClusters, maxnSUClusters, bConcavity, (bTorus ? maxnSTClusters : 0),
		visualizationData.bVisualizeConvexClusters, visualizationData.bVisualizeConcaveClusters, visualizationData.bVisualizeSurfels);

	RECOG::CreateObjectGraphFromSceneClusters(SClusters, pObjects);

	pObjects->GetVertices();

	// Classify all clusters.

	maxBoundingSphereRadius = 10000.0f;
	bGnd = false;
	bDepthImage = false;

	descriptorMem.Clear();

	Init(pMesh);

	Array<int> iSegmentArray;

	iSegmentArray.n = 1;
	iSegmentArray.Element = new int[iSegmentArray.n];

	std::vector<void *>hypotheses;

	//hypotheses.clear();

	QList<RECOG::VN_::Hypothesis> hypothesisList;
	QList<RECOG::VN_::Hypothesis> *pHypothesisList = &hypothesisList;

	int iHypothesis = 0;

	int iObject;
	SURFEL::Object *pObject;
	VN_::Hypothesis *pHypothesis;

	for (iObject = 0; iObject < pObjects->objectArray.n; iObject++)
	{
		pObject = pObjects->objectArray.Element + iObject;

		if (pObject->flags & RVLPCSEGMENT_OBJECT_FLAG_CONCAVE)
			continue;

		iSegmentArray.Element[0] = iObject;

		if (!Classify(pMesh, pObjects, NULL, iSegmentArray, fitData, pHypothesisList, D, bD, f, R, pMem))
			continue;

		// Add hypothesis to the hypothesis list.

		pHypothesis = hypothesisList.pFirst;

		while (pHypothesis)
		{
			hypotheses.push_back(pHypothesis);

			pHypothesis = pHypothesis->pNext;
		}
	}

	// Sort hypotheses.

	Array<SortIndex<float>> sortedHypotheses;

	sortedHypotheses.Element = new SortIndex<float>[hypotheses.size()];

	int i;

	for (i = 0; i < hypotheses.size(); i++)
	{
		sortedHypotheses.Element[i].idx = i;
		sortedHypotheses.Element[i].cost = ((RECOG::VN_::Hypothesis *)(hypotheses[i]))->cost;
	}

	sortedHypotheses.n = hypotheses.size();

	BubbleSort<SortIndex<float>>(sortedHypotheses);

	// Save hypotheses to a file.

	if (bHypothesisLog)
	{
		FILE *fpHypothesisLog = fopen((std::string(resultsFolder) + "\\hypotheses.log").data(), "w");

		for (i = 0; i < sortedHypotheses.n; i++)
		{
			iHypothesis = sortedHypotheses.Element[i].idx;

			pHypothesis = (VN_::Hypothesis *)(hypotheses[iHypothesis]);

			fprintf(fpHypothesisLog, "%03d: Segment %d Class %d E=%f\n", iHypothesis, pHypothesis->iSegmentArray.Element[0], pHypothesis->iClass, pHypothesis->cost);
		}

		if (bHypothesisLog)
			fclose(fpHypothesisLog);
	}

	delete[] iSegmentArray.Element;

	// Visualization
	
	Visualizer visualizer;

	visualizer.Create();

	visualizer.SetMesh(pMesh);

	float minCost = -1.0f;

	RECOG::VN_::Hypothesis *pBestHypothesis = NULL;

	for (iHypothesis = 0; iHypothesis < hypotheses.size(); iHypothesis++)
	{
		pHypothesis = (RECOG::VN_::Hypothesis *)(hypotheses[iHypothesis]);

		if (minCost < 0.0f || pHypothesis->cost < minCost)
		{
			minCost = pHypothesis->cost;

			pBestHypothesis = pHypothesis;
		}
	}

	pHypothesis = pBestHypothesis;

	bool bContinue = true;

	float t[3];

	RVLNULL3VECTOR(t);

	uchar selectionColor[3] = { 0, 0, 255 };

	uchar defaultColor[3] = { 255, 255, 0 };

	RECOG::VN_::Hypothesis *pPrevHypothesis = NULL;

	int iSurfel;
	int j;
	float *a;
	RECOG::ClassData *pClass;
	float *d;

	while (bContinue)
	{
		do
		{
			printf("Enter hypothesis ID: ");
			scanf("%d", &iHypothesis);
		} while (iHypothesis < -1 || iHypothesis > (int)(hypotheses.size()));

		if (iHypothesis == -1)
			break;

		if (pPrevHypothesis)
		{
			for (iSurfel = 0; iSurfel < pPrevHypothesis->iSurfelArray.n; iSurfel++)
				visualizer.PaintPointSet(&(pSurfels->NodeArray.Element[pPrevHypothesis->iSurfelArray.Element[iSurfel]].PtList),
				pMesh->pPolygonData, defaultColor);

			visualizer.renderer->RemoveAllViewProps();

			visualizer.SetMesh(pMesh);
		}

		pHypothesis = (RECOG::VN_::Hypothesis *)(hypotheses[iHypothesis]);

		if (pHypothesis)
		{
			for (iSurfel = 0; iSurfel < pHypothesis->iSurfelArray.n; iSurfel++)
				visualizer.PaintPointSet(&(pSurfels->NodeArray.Element[pHypothesis->iSurfelArray.Element[iSurfel]].PtList),
				pMesh->pPolygonData, selectionColor);

			pClass = classArray.Element + pHypothesis->iClass;

			d = new float[pClass->M.w];

			RVLMULMXTVECT(pClass->M.Element, pHypothesis->q, pClass->M.h, pClass->M.w, d, i, j, a);

			convexClustering.DisplayCTI(&visualizer, sampledUnitSphere.Element, sampledUnitSphere.h, d, pHypothesis->R, t);

			delete[] d;

			pPrevHypothesis = pHypothesis;
		}

		visualizer.Run();
	}
}

void VNClassifier::DetectPrimitives(
	Mesh *pMesh,
	RECOG::VN_::FitData *fitData,
	QList<RECOG::VN_::Hypothesis> *pHypothesisList,
	CRVLMem *pMem)
{

}

void VNClassifier::Interpretation()
{
	// Parameters

	int nIterations = simulatedAnnealingnIterationsPerHypothesis * hypotheses.n;
	int nPerturbances = 5;
	int perturbanceProbability = 50;

	// Initial interpretation by greedy search.

	GreedyInterpretation();

	bool *X = new bool[hypotheses.n];

	memset(X, 0, hypotheses.n * sizeof(bool));

	int i;

	for (i = 0; i < interpretation.n; i++)
		X[interpretation.Element[i]] = true;

	// 

	bool *X_ = new bool[hypotheses.n];

	bool *XBest = new bool[hypotheses.n];

	int idxRange = 100 * hypotheses.n / perturbanceProbability;

	InitInterpretationCostComputation();

	// Only for debugging purpose!
	
	//memset(X_, 0, hypotheses.n * sizeof(bool));
	////X[0] = X[1] = X[3] = X[4] = X[8] = X[11] = X[12] = X[17] = true;
	////X[0] = X[1] = X[3] = X[14] = X[8] = X[11] = X[12] = X[17] = true;
	//X_[  0] = true;
	//X_[  1] = true;
	//X_[  2] = true;
	//X_[  3] = true;
	//X_[  5] = true;
	//X_[  7] = true;
	//X_[  8] = true;
	//X_[ 20] = true;
	//X_[ 21] = true;
	//X_[ 22] = true;
	//X_[ 28] = true;
	//X_[ 29] = true;
	//X_[ 37] = true;
	//X_[ 38] = true;
	//X_[ 43] = true;
	//X_[ 52] = true;
	//X_[ 72] = true;
	//X_[ 96] = true;
	//X_[101] = true;
	//X_[108] = true;

	//float cost1 = InterpretationCost(X_);

	//memset(X_, 0, hypotheses.n * sizeof(bool));
	//X_[0] = true;
	//X_[1] = true;
	//X_[2] = true;
	//X_[3] = true;
	//X_[7] = true;
	//X_[10] = true;
	//X_[20] = true;
	//X_[21] = true;
	//X_[22] = true;
	//X_[28] = true;
	//X_[29] = true;
	//X_[37] = true;
	//X_[38] = true;
	//X_[43] = true;
	//X_[52] = true;
	//X_[72] = true;
	//X_[96] = true;
	//X_[101] = true;
	//X_[108] = true;

	//float cost2 = InterpretationCost(X_);

	//memcpy(X, X_, hypotheses.n * sizeof(bool));

	// Interpretation using local search.

	memcpy(XBest, X, hypotheses.n * sizeof(bool));

	float minCost = InterpretationCost(X);

	int j, k, iHypothesis;
	float cost;

	for (k = 0; k < nIterations; k++)
	{
		for (i = 0; i < simulatedAnnealingnSamples; i++)
		{
			memcpy(X_, X, hypotheses.n * sizeof(bool));

			for (j = 0; j < nPerturbances; j++)
			{
				iHypothesis = rand() % idxRange;

				if (iHypothesis < hypotheses.n)
					X_[iHypothesis] = !X_[iHypothesis];
			}

			cost = InterpretationCost(X_);

			if (cost < minCost)
			{
				minCost = cost;

				memcpy(XBest, X_, hypotheses.n * sizeof(bool));
			}
		}

		memcpy(X, XBest, hypotheses.n * sizeof(bool));
	}

	DeallocateInterpretationCostComputationMem();

	// Copy X to interpretation.

	interpretation.n = 0;

	for (iHypothesis = 0; iHypothesis < hypotheses.n; iHypothesis++)
		if (X[iHypothesis])
			interpretation.Element[interpretation.n++] = iHypothesis;

	// Free memory.

	delete[] X;
	delete[] X_;
	delete[] XBest;
}

void VNClassifier::GreedyInterpretation()
{
	// Sort hypotheses.

	RVL_DELETE_ARRAY(sortedHypotheses.Element);

	sortedHypotheses.Element = new SortIndex<float>[hypotheses.n];

	int i;

	for (i = 0; i < hypotheses.n; i++)
	{
		sortedHypotheses.Element[i].idx = i;
		sortedHypotheses.Element[i].cost = hypotheses.Element[i]->cost;
	}

	sortedHypotheses.n = hypotheses.n;

	BubbleSort<SortIndex<float>>(sortedHypotheses);

	// Hypothesis pruning.

	RVL_DELETE_ARRAY(interpretation.Element);

	interpretation.Element = new int[hypotheses.n];

	interpretation.n = 0;

	bool *bSurfelOccupied = new bool[pSurfels->NodeArray.n];

	memset(bSurfelOccupied, 0, pSurfels->NodeArray.n * sizeof(bool));

	int k, iHypothesis;
	RECOG::VN_::Hypothesis *pHypothesis;
	SURFEL::Object *pObject_;
	QLIST::Index *piElement;
	int occupied;

	for (i = 0; i < hypotheses.n; i++)
	{
		iHypothesis = sortedHypotheses.Element[i].idx;

		pHypothesis = hypotheses.Element[iHypothesis];

		occupied = 0;

		for (k = 0; k < pHypothesis->iSegmentArray.n; k++)
		{
			pObject_ = pObjects->objectArray.Element + pHypothesis->iSegmentArray.Element[k];

			piElement = pObject_->surfelList.pFirst;

			while (piElement)
			{
				if (bSurfelOccupied[piElement->Idx])
					occupied += pSurfels->NodeArray.Element[piElement->Idx].size;

				piElement = piElement->pNext;
			}
		}

		if (occupied * 100 / pHypothesis->support > collisionPercThr)
			continue;

		for (k = 0; k < pHypothesis->iSegmentArray.n; k++)
		{
			pObject_ = pObjects->objectArray.Element + pHypothesis->iSegmentArray.Element[k];

			piElement = pObject_->surfelList.pFirst;

			while (piElement)
			{
				bSurfelOccupied[piElement->Idx] = true;

				piElement = piElement->pNext;
			}
		}

		interpretation.Element[interpretation.n++] = iHypothesis;
	}

	delete[] bSurfelOccupied;
}

// Move to Util.h

namespace RVL
{
	int Area(Rect<int> rect)
	{
		return (rect.maxx - rect.minx + 1) * (rect.maxy - rect.miny + 1);
	}
}

void VNClassifier::ImageSegmentation(
	Mesh *pMesh,
	Array2D<int> &objectMap)
{
	float ROIBorderSize = 10;

	Rect<int> cropWin;
	cropWin.minx = 0;
	cropWin.maxx = camera.w - 1;
	cropWin.miny = 0;
	cropWin.maxy = camera.h - 1;

	int nPix = camera.w * camera.h;

	objectMap.w = camera.w;
	objectMap.h = camera.h;

	objectMap.Element = new int[nPix];

	float *ez = new float[nPix];

	bool *bValidPt = new bool[nPix];

	int iPix = 0;

	int u, v;
	float s, z, fTmp, e;
	float r[3];	
	Point *pPt;

	for (v = 0; v < camera.h; v++)
		for (u = 0; u < camera.w; u++, iPix++)
		{
			bValidPt[iPix] = false;

			objectMap.Element[iPix] = -2;

			ez[iPix] = 10.0f;

			//pPt = pMesh->NodeArray.Element + iPix;

			////bValidPt[iPix] = (RVLDOTPRODUCT3(pPt->N, pPt->N) >= 1e-10);
			//r[0] = ((float)u - camera.uc) / camera.fu;
			//r[1] = ((float)v - camera.vc) / camera.fv;
			//r[2] = 1.0f;

			//fTmp = RVLDOTPRODUCT3(NGnd, r);

			//if (fTmp < -1e-10)
			//{
			//	s = dGnd / fTmp;

			//	z = s * r[2];

			//	if (z <= maxz)
			//	{
			//		objectMap.Element[iPix] = -1;

			//		if (bValidPt[iPix])
			//		{
			//			e = RVLDOTPRODUCT3(NGnd, pPt->P) - dGnd;

			//			ez[iPix] = RVLABS(e);
			//		}
			//		else
			//			ez[iPix] = z;
			//	}
			//}
		}

	int maxnVNFeatures = 0;

	int iHypothesis;
	VN_::Hypothesis *pHypothesis;

	for (iHypothesis = 0; iHypothesis < interpretation.n; iHypothesis++)
	{
		pHypothesis = hypotheses.Element[interpretation.Element[iHypothesis]];

		if (pHypothesis->pVN->featureArray.n > maxnVNFeatures)
			maxnVNFeatures = pHypothesis->pVN->featureArray.n;
	}

	float *SDF = new float[maxnVNFeatures];

	float R[9];

	RVLUNITMX3(R);

	Rect<float> fROI;
	Rect<int> ROI;
	int iActiveFeature;
	float rM[3];
	VN_::SurfaceRayIntersection surfaceRayIntersection;

	for (iHypothesis = 0; iHypothesis < interpretation.n; iHypothesis++)
	{
		pHypothesis = hypotheses.Element[interpretation.Element[iHypothesis]];

		pSurfels->GetDepthImageROI(pHypothesis->iVertexArray, camera, fROI);

		ROI.minx = (int)(fROI.minx);
		ROI.maxx = (int)(fROI.maxx);
		ROI.miny = (int)(fROI.miny);
		ROI.maxy = (int)(fROI.maxy);

		ExpandRect<int>(&ROI, ROIBorderSize);

		CropRect<int>(ROI, cropWin);

		for (v = ROI.miny; v <= ROI.maxy; v++)
			for (u = ROI.minx; u <= ROI.maxx; u++)
			{
				//if (u == 268 && v == 263)
				//	int debug = 0;

				iPix = u + v * camera.w;

				pPt = pMesh->NodeArray.Element + iPix;

				if (bValidPt[iPix])
				{
					e = pHypothesis->pVN->Evaluate(pPt->P, SDF, iActiveFeature, true, pHypothesis->d);

					e = RVLABS(e);

					if (objectMap.Element[iPix] == -2 || e < ez[iPix])
					{
						ez[iPix] = e;

						objectMap.Element[iPix] = iHypothesis;
					}
				}
				else
				{
					r[0] = ((float)u - camera.uc) / camera.fu;
					r[1] = ((float)v - camera.vc) / camera.fv;
					r[2] = 1.0f;

					RVLNORM3(r, fTmp);

					RVLMULMX3X3TVECT(R, r, rM);

					RVLNORM3(rM, fTmp);

					surfaceRayIntersection = pHypothesis->pVN->Project(pHypothesis->d, rM);

					z = surfaceRayIntersection.s;

					if (z < ez[iPix])
					{
						ez[iPix] = z;

						objectMap.Element[iPix] = iHypothesis;
					}
				}
			}
	}

	delete[] ez;
	delete[] SDF;
	delete[] bValidPt;

#ifdef NEVER
	Rect<int> *ROI = new Rect<int>[interpretation.n];

	Rect<int> cropWin;
	cropWin.minx = 0;
	cropWin.maxx = camera.w - 1;
	cropWin.miny = 0;
	cropWin.maxy = camera.h - 1;

	int nPixels = camera.w * camera.h;

	int nAssignments = nPixels;

	int iHypothesis;
	VN_::Hypothesis *pHypothesis;
	Rect<float> fROI;
	Rect<int> *pROI;

	for (iHypothesis = 0; iHypothesis < interpretation.n; iHypothesis++)
	{
		pHypothesis = interpretation.Element[iHypothesis];

		pSurfels->GetDepthImageROI(pHypothesis->iVertexArray, camera, fROI);

		pROI = ROI + iHypothesis;

		pROI->minx = (int)(fROI.minx);
		pROI->maxx = (int)(fROI.maxx);
		pROI->miny = (int)(fROI.miny);
		pROI->maxy = (int)(fROI.maxy);

		ExpandRect<int>(pROI, ROIBorderSize);

		CropRect<int>(*pROI, cropWin);

		nAssignments += Area(*pROI);
	}

	QLIST::Entry<Pair<int, float>> *assignmentMem = new QLIST::Entry<Pair<int, float>>[nAssignments];

	QLIST::Entry<Pair<int, float>> *pAssignment = assignmentMem;

	QList<QLIST::Entry<Pair<int, float>>> *assignmentList = new QList<QLIST::Entry<Pair<int, float>>>[nPixels];

	QList<QLIST::Entry<Pair<int, float>>> *pAssignmentList = assignmentList;

	int iPix = 0;

	int u, v;
	float s, z, fTmp;
	float r[3];	
	Point *pPt;

	for (v = 0; v < camera.h; v++)
		for (u = 0; u < camera.w; u++, iPix++, pAssignmentList++)
		{
			RVLQLIST_INIT(pAssignmentList);

			r[0] = ((float)u - camera.uc) / camera.fu;
			r[1] = ((float)v - camera.vc) / camera.fv;
			r[2] = 1.0f;

			fTmp = RVLDOTPRODUCT3(NGnd, r);

			if (fTmp > 1e-10)
			{
				s = dGnd / fTmp;

				z = s * r[2];

				if (z <= maxz)
				{
					RVLQLIST_ADD_ENTRY(pAssignmentList, pAssignment);

					pAssignment->data.a = -1;

					pPt = pMesh->NodeArray.Element + iPix;

					pAssignment->data.b = (RVLDOTPRODUCT3(pPt->N, pPt->N) >= 1e-10 ? RVLDOTPRODUCT3(NGnd, pPt->P) - dGnd : z);

					pAssignment++;
				}
			}
		}

	for (iHypothesis = 0; iHypothesis < interpretation.n; iHypothesis++)
	{
		pHypothesis = interpretation.Element[iHypothesis];

		pROI = ROI + iHypothesis;

		for (v = pROI->miny; v <= pROI->maxy; v++)
			for (u = pROI->minx; u <= pROI->maxx; u++)
			{
				iPix = u + v * camera.w;

				pPt = pMesh->NodeArray.Element + iPix;

				pAssignmentList = assignmentList + iPix;

				if (RVLDOTPRODUCT3(pPt->N, pPt->N) >= 1e-10)
				{
					pHypothesis->pVN->Evaluate(pPt->P, SDF, pHypothesis->d);
				}
			}
	}

	delete[] assignmentList;
	delete[] assignmentMem;
#endif
}

void VNClassifier::ComputeClassPropertiesForGrasping(int iClass, VN *pModel)
{
	

}
void VNClassifier::SampleSO3()
{
	int nRotSamples = 16 * sampledUnitSphere.h;

	SO3Samples.a = 3;
	SO3Samples.b = 3;
	SO3Samples.c = nRotSamples;

	SO3Samples.Element = new float[SO3Samples.a * SO3Samples.b * SO3Samples.c];

	float R[9];

	float *X = R;
	float *Y = R + 3;
	float *Z = R + 6;

	float dphi = 2.0f * PI / 16.0f;

	int iRotSample = 0;

	int i, j, k, i_;
	float *Z_;
	float X_[3], Y_[3], VTmp[3];
	float fTmp;
	float phi, cs, sn;
	float *R__;

	for (i = 0; i < sampledUnitSphere.h; i++)
	{
		Z_ = sampledUnitSphere.Element + sampledUnitSphere.w * i;

		RVLCOPY3VECTOR(Z_, Z);

		RVLORTHOGONAL3(Z_, X_, i_, j, k, fTmp);

		RVLCROSSPRODUCT3(Z_, X_, Y_);

		for (j = 0; j < 16; j++)
		{
			phi = (float)j * dphi;

			cs = cos(phi);
			sn = sin(phi);

			RVLSCALE3VECTOR(X_, cs, X);
			RVLSCALE3VECTOR(Y_, -sn, VTmp);
			RVLSUM3VECTORS(X, VTmp, X);

			RVLSCALE3VECTOR(X_, sn, Y);
			RVLSCALE3VECTOR(Y_, cs, VTmp);
			RVLSUM3VECTORS(Y, VTmp, Y);

			R__ = SO3Samples.Element + 9 * iRotSample;

			RVLCOPYMX3X3(R, R__);

			iRotSample++;
		}
	}
}

float VNClassifier::ProjectToLatentSubspace(
	float *d,
	uchar *bd,
	float *dU,
	uchar *bdU,
	ClassData *pClass,
	float *qOpt,
	int &iR)
{
	VN *pModel = models[pClass->iMetaModel];

	Array2D<float> M = pClass->M;

	int m = M.h;

	int m2 = m * m;

	Array<int> iV;

	iV.Element = new int[M.w];

	cv::Mat A;

	A.create(m, m, CV_32FC1);

	float *A_ = (float *)(A.data);

	cv::Mat b;

	b.create(m, 1, CV_32FC1);

	float *b_ = (float *)(b.data);

	float *a = new float[m];

	cv::Mat q;

	q.create(m, 1, CV_32FC1);

	float *q_ = (float *)(q.data);

	float *dSV = new float[M.w];

	float minE = -1.0f;

	int i, j, l, iN;
	float *d__, *dU__, *dS;
	uchar *bd__, *bdU__, *bdS;
	int iFeature;
	float *M_;
	float dS_;
	VN_::ModelCluster *pModelCluster;
	float dM, e, E;

	//for (int k = 0; k < 1000; k++)
	//{
	for (i = 0; i < SO3Samples.c; i++)
	{
		d__ = d + i * sampledUnitSphere.h;
		bd__ = bd + i * sampledUnitSphere.h;
		dU__ = dU + i * sampledUnitSphere.h;
		bdU__ = bdU + i * sampledUnitSphere.h;

		memset(A_, 0, m2 * sizeof(float));

		memset(b_, 0, m * sizeof(float));

		iV.n = 0;

		pModelCluster = pModel->modelClusterList.pFirst;

		while (pModelCluster)
		{
			if (pModelCluster->type == RVLVN_CLUSTER_TYPE_CONVEX)
			{
				dS = d__;
				bdS = bd__;
			}
			else if (pModelCluster->type == RVLVN_CLUSTER_TYPE_CONCAVE)
			{
				dS = dU__;
				bdS = bdU__;
			}

			for (iFeature = pModelCluster->iFeatureInterval.a; iFeature <= pModelCluster->iFeatureInterval.b; iFeature++)
			{
				iN = pModelCluster->iN[iFeature - pModelCluster->iFeatureInterval.a];

				if (bdS[iN])
				{
					M_ = M.Element + iFeature;

					for (j = 0; j < m; j++)
						a[j] = M_[j * M.w];

					for (l = 0; l < m; l++)
						for (j = l; j < m; j++)
							A_[l * m + j] += (a[l] * a[j]);

					dS_ = dS[iN];

					for (j = 0; j < m; j++)
						b_[j] += (a[j] * dS_);

					iV.Element[iV.n] = iFeature;

					dSV[iV.n] = dS_;

					iV.n++;
				}
			}

			pModelCluster = pModelCluster->pNext;
		}

		for (l = 1; l < m; l++)
			for (j = 0; j < l; j++)
				A_[l * m + j] = A_[j * m + l];

		cv::solve(A, b, q);

		E = 0.0f;

		for (j = 0; j < iV.n; j++)
		{
			M_ = M.Element + iV.Element[j];

			dM = 0.0f;

			for (l = 0; l < m; l++)
				dM += M_[l * M.w] * q_[l];

			e = dSV[j] - dM;

			E += e * e;
		}

		if (minE < 0.0f || E < minE)
		{
			minE = E;

			iR = i;

			memcpy(qOpt, q_, M.h * sizeof(float));
		}
	}	// for every descriptor

	delete[] a;
	delete[] iV.Element;
	delete[] dSV;

	return minE;
}

float VNClassifier::ProjectToLatentSubspace(
	float *d,
	uchar *bd,
	RECOG::ClassData *pClass,
	RECOG::VN_::FitData *pFitData,
	float *qOpt,
	int &iR,
	int &nb,
	bool &bAllDoFsVisible)
{
	VN *pModel = models[pClass->iMetaModel];

	//int nqs = M.h - 4;

	//float *qs = new float[nqs];

	//float *qs_ = q_ + 4;

	float minE = -1.0f;

	bAllDoFsVisible = false;

	//printf("lambda=%f, sqrtbeta=%f\n", lambda_, sqrtbeta);

	//FILE *fpDebug = fopen("C:\\RVL\\ExpRez\\q.txt", "w");

	//FILE *fpDebug2 = fopen("C:\\RVL\\ExpRez\\A.txt", "w");

	int i;
	float *dS;
	uchar *bdS;
	float E;

	//for (int k = 0; k < 1000; k++)
	//{
	for (i = 0; i < SO3Samples.c; i++)
	{
		//if (i == 272)
		//	int debug = 0;

		dS = d + i * sampledUnitSphere.h;
		bdS = bd + i * sampledUnitSphere.h;

		if (FitLS(dS, bdS, pClass, pFitData, E))
		{
			bAllDoFsVisible = true;

			if (minE < 0.0f || E < minE)
			{
				minE = E;

				iR = i;

				memcpy(qOpt, pFitData->q_, pClass->M.h * sizeof(float));

				nb = pFitData->iV.n;
			}
		}
	}	// for every descriptor

	//fclose(fpDebug);
	//fclose(fpDebug2);

	//delete[] qs;

	return minE;
}

//Requires prealocated variable allbuffer since results wil be saved into her sequentially. (tuple<E, q, iR, nb, bAllDoFsVisible>)
void VNClassifier::ProjectToLatentSubspace(
	float *d,
	uchar *bd,
	RECOG::ClassData *pClass,
	RECOG::VN_::FitData *pFitData,
	std::vector<std::tuple<float, float*, int, int, bool>> *allbuffer)
{
	VN *pModel = models[pClass->iMetaModel];

	float minE = -1.0f;

	bool bAllDoFsVisible = false;

	int i;
	float *dS;
	uchar *bdS;
	float E;

	for (i = 0; i < SO3Samples.c; i++)
	{
		dS = d + i * sampledUnitSphere.h;
		bdS = bd + i * sampledUnitSphere.h;

		if (FitLS(dS, bdS, pClass, pFitData, E))
		{
			bAllDoFsVisible = true;

			std::get<0>((*allbuffer)[i]) = E;
			memcpy(std::get<1>((*allbuffer)[i]), pFitData->q_, pClass->M.h * sizeof(float));
			std::get<2>((*allbuffer)[i]) = i;
			std::get<3>((*allbuffer)[i]) = pFitData->iV.n;
			std::get<4>((*allbuffer)[i]) = bAllDoFsVisible;
		}
	}	// for every descriptor

}


bool VNClassifier::FitLS(
	float *d,
	uchar *bd,
	RECOG::ClassData *pClass,
	VN_::FitData *pData,
	float &E)
{
	Array2D<float> M = pClass->M;
	float *A_ = pData->A_;
	float *b_ = pData->b_;
	float *q_ = pData->q_;
	float *a = pData->a;
	int m = M.h;

	memset(A_, 0, m * m * sizeof(float));

	memset(b_, 0, m * sizeof(float));

	pData->iV.n = 0;
	float w = 0.0f;
	float fn = (float)(M.w);
	bool bFullRank = false;

	int i, j, l;
	int iFeature;
	float dS_;
	float beta__, lambda__;
	float *M_;
	//float *AH;
	float dM, e, maxe;

	for (iFeature = 0; iFeature < M.w; iFeature++)
	{
		dS_ = d[iFeature];

		//if (bdS[iFeature])
		if (!bDepthImage || dS_ < 0.0f)
		{
			if (!bDepthImage || bd[iFeature])
			{
				beta__ = 1.0f;

				pData->iV.Element[pData->iV.n] = iFeature;

				pData->dSV[pData->iV.n] = dS_;

				pData->iV.n++;
			}
			else
				beta__ = pData->sqrtbeta;

			M_ = M.Element + iFeature;

			for (j = 0; j < m; j++)
				a[j] = beta__ * M_[j * M.w];

			for (l = 0; l < m; l++)
				for (j = l; j < m; j++)
					A_[l * m + j] += (a[l] * a[j]);

			for (j = 0; j < m; j++)
				b_[j] += (beta__ * a[j] * dS_);

			w += (beta__ * beta__);
		}
	}

	lambda__ = pData->lambda1 * (fn - w) + pData->lambda0;

	for (l = 1; l < m; l++)
	{
		for (j = 0; j < l; j++)
			A_[l * m + j] = A_[j * m + l];

		if (l > fitParams.regularization)
			A_[l * m + l] += lambda__;
	}

	//PrintMatrix<float>(fpDebug2, A_, 1, m2);

	if (cv::solve(pData->A, pData->b, pData->q))
	{
		//PrintMatrix<float>(fpDebug, q_, 1, m);

		if (q_[3] >= 0.0f)
		{
			bFullRank = true;

			E = 0.0f;

			for (j = 0; j < pData->iV.n; j++)
			{
				M_ = M.Element + pData->iV.Element[j];

				dM = 0.0f;

				for (l = 0; l < m; l++)
					dM += (M_[l * M.w] * q_[l]);

				e = pData->dSV[j] - dM;

				E += (e * e);
			}

			//E /= ((float)(pData->iV.n));

			//s = q_[3];

			////RVLSCALEVECTOR(qs_, s, qs, nqs, j);

			//maxe = 0.0f;

			//for (j = 0; j < pClass->nHull; j++)
			//{
			//	AH = pClass->A + nqs * j;

			//	RVLDOTPRODUCT(AH, qs, nqs, e, l);

			//	e -= (s * pClass->b[j]);

			//	if (e > maxe)
			//		maxe = e;
			//}

			//E += (maxe * maxe);

			return true;
		}	// If q[3] > 0
		else
			return false;
	}	// If system A, b, q is solvable. 
	else
		return false;
}

bool VNClassifier::FitRotLS(
	Array<int> iVertexArray,
	float o,
	float *NM,
	RECOG::ClassData *pClass,
	float *RIn,
	int nIterations,
	bool bNormalTest,
	bool bVisibilityTest,
	RECOG::VN_::FitData *pData,
	float *qOpt,
	float *ROpt,
	float &E)
{
	float lambdaRotUp = 2.0f;
	float lambdaRotDown = 1.0f / lambdaRotUp;

	Array2D<float> M = pClass->M;
	float *A_ = pData->A_;
	float *b_ = pData->b_;
	float *x = pData->q_;
	float *q_ = x + 3;
	float *a = pData->a;
	float *aq = a + 3;
	int m = M.h;
	int nx = m + 3;

	// R <- RIn

	float R[9];

	RVLCOPYMX3X3(RIn, R);

	// q <- q - [pData->T * R' * pData->Pc; zeros(m - 3, 1)]

	float PcM[3];

	RVLMULMX3X3TVECT(R, pData->Pc, PcM);

	float qc[3];

	float *T = pClass->T;

	RVLMULMX3X3VECT(T, PcM, qc);

	RVLDIF3VECTORS(q_, qc, q_);

	// main loop

	memcpy(qOpt, q_, m * sizeof(float));

	//pData->iV.n = 0;
	float fn = (float)(M.w);
	bool bFullRank = false;
	float E_ = -1.0f;

	int i, j, k, l;
	int iFeature;
	float dS, dSc;
	float beta__, lambdaRot, lambda__;
	float *M_;
	//float *AH;
	float dM, e, maxe, dSmax, d_, dSmaxValid;
	float NS[3];
	float *NM_;
	SURFEL::Vertex *pVertex;
	float *P;
	bool bdmax, bdmaxValid;
	float distFromNHull;
	bool bd;
	int iVertex, iVertex_;
	float th;
	float z[3];
	float dR[9], R_[9], R__[9], V3Tmp[3];
	float w;
	float maxTrA, fTmp;
	int nb;
	float beta2;

	for (k = 0; k < nIterations; k++)
	{
		w = 0.0f;
		E = 0.0f;
		nb = 0;

		memset(A_, 0, nx * nx * sizeof(float));

		memset(b_, 0, nx * sizeof(float));

		//FILE *fpDebug = fopen("C:\\RVL\\ExpRez\\a_.txt", "w");
		//FILE *fpDebug3 = fopen("C:\\RVL\\ExpRez\\P.txt", "w");
		//FILE *fpDebug4 = fopen("C:\\RVL\\ExpRez\\debug.txt", "w");

		for (iFeature = 0; iFeature < M.w; iFeature++)
		{
			//fprintf(fpDebug4, "iFeature=%d\n", iFeature);

			NM_ = NM + 3 * iFeature;

			RVLMULMX3X3VECT(R, NM_, NS);

			bdmax = bdmaxValid = false;

			for (j = 0; j < iVertexArray.n; j++)
			{
				iVertex_ = iVertexArray.Element[j];

				//fprintf(fpDebug4, "iVertex_=%d\n", iVertex_);

				pVertex = pSurfels->vertexArray.Element[iVertex_];

				P = pVertex->P;

				d_ = RVLDOTPRODUCT3(NS, P);

				if (bdmax)
				{
					if (o * d_ > o * dSmax)
					{
						dSmax = d_;

						iVertex = iVertex_;
					}
				}
				else
				{
					dSmax = d_;

					iVertex = iVertex_;

					bdmax = true;
				}

				if (bNormalTest)
				{
					if (d_ < 0.0f || !bVisibilityTest)
					{
						distFromNHull = pSurfels->DistanceFromNormalHull(pVertex->normalHull, NS);

						if (distFromNHull <= maxDistFromNHull)
						{
							if (bdmaxValid)
							{
								if (o * d_ > o * dSmaxValid)
									dSmaxValid = d_;
							}
							else
							{
								dSmaxValid = d_;

								bdmaxValid = true;
							}
						}
					}
				}
			}	// for each vertex

			if (dSmax >= 0.0f)
				continue;

			bd = false;

			if (bdmaxValid)
				bd = (o * (dSmax - dSmaxValid) <= maxedmax);

			// Comment this!
			//bd = true;

			if (bd)
			{
				beta__ = 1.0f;

				//pData->iV.Element[pData->iV.n] = iFeature;

				//pData->dSV[pData->iV.n] = dS;

				//pData->iV.n++;
			}
			else
				beta__ = pData->sqrtbeta;

			beta2 = beta__ * beta__;

			dS = dSmax;

			dSc = dS - RVLDOTPRODUCT3(NS, pData->Pc);

			P = pData->P + 3 * iVertex;

			//fprintf(fpDebug3, "%d\t", iVertex);

			//PrintMatrix<float>(fpDebug3, P, 1, 3);

			RVLCROSSPRODUCT3(P, NS, a);

			RVLSCALE3VECTOR(a, beta__, a);

			M_ = M.Element + iFeature;

			for (j = 0; j < m; j++)
				aq[j] = beta__ * M_[j * M.w];

			for (l = 0; l < nx; l++)
				for (j = l; j < nx; j++)
					A_[l * nx + j] += (a[l] * a[j]);

			//PrintMatrix<float>(fpDebug, a, 1, nx);

			//fTmp = beta__ * e;
			fTmp = beta__ * dSc;

			RVLSCALE3VECTOR(a, fTmp, V3Tmp);

			RVLSUM3VECTORS(b_, V3Tmp, b_);

			for (j = 3; j < nx; j++)
				b_[j] += (beta__ * a[j] * dSc);

			w += beta2;

			dM = 0.0f;

			for (l = 0; l < m; l++)
				dM += (M_[l * M.w] * q_[l]);

			e = dSc - dM;

			//if (bd)
			{
				E += (beta2 * e * e);
				nb++;
			}
		}	// for (iFeature = 0; iFeature < M.w; iFeature++) 

		//fclose(fpDebug);
		//fclose(fpDebug3);
		//fclose(fpDebug4);

		E /= ((float)nb);

		if (E_ > 0.0f && E > E_)
		{
			RVLCOPYMX3X3(R_, R);

			if (k >= nIterations - 1)
				break;
			else
				lambdaRot *= lambdaRotUp;
		}
		else if (E_ < 0.0f)
		{
			maxTrA = 0.0;

			for (i = 3; i < nx; i++)
			{
				fTmp = A_[i * nx + i];

				if (fTmp > maxTrA)
					maxTrA = fTmp;
			}

			lambdaRot = 0.1f * maxTrA;

			E_ = E;

			RVLCOPYMX3X3(R, R_);
		}
		else
		{
			lambdaRot *= lambdaRotDown;

			E_ = E;

			RVLCOPYMX3X3(R, R_);

			memcpy(qOpt, q_, m * sizeof(float));
		}

		if (k >= nIterations)
			break;

		lambda__ = pData->lambda1 * (fn - w) + pData->lambda0;

		for (l = 0; l < 3; l++)
			A_[l * nx + l] += lambdaRot;

		for (l = 1; l < nx; l++)
		{
			for (j = 0; j < l; j++)
				A_[l * nx + j] = A_[j * nx + l];

			if (l > 6)
				A_[l * nx + l] += lambda__;
		}

		//FILE *fpDebug2 = fopen("C:\\RVL\\ExpRez\\A.txt", "w");

		//PrintMatrix<float>(fpDebug2, A_, nx, nx);

		//fclose(fpDebug2);
		
		if (!cv::solve(pData->A, pData->b, pData->q))
		{
			RVLCOPYMX3X3(R_, R);

			break;
		}

		//int k_ = 0;

		//while (!cv::solve(pData->A, pData->b, pData->q) && k_ < 10)
		//{
		//	for (l = 0; l < 3; l++)
		//		A_[l * nx + l] += lambdaRot;	

		//	lambdaRot *= lambdaRotUp;

		//	k_++;
		//}

		//if (k_ >= 10)
		//	return false;

		// Only for debugging purpose!



		///

		if (q_[3] < 0.0f)
			return false;

		//PrintMatrix<float>(fpDebug, q_, 1, m);

		bFullRank = true;

		// dphi = || dphi ||

		th = sqrt(RVLDOTPRODUCT3(x, x));

		//printf("th=%f\n", th);

		if (RVLABS(th) >= 1e-8)
		{
			// z <- dphi / || dphi ||

			RVLSCALE3VECTOR2(x, th, z);

			// dR <- Rot(z, th) (angle axis to rotation matrix)

			AngleAxisToRot<float>(z, th, dR);

			// R <- dR * R

			RVLMXMUL3X3(dR, R, R__);

			RVLCOPYMX3X3(R__, R);
		}
	}	// for (k = 0; k < nIterations; k++)

	// ROpt <- R

	RVLCOPYMX3X3(R, ROpt);

	// q <- q - [pData->T * R' * pData->Pc; zeros(m - 3, 1)]

	RVLMULMX3X3TVECT(ROpt, pData->Pc, PcM);

	RVLMULMX3X3VECT(T, PcM, qc);

	RVLSUM3VECTORS(qOpt, qc, qOpt);

	return true;
}

// Mathematics for this function is explained in ARP3D.TR4.11, CTI Fitting II

float VNClassifier::FitCTI(
	Mesh *pMesh,
	Array<int> iSurfelArray,
	Array<int> iVertexArray,
	Camera camera,
	int nIterations,
	float *R,
	float *d,
	uchar *bd)
{
	nIterations = 5;

	float lambdaUp = 2.0f;
	float lambdaDown = 1.0f / lambdaUp;
	float lambda = 0.1f;

	SURFEL::SceneSamples sceneSamples;

	sceneSamples.PGnd = NULL;

	pSurfels->SampleSurfelSet(pMesh, iSurfelArray, iVertexArray, camera, sceneSamples, true, false);

	float R0[9];

	RVLUNITMX3(R0);

	float *d_ = new float[sampledUnitSphere.h];

	uchar *bd_ = new uchar[sampledUnitSphere.h];

	float *d__ = new float[sampledUnitSphere.h];

	float *NS = new float[sampledUnitSphere.h * sampledUnitSphere.w];

	int *iV = new int[sampledUnitSphere.h];

	cv::Mat A;

	A.create(3, 3, CV_32FC1);

	float *A_ = (float *)(A.data);

	cv::Mat b;

	b.create(3, 1, CV_32FC1);

	float *b_ = (float *)(b.data);

	cv::Mat x(3, 1, CV_32FC1);

	float *x_ = (float *)(x.data);

	Array<SURFEL::NormalHullElement> NHull;

	NHull.Element = NULL;
	NHull.n = 0;

	//float *dFace = new float[sampledUnitSphere.h];

	int *nFace = new int[sampledUnitSphere.h];

	float *PcFace = new float[3 * sampledUnitSphere.h];

	int *iAssociatedFace = new int[sceneSamples.PtArray.n];

	//float U[5][3] = {
	//	{ 0.0f, 0.0f, 0.0f },
	//	{ 1.0f, 0.0f, 0.0f },
	//	{ 0.0f, 1.0f, 0.0f },
	//	{ -1.0f, 0.0f, 0.0f },
	//	{ 0.0f, -1.0f, 0.0f } };

	//float phi = 0.125 * PI;

	int iZ_[3] = { 6, 8, 57 };

	float *Z0 = R0 + 6;

	float EOpt = -1.0f;

	int i, k, iX, iZ;
	int iFace;
	float *P;
	float *NM, *NS_, *PcFace_, *Z;
	float e, e2, maxe, th, E, E_, fnFace, phi, sf, ca, sa, al;	
	bool bFirst;
	int iAssociatedFace_;
	float dP[3], J[3], A__[9], b__[3], z[3], dR[9], R_[9], R__[9], ROpt[9], RZ[9], RX[9], V3Tmp[3];
	SURFEL::Vertex *pVertex;

	for (iZ = 0; iZ < 4; iZ++)
	//iZ = 3;
	{
		if (iZ == 0)
		{
			RVLUNITMX3(RZ)
		}
		else
		{
			Z = sampledUnitSphere.Element + 3 * iZ_[iZ - 1];

			RVLCROSSPRODUCT3(Z0, Z, V3Tmp);

			RVLNORM3(V3Tmp, sf);

			phi = asin(sf);

			AngleAxisToRot<float>(V3Tmp, phi, RZ);
		}

		for (iX = 0; iX < 3; iX++)
		//iX = 1;
		{
			al = 0.125 * PI * (float)iX;

			ca = cos(al);
			sa = sin(al);

			RVLROTZ(ca, sa, RX);

			RVLMXMUL3X3(RZ, RX, R);

			//float RGT_[9];

			//float *XGT_ = RGT_;
			//float *YGT_ = RGT_ + 3;
			//float *ZGT_ = RGT_ + 6;

			//RVLSET3VECTOR(ZGT_, -0.067668, -0.735154, -0.674515);
			//RVLSET3VECTOR(XGT_, -0.51496, 0.560113, -0.615370);
			//RVLCROSSPRODUCT3(ZGT_, XGT_, YGT_);

			//float debug = RVLMULROWCOL3(RGT_, R, 0, 0);

			//float RGT[9];

			//RVLCOPYMX3X3T(RGT_, RGT);

			//GetAngleAxis(RGT, z, phi);

			//float fTmp;

			//RVLNORM3(z, fTmp);

			//AngleAxisToRot<float>(z, phi, dR);

			//RVLMXMUL3X3(dR, R0, R);

			//RVLCOPYMX3X3(R0, R);

			E_ = -1.0f;

			for (k = 0; k < nIterations; k++)
			{
				//if (k == 0)
				ComputeDescriptor(iVertexArray, NHull, R, d_, bd_, NS, iV);

				RVLNULLMX3X3(A_);

				RVLNULL3VECTOR(b_);

				//memset(dFace, 0, sampledUnitSphere.h * sizeof(float));

				memset(nFace, 0, sampledUnitSphere.h * sizeof(int));

				memset(PcFace, 0, 3 * sampledUnitSphere.h * sizeof(float));

				E = 0.0f;

				for (i = 0; i < sceneSamples.PtArray.n; i++)
				{
					if (sceneSamples.status[i] != 1)
						continue;

					P = sceneSamples.PtArray.Element[i];

					bFirst = true;

					for (iFace = 0; iFace < sampledUnitSphere.h; iFace++)
					{
						if (d_[iFace] >= 0.0f)
							continue;

						NS_ = NS + 3 * iFace;

						//NM = sampledUnitSphere.Element + 3 * iFace;

						//RVLMULMX3X3VECT(R, NM, NS_);

						e = RVLDOTPRODUCT3(NS_, P) - d_[iFace];

						if (bFirst)
						{
							maxe = e;

							iAssociatedFace_ = iFace;

							bFirst = false;
						}
						else if (e > maxe)
						{
							maxe = e;

							iAssociatedFace_ = iFace;
						}
					}

					NS_ = NS + 3 * iAssociatedFace_;

					//dFace[iAssociatedFace] += (RVLDOTPRODUCT3(NS_, P) - sceneSamples.g[i]);

					nFace[iAssociatedFace_]++;

					PcFace_ = PcFace + 3 * iAssociatedFace_;

					RVLSUM3VECTORS(PcFace_, P, PcFace_);

					iAssociatedFace[i] = iAssociatedFace_;
				}	// for every sample point

				for (iFace = 0; iFace < sampledUnitSphere.h; iFace++)
				{
					if (nFace[iFace] == 0)
						continue;

					fnFace = (float)(nFace[iFace]);

					//d_[iFace] = dFace[iFace] / fnFace;

					PcFace_ = PcFace + 3 * iFace;

					RVLSCALE3VECTOR2(PcFace_, fnFace, PcFace_);

					NS_ = NS + 3 * iFace;

					d_[iFace] = RVLDOTPRODUCT3(NS_, PcFace_);
				}

				for (i = 0; i < sceneSamples.PtArray.n; i++)
				{
					if (sceneSamples.status[i] != 1)
						continue;

					P = sceneSamples.PtArray.Element[i];

					iAssociatedFace_ = iAssociatedFace[i];

					//pVertex = pSurfels->vertexArray.Element[iV[iAssociatedFace]];

					//RVLDIF3VECTORS(P, pVertex->P, dP);

					PcFace_ = PcFace + 3 * iAssociatedFace_;

					RVLDIF3VECTORS(P, PcFace_, dP);

					NS_ = NS + 3 * iAssociatedFace_;

					RVLCROSSPRODUCT3(dP, NS_, J);

					e = RVLDOTPRODUCT3(NS_, dP) - sceneSamples.g[i];

					E += (e * e);

					RVLSCALE3VECTOR(J, e, b__);

					RVLSUM3VECTORS(b_, b__, b_);

					RVLVECTCOV3(J, A__);

					RVLSUMMX3X3(A_, A__, A_);
				}	// for every sample point

				if (E_ > 0.0f && E > E_)
				{
					RVLCOPYMX3X3(R_, R);

					if (k >= nIterations - 1)
					{
						memcpy(d_, d__, sampledUnitSphere.h * sizeof(float));

						break;
					}
					else
						lambda *= lambdaDown;
				}
				else
				{
					//lambda *= lambdaUp;

					if (lambda > 1.0f)
						lambda = 1.0f;

					E_ = E;

					RVLCOPYMX3X3(R, R_);

					memcpy(d__, d_, sampledUnitSphere.h * sizeof(float));
				}

				if (k >= nIterations - 1)
					break;

				RVLCOMPLETESIMMX3(A_);

				cv::solve(A, b, x);

				//RVLSCALE3VECTOR(x_, lambda, x_);

				// dphi = || dphi ||

				th = sqrt(RVLDOTPRODUCT3(x_, x_));

				if (RVLABS(th) >= 1e-8)
				{
					// z <- dphi / || dphi ||

					RVLSCALE3VECTOR2(x_, th, z);

					// dR <- Rot(z, th) (angle axis to rotation matrix)

					if (th > 0.1f)
						th = 0.1f;
					else if (th < -0.1f)
						th = -0.1f;

					AngleAxisToRot<float>(z, th, dR);

					// R <- dR * R

					RVLMXMUL3X3(dR, R, R__);

					RVLCOPYMX3X3(R__, R);
				}
				else
					break;
			}	// optimization iteration.

			if (EOpt < 0.0f || E < EOpt)
			{
				EOpt = E;

				RVLCOPYMX3X3(R, ROpt);

				memcpy(d, d_, sampledUnitSphere.h * sizeof(float));

				memcpy(bd, bd_, sampledUnitSphere.h * sizeof(uchar));
			}
		}	// for (iX = 0; iX < 3; iX++)
	}	// for (iZ = 0; iZ < 4; iZ++)

	SURFEL::DeleteSceneSamples(sceneSamples);

	delete[] d_;
	delete[] bd_;
	delete[] d__;
	delete[] NS;
	delete[] iV;
	//delete[] dFace;
	delete[] nFace;
	delete[] PcFace;
	delete[] iAssociatedFace;

	RVLCOPYMX3X3(ROpt, R);

	return EOpt;
}

// Function FitCTI computes localy optimal rotation matrix R, which aligns convex template stored in sampledUnitSphere with
// the face normals of the convex mesh pConvexHull. The initial rotation matrix RInit is provided as an argument.
// The mathematics for FitCTI is described in ARP3D.TR3.9.

float VNClassifier::FitCTI(
	Mesh *pConvexHull,
	float *U,
	float *RInit,
	bool *mask,
	int nIterations,
	float *R,
	int *faceCorrespondence,
	Array<int> *piModelNormalArray)
{
	RVLCOPYMX3X3(RInit, R);

	float kR = 0.5f / (sigmaR * sigmaR);

	float V3Tmp2[3];
	float w, maxw;
	float xx, yy, zz, xy, zx, yz;
	int i, j, k, i_, j_, k_, iCorrespondence, iNormal;
	float *uS, *a;
	float uM[3], V3Tmp[3], b[3], z[3];
	MESH::Face *pFace;
	float dR[9], R_[9], U_[9], M3x3Tmp[9];
	float detU;
	float invU[9];
	float dphi[3];
	float th;

	for (k = 0; k < nIterations; k++)
	{
		RVLNULL3VECTOR(V3Tmp2);

		if (U == NULL)
			RVLNULLMX3X3(invU);

		for (j = 0; j < pConvexHull->iVisibleFaces.n; j++)
		{
			// uS <- normal of the j-th visible face

			pFace = pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[j]];

			uS = pFace->N;

			if (piModelNormalArray)
			{
				RVLMULMX3X3TVECT(R, uS, uM);

				maxw = -2.0f;

				for (i = 0; i < piModelNormalArray->n; i++)
				{
					iNormal = piModelNormalArray->Element[i];

					a = sampledUnitSphere.Element + 3 * iNormal;

					w = RVLDOTPRODUCT3(a, uM);

					if (w > maxw)
					{
						maxw = w;

						iCorrespondence = iNormal;
					}
				}
			}
			else
			RVLVN_GET_CLOSEST_CONVEX_TEMPLATE_ELEMENT(convexTemplateLUT, uS, R, uM, i_, j_, k_, iCorrespondence);

			faceCorrespondence[j] = iCorrespondence;
			
			if (mask[iCorrespondence])
			{
				// a <- the closest template vector

				a = sampledUnitSphere.Element + 3 * iCorrespondence;

				// V3Tmp2 <- V3Tmp2 + w * (uM x a), where w = the area of the j-th visible face * exp(-||a - uM||^2 / (2 * sigmaR^2))

				RVLDIF3VECTORS(a, uM, V3Tmp);

				w = pFace->Area * exp(-kR * RVLDOTPRODUCT3(V3Tmp, V3Tmp));

				RVLCROSSPRODUCT3(uM, a, V3Tmp);

				RVLSCALE3VECTOR(V3Tmp, w, V3Tmp);

				RVLSUM3VECTORS(V3Tmp2, V3Tmp, V3Tmp2);

				if (U == NULL)
				{					
					xx = w * uS[0] * uS[0];
					yy = w * uS[1] * uS[1];
					zz = w * uS[2] * uS[2];
					xy = w * uS[0] * uS[1];
					yz = w * uS[1] * uS[2];
					zx = w * uS[2] * uS[0];

					RVLMXEL(M3x3Tmp, 3, 0, 0) = -(yy + zz);
					RVLMXEL(M3x3Tmp, 3, 0, 1) = xy;
					RVLMXEL(M3x3Tmp, 3, 0, 2) = -zx;
					RVLMXEL(M3x3Tmp, 3, 1, 1) = -(zz + xx);
					RVLMXEL(M3x3Tmp, 3, 1, 2) = yz;
					RVLMXEL(M3x3Tmp, 3, 2, 2) = -(xx + yy);

					RVLSUMMX3X3UT(invU, M3x3Tmp, invU);
				}
			}
		}

		// Only for debugging purpose!

		//if (pConvexHull->iVisibleFaces.n >= 31)
		//	if (RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[19] + 2]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[22] + 2]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[25] + 2]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[26] + 2]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[31] + 2]) < 1e-3 ||
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[19] + 0]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[22] + 0]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[25] + 0]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[26] + 0]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[31] + 0]) < 1e-3 ||
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[19] + 1]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[22] + 1]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[25] + 1]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[26] + 1]) < 1e-3 &&
		//		RVLABS(sampledUnitSphere.Element[3 * faceCorrespondence[31] + 1]) < 1e-3)
		//		int debug = 0;

		//if (pConvexHull->iVisibleFaces.n >= 19)
		//	if (faceCorrespondence[1] % 11 == 0 && 
		//		faceCorrespondence[6] % 11 == 0 && 
		//		faceCorrespondence[10] % 11 == 0 && 
		//		faceCorrespondence[11] % 11 == 0 && 
		//		faceCorrespondence[13] % 11 == 0)
		//		int debug = 0;

		//if (pConvexHull->iVisibleFaces.n >= 12)
		//	if (faceCorrespondence[0] % 11 == 0 && 
		//		faceCorrespondence[1] % 11 == 0 &&
		//		faceCorrespondence[2] % 11 == 0 &&
		//		faceCorrespondence[11] % 11 == 0)
		//		int debug = 0;

		///

		// b <- R * V3Tmp2

		RVLMULMX3X3VECT(R, V3Tmp2, b);

		// dphi = -inv(U) * b

		if (U == NULL)
		{			
			RVLINVCOV3(invU, U_, detU);

			RVLCOMPLETESIMMX3(U_);

			RVLMULMX3X3VECT(U_, b, dphi);
		}
		else
		{
		RVLMULMX3X3VECT(U, b, dphi);
		}

		RVLNEGVECT3(dphi, dphi);

		// Only for debugging purpose!

		//float E = 0.0f;

		//float e;

		//for (j = 0; j < pConvexHull->iVisibleFaces.n; j++)
		//{
		//	pFace = pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[j]];

		//	uS = pFace->N;

		//	RVLMULMX3X3TVECT(R, uS, uM);

		//	a = sampledUnitSphere.Element + 3 * faceCorrespondence[j];

		//	RVLDIF3VECTORS(uM, a, V3Tmp);

		//	e = RVLDOTPRODUCT3(V3Tmp, V3Tmp);

		//	E += (pFace->Area * e);
		//}

		///

		// dphi = || dphi ||

		th = sqrt(RVLDOTPRODUCT3(dphi, dphi));

		if (RVLABS(th) >= 1e-8)
		{
			// z <- dphi / || dphi ||
		
			RVLSCALE3VECTOR2(dphi, th, z);

			// dR <- Rot(z, th) (angle axis to rotation matrix)

			AngleAxisToRot<float>(z, th, dR);

			// R <- dR' * R

			RVLMXMUL3X3T1(dR, R, R_);

			RVLCOPYMX3X3(R_, R);
		}
		else
			break;

		// Only for debugging purpose!

		//E = 0.0f;
		//float E2 = 0.0f;
		//float E3 = 0.0f;

		//for (j = 0; j < pConvexHull->iVisibleFaces.n; j++)
		//{
		//	pFace = pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[j]];

		//	uS = pFace->N;

		//	// e <- || R' * uS - a ||^2 = || (dR' * RInit)' * uS - a ||^2 = || RInit' * dR * uS - a ||^2 = || RInit * a - dR * uS ||^2

		//	RVLMULMX3X3TVECT(R, uS, uM);

		//	a = sampledUnitSphere.Element + 3 * faceCorrespondence[j];

		//	if (j == 31)
		//	{
		//		if (RVLABS(a[2]) < 1e-3)
		//			int debug = 0;
		//	}

		//	RVLDIF3VECTORS(uM, a, V3Tmp);

		//	e = RVLDOTPRODUCT3(V3Tmp, V3Tmp);

		//	E += (pFace->Area * e);

		//	//// e <- || RInit * a - uS - dphi x uS ||^2

		//	//RVLMULMX3X3VECT(RInit, a, V3Tmp);

		//	//RVLDIF3VECTORS(V3Tmp, uS, V3Tmp);

		//	//RVLCROSSPRODUCT3(dphi, uS, V3Tmp2);

		//	//RVLDIF3VECTORS(V3Tmp, V3Tmp2, V3Tmp);

		//	//e = RVLDOTPRODUCT3(V3Tmp, V3Tmp);

		//	//// E2 <- E2 + e

		//	//E2 += (pFace->Area * e);

		//	//// e <- || Rinit * a - dR * uS ||^2

		//	//RVLMULMX3X3VECT(RInit, a, V3Tmp);

		//	//RVLMULMX3X3VECT(dR, uS, V3Tmp2);

		//	//RVLDIF3VECTORS(V3Tmp, V3Tmp2, V3Tmp);

		//	//e = RVLDOTPRODUCT3(V3Tmp, V3Tmp);

		//	//// E2 <- E2 + e

		//	//E3 += (pFace->Area * e);
		//}

		//int debug = 0;

		///
	}	// for (k = 0; k < nIterations; k++)

	// Compute the fitting score.

	float score = 0.0f;

	for (j = 0; j < pConvexHull->iVisibleFaces.n; j++)
	{
		pFace = pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[j]];

		uS = pFace->N;

		if (piModelNormalArray)
		{
			RVLMULMX3X3TVECT(R, uS, uM);

			maxw = -2.0f;

			for (i = 0; i < piModelNormalArray->n; i++)
			{
				iNormal = piModelNormalArray->Element[i];

				a = sampledUnitSphere.Element + 3 * iNormal;

				w = RVLDOTPRODUCT3(a, uM);

				if (w > maxw)
				{
					maxw = w;

					iCorrespondence = iNormal;
				}
			}
		}
		else
			RVLVN_GET_CLOSEST_CONVEX_TEMPLATE_ELEMENT(convexTemplateLUT, uS, R, uM, i_, j_, k_, iCorrespondence);

		faceCorrespondence[j] = iCorrespondence;

		if (mask[iCorrespondence])
		{
			a = sampledUnitSphere.Element + 3 * iCorrespondence;

			RVLDIF3VECTORS(a, uM, V3Tmp);

			w = pFace->Area * exp(-kR * RVLDOTPRODUCT3(V3Tmp, V3Tmp));

			score += w;
		}
	}

	return score;
}

void VNClassifier::InitNormalAlignment(
	Array<MESH::Face *> faces,
	Array<int> iFaceArray,
	float *U)
{
	float invU[9];

	RVLNULLMX3X3(invU);

	int i;
	float M3x3Tmp[9];
	float area;
	float *u;
	float xx, yy, zz, xy, zx, yz;
	MESH::Face *pFace;

	for (i = 0; i < iFaceArray.n; i++)
	{
		pFace = faces.Element[iFaceArray.Element[i]];

		u = pFace->N;

		area = pFace->Area;

		xx = area * u[0] * u[0];
		yy = area * u[1] * u[1];
		zz = area * u[2] * u[2];
		xy = area * u[0] * u[1];
		yz = area * u[1] * u[2];
		zx = area * u[2] * u[0];

		RVLMXEL(M3x3Tmp, 3, 0, 0) = -(yy + zz);
		RVLMXEL(M3x3Tmp, 3, 0, 1) = xy;
		RVLMXEL(M3x3Tmp, 3, 0, 2) = -zx;
		RVLMXEL(M3x3Tmp, 3, 1, 1) = -(zz + xx);
		RVLMXEL(M3x3Tmp, 3, 1, 2) = yz;
		RVLMXEL(M3x3Tmp, 3, 2, 2) = -(xx + yy);

		RVLSUMMX3X3UT(invU, M3x3Tmp, invU);
	}

	float detU;

	RVLINVCOV3(invU, U, detU);

	RVLCOMPLETESIMMX3(U);
}

float VNClassifier::Align(
	Array<MESH::Face *> faceArray,
	Array<MESH::Face *> refFaceArray,
	float *R)
{
	float sigTh = 6.0f * DEG2RAD;

	int nIterations = 2;

	float varTh = sigTh * sigTh;

	float csThr = cos(3.0f * sigTh);

	int iFace;

	float totalRefArea = 0.0f;

	for (iFace = 0; iFace < refFaceArray.n; iFace++)
		totalRefArea += refFaceArray.Element[iFace]->Area;

	Array<int> iFaceArray;

	iFaceArray.Element = new int[faceArray.n];

	float *area = new float[faceArray.n];

	float totalArea = 0.0f;

	for (iFace = 0; iFace < faceArray.n; iFace++)
	{
		area[iFace] = faceArray.Element[iFace]->Area;

		totalArea += area[iFace];
	}

	int *iCorresp = new int[faceArray.n];

	float *uM_ = new float[3 * faceArray.n];

	float bestScore = 0.0f;

	int j, k, iR, iFace_, iCorresp_;
	float cs, maxcs, w, e, th, score;
	float *RInit, *uS, *uM, *a;
	float U[9], R_[9], R__[9], V3Tmp[3], V3Tmp2[3], b[3], dphi[3], z[3], dR[9];
	MESH::Face *pFace, *pFace_;

	for (iR = 0; iR < SO3Samples.c; iR++)
	{
		RInit = SO3Samples.Element + 9 * iR;

		RVLCOPYMX3X3(RInit, R_);

		for (k = 0; k <= nIterations; k++)
		{
			iFaceArray.n = 0;

			score = 0.0f;

			for (iFace = 0; iFace < faceArray.n; iFace++)
			{
				pFace = faceArray.Element[iFace];

				// uS <- normal of the j-th visible face

				uS = pFace->N;

				// uM <- normal of the j-th visible face transformed by R_

				uM = uM_ + 3 * iFace;

				RVLMULMX3X3VECT(R_, uS, uM);

				maxcs = csThr;

				iCorresp_ = -1;

				w = 0.0f;

				for (iFace_ = 0; iFace_ < refFaceArray.n; iFace_++)
				{
					pFace_ = refFaceArray.Element[iFace_];

					cs = RVLDOTPRODUCT3(uM, pFace_->N);

					if (cs >= csThr)
					{
						e = acos(maxcs);

						e *= e;

						w += (pFace_->Area * exp(-e / varTh));

						if (cs > maxcs)
						{
							maxcs = cs;

							iCorresp_ = iFace_;
						}
					}
				}
				
				iCorresp[iFace] = iCorresp_;

				if (iCorresp_ >= 0)
				{
					iFaceArray.Element[iFaceArray.n++] = iFace;

					pFace->Area = area[iFace] * w;

					score += pFace->Area;
				}
			}

			//if (iFaceArray.n == 76)
			//	int debug = 0;

			if (k == nIterations)
				break;

			InitNormalAlignment(faceArray, iFaceArray, U);

			RVLNULL3VECTOR(V3Tmp2);

			for (j = 0; j < iFaceArray.n; j++)
			{
				iFace = iFaceArray.Element[j];				

				pFace = faceArray.Element[iFace];

				uS = pFace->N;

				// uM <- normal of the j-th visible face transformed by R_

				uM = uM_ + 3 * iFace;

				// a <- the closest reference vector

				pFace_ = refFaceArray.Element[iCorresp[iFace]];

				a = pFace_->N;

				// V3Tmp2 <- V3Tmp2 + w * (uM x a), where w = the area of the j-th face

				RVLCROSSPRODUCT3(uM, a, V3Tmp);

				RVLSCALE3VECTOR(V3Tmp, pFace->Area, V3Tmp);

				RVLSUM3VECTORS(V3Tmp2, V3Tmp, V3Tmp2);
			}

			// b <- R_ * V3Tmp2

			RVLMULMX3X3TVECT(R_, V3Tmp2, b);

			// dphi = -inv(U) * b

			RVLMULMX3X3VECT(U, b, dphi);

			RVLNEGVECT3(dphi, dphi);

			// Only for debugging purpose!

			//float E = 0.0f;

			//float e;

			//for (j = 0; j < pConvexHull->iVisibleFaces.n; j++)
			//{
			//	pFace = pConvexHull->faces.Element[pConvexHull->iVisibleFaces.Element[j]];

			//	uS = pFace->N;

			//	RVLMULMX3X3TVECT(R, uS, uM);

			//	a = sampledUnitSphere.Element + 3 * faceCorrespondence[j];

			//	RVLDIF3VECTORS(uM, a, V3Tmp);

			//	e = RVLDOTPRODUCT3(V3Tmp, V3Tmp);

			//	E += (pFace->Area * e);
			//}

			///

			// dphi = || dphi ||

			th = sqrt(RVLDOTPRODUCT3(dphi, dphi));

			if (RVLABS(th) >= 1e-8)
			{
				// z <- dphi / || dphi ||

				RVLSCALE3VECTOR2(dphi, th, z);

				// dR <- Rot(z, th) (angle axis to rotation matrix)

				AngleAxisToRot<float>(z, th, dR);

				// R_ <- dR' * R_

				RVLMXMUL3X3T1(dR, R_, R__);

				RVLCOPYMX3X3(R__, R_);
			}
			else
				break;
		}	// for (k = 0; k < nIterations; k++)

		if (score > bestScore)
		{
			bestScore = score;

			RVLCOPYMX3X3(R_, R);
		}
	}	// for (iR = 0; iR < SO3Samples.c; iR++)

	for (iFace = 0; iFace < faceArray.n; iFace++)
		faceArray.Element[iFace]->Area = area[iFace];

	delete[] iFaceArray.Element;
	delete[] area;
	delete[] iCorresp;
	delete[] uM_;

	return bestScore / (totalArea * totalRefArea);
}

void VNClassifier::SetSceneFileName(char *sceneFileName_)
{
	RVLCopyString(sceneFileName_, &sceneFileName);
}

void VNClassifier::SaveDescriptor(
	FILE *fp,
	float *d,
	bool *bd,
	int iModel,
	int iMetaModel)
{
	VN *pModel = models[iMetaModel];

	fprintf(fp, "%d\t%d\t", iModel, iMetaModel);

	int i;

	for (i = 0; i < pModel->featureArray.n; i++)
		fprintf(fp, "%f\t", d[i]);

	for (i = 0; i < pModel->featureArray.n; i++)
		fprintf(fp, "%d\t", (int)(bd[i]));

	fprintf(fp, "\n");
}

void VNClassifier::SaveDescriptor(
	FILE *fp,
	float *d,
	uchar *bd,
	float *R,
	int iCluster,
	int type)
{
	float t[3];

	RVLNULL3VECTOR(t);

	fprintf(fp, "%d\t%d\t", type, iCluster);

	int i;

	for (i = 0; i < 9; i++)
		fprintf(fp, "%f\t", R[i]);

	for (i = 0; i < 3; i++)
		fprintf(fp, "%f\t", t[i]);

	for (i = 0; i < sampledUnitSphere.h; i++)
		fprintf(fp, "%f\t", d[i]);

	for (i = 0; i < sampledUnitSphere.h; i++)
		fprintf(fp, "%d\t", (int)(bd[i]));

	for (i = 0; i < sampledUnitSphere.h; i++)
		fprintf(fp, "%f\t", 0.0f);

	for (i = 0; i < 3; i++)
		fprintf(fp, "%f\t", 0.0f);

	fprintf(fp, "\n");
}

//Vidovic - commented on 3.10.2018 - CHECK with Cupec
/*void VNClassifier::SaveDescriptor(
	FILE *fp,
	float *d,
	uchar *bd,
	int iModel,
	int iMetaModel)
{
	VN *pModel = models[iMetaModel];

	fprintf(fp, "%d\t%d\t", iModel, iMetaModel);

	int i;

	for (i = 0; i < pModel->featureArray.n; i++)
		fprintf(fp, "%f\t", d[i]);

	for (i = 0; i < pModel->featureArray.n; i++)
		fprintf(fp, "%d\t", (int)(bd[i]));

	fprintf(fp, "\n");
}*/

void VNClassifier::SaveDescriptors(
	FILE *fp,
	float *d,
	uchar *bd,
	int iCluster,
	int type)
{
	float t[3];

	RVLNULL3VECTOR(t);

	float *d_ = d;
	uchar *bd_ = bd;
	float *R = SO3Samples.Element;
	float *R_ = R;

	int i, j;

	for (j = 0; j < SO3Samples.c; j++, d_ += sampledUnitSphere.h, bd_ += sampledUnitSphere.h, R_ += 9)
	{
		fprintf(fp, "%d\t%d\t", type, iCluster);

		for (i = 0; i < 9; i++)
			fprintf(fp, "%f\t", R_[i]);

		for (i = 0; i < 3; i++)
			fprintf(fp, "%f\t", t[i]);

		for (i = 0; i < sampledUnitSphere.h; i++)
			fprintf(fp, "%f\t", d_[i]);

		for (i = 0; i < sampledUnitSphere.h; i++)
			fprintf(fp, "%d\t", (int)(bd_[i]));

		for (i = 0; i < sampledUnitSphere.h; i++)
			fprintf(fp, "%f\t",0.0f);

		for (i = 0; i < 3; i++)
			fprintf(fp, "%f\t", 0.0f);

		fprintf(fp, "\n");
	}
}

void VNClassifier::SaveDescriptors(
	FILE *fp,
	float *d,
	uchar *bd,
	int nDescriptors,
	int iCluster,
	int type)
{
	float t[3];

	RVLNULL3VECTOR(t);

	float R[9];

	RVLUNITMX3(R)

	float *d_ = d;
	uchar *bd_ = bd;

	int i, j;

	for (j = 0; j < nDescriptors; j++, d_ += sampledUnitSphere.h, bd_ += sampledUnitSphere.h)
	{
		fprintf(fp, "%d\t%d\t", type, iCluster);

		for (i = 0; i < 9; i++)
			fprintf(fp, "%f\t", R[i]);

		for (i = 0; i < 3; i++)
			fprintf(fp, "%f\t", t[i]);

		for (i = 0; i < sampledUnitSphere.h; i++)
			fprintf(fp, "%f\t", d_[i]);

		for (i = 0; i < sampledUnitSphere.h; i++)
			fprintf(fp, "%d\t", (int)(bd_[i]));

		for (i = 0; i < sampledUnitSphere.h; i++)
			fprintf(fp, "%f\t", 0.0f);

		for (i = 0; i < 3; i++)
			fprintf(fp, "%f\t", 0.0f);

		fprintf(fp, "\n");
	}
}

void VNClassifier::SaveVertices(
	FILE *fp,
	RECOG::PSGM_::Cluster * pSCluster)
{
	int iVertex, nVertices;
	nVertices = pSCluster->iVertexArray.n;

	SURFEL::Vertex *pVertex;

	fprintf(fp, "%d\n", pSCluster->iVertexArray.n);
	
	for (iVertex = 0; iVertex < nVertices; iVertex++)
	{
		pVertex = pSurfels->vertexArray.Element[iVertex];

		fprintf(fp, "%f\t%f\t%f\n", pVertex->P[0], pVertex->P[1], pVertex->P[2]);
	}
}

void VNClassifier::SaveLatentVector(
	FILE *fp,
	float *q,
	int iClass)
{
	fprintf(fp, "%d\t", iClass);

	ClassData *pClass = classArray.Element + iClass;

	PrintMatrix(fp, q, 1, pClass->M.h);
}

void VNClassifier::LoadDescriptor(
	FILE *fp,
	float *&d,
	bool *&bd,
	int &iModel,
	int &iMetaModel)
{
	fscanf(fp, "%d\t%d\t", &iModel, &iMetaModel);

	VN *pModel = models[iMetaModel];

	d = new float[pModel->featureArray.n];

	int i;

	for (i = 0; i < pModel->featureArray.n; i++)
		fscanf(fp, "%f\t", d + i);

	bd = new bool[pModel->featureArray.n];

	int ibd;

	for (i = 0; i < pModel->featureArray.n; i++)
	{
		fscanf(fp, "%d\t", &ibd);

		bd[i] = (bool)ibd;
	}
}

void VNClassifier::LoadLatentVector(
	FILE *fp,
	float *&q,
	int &iClass)
{
	fscanf(fp, "%d\t", &iClass);

	ClassData *pClass = classArray.Element + iClass;

	q = new float[pClass->M.h];

	int i;

	for (i = 0; i < pClass->M.h; i++)
		fscanf(fp, "%f\t", q + i);
}

void VNClassifier::LoadShapeSpace(char *shapeSpaceFileName)
{
	FILE *fp = fopen(shapeSpaceFileName, "r");

	if (fp == NULL)
		return;

	bShapeSpace = true;

	char shapeSpaceFilePath[500];

	strcpy(shapeSpaceFilePath, shapeSpaceFileName);

	char *pFileName = strrchr(shapeSpaceFilePath, '\\') + 1;

	int iFeature, j, l;
	char line[500];
	ClassData *pClass;
	int iClass;
	char shapeSpaceFileName_[500];
	FILE *fp_;
	VN *pModel;
	float *N, *M_;

	while (true)
	{
		fgets(line, 500, fp);

		if (sscanf(line, "%d\t%s\n", &iClass, shapeSpaceFileName_) == 2)
		{
			if (iClass >= 0 && iClass < classArray.n)
			{
				pClass = classArray.Element + iClass;

				strcpy(pFileName, shapeSpaceFileName_);

				fp_ = fopen(shapeSpaceFilePath, "rb");

				fread(&(pClass->M.w), sizeof(int), 1, fp_);
				fread(&(pClass->M.h), sizeof(int), 1, fp_);
				pClass->M.Element = new float[pClass->M.w * pClass->M.h];
				fread(pClass->M.Element, sizeof(float), pClass->M.w * pClass->M.h, fp_);
				int m = pClass->M.h - 4;
				pClass->qMin = new float[m];
				fread(pClass->qMin, sizeof(float), m, fp_);
				pClass->qMax = new float[m];
				fread(pClass->qMax, sizeof(float), m, fp_);
				fread(&(pClass->nHull), sizeof(int), 1, fp_);
				pClass->A = new float[pClass->nHull * m];
				fread(pClass->A, sizeof(float), pClass->nHull * m, fp_);
				pClass->b = new float[pClass->nHull];
				fread(pClass->b, sizeof(float), pClass->nHull, fp_);
				fread(&(pClass->nq0), sizeof(int), 1, fp_);
				pClass->q0 = new float[pClass->nq0 * m];
				fread(pClass->q0, sizeof(float), pClass->nq0 * m, fp_);
				RVLNULLMX3X3(pClass->T);
				pModel = models[pClass->iMetaModel];
				for (iFeature = 0; iFeature < pModel->featureArray.n; iFeature++)
				{
					N = pModel->featureArray.Element[iFeature].N;

					M_ = pClass->M.Element + iFeature;

					for (l = 0; l < 3; l++)
						for (j = l; j < 3; j++)
							pClass->T[l * 3 + j] += (N[j] * M_[l * pModel->featureArray.n]);
				}			
				RVLCOMPLETESIMMX3(pClass->T);

				fclose(fp_);
			}
		}
		else if (strcmp(line, "end") == 0)
			break;
	}

	fclose(fp);
}

void VNClassifier::InitDisplay(
	Visualizer *pVisualizer,
	Mesh *pMesh,
	unsigned char *selectionColor)
{
	pVisualizer->normalLength = 10.0;

	pVisualizer->SetMesh(pMesh);

	visualizationData.pMesh = pMesh;
	visualizationData.pSurfels = pSurfels;
	visualizationData.pClassifier = this;
	visualizationData.pVisualizer = pVisualizer;
	RVLCOPY3VECTOR(selectionColor, visualizationData.selectionColor);
	visualizationData.iSelectedCluster = -1;
	pSurfels->DisplayData.keyPressUserFunction = &RECOG::VN_::keyPressUserFunction;
	pSurfels->DisplayData.mouseRButtonDownUserFunction = &RECOG::VN_::mouseRButtonDownUserFunction;
	pSurfels->DisplayData.vpUserFunctionData = &visualizationData;

	pSurfels->InitDisplay(pVisualizer, pMesh, pSurfelDetector);
}

void VNClassifier::ResetClusterColor(int iCluster)
{
	RECOG::ResetClusterColor(visualizationData.pVisualizer, visualizationData.pMesh, pSurfels, iCluster, SClusters, 
		visualizationData.clusterColor, clusterMap);
}

void VNClassifier::DisplayClusters()
{
	RECOG::DisplayClusters(visualizationData.pVisualizer, SClusters, visualizationData.pMesh, pSurfels, visualizationData.clusterColor, 
		clusterMap);
}

void VNClassifier::DisplaySelectedCluster(int iCluster)
{
	RECOG::DisplaySelectedCluster(visualizationData.pVisualizer, visualizationData.pMesh, pSurfels, iCluster, SClusters,
		visualizationData.iSelectedCluster, clusterMap, visualizationData.selectionColor, visualizationData.clusterColor);
		}

void VN_::_3DNetDatabaseClasses(VNClassifier *pClassifier)
{
	pClassifier->classArray.n = 10;

	pClassifier->classArray.Element = new RECOG::ClassData[pClassifier->classArray.n];

	RECOG::ClassData *pClass;

	// class apple
	pClass = pClassifier->classArray.Element;
	pClass->iMetaModel = RVLVN_METAMODEL_CONVEX;
	pClass->iFirstInstance = 0;
	pClass->nInstances = 12;
	pClass->iRefInstance = pClass->iFirstInstance; //by default


	pClass->VNGraspNormals_.n = 1;
	pClass->VNGraspNormals_.Element = new RECOG::VNGraspNormals[pClass->VNGraspNormals_.n];
	pClass->VNGraspNormals_.Element[0].n1 = 11;//parallel planes for gripping
	pClass->VNGraspNormals_.Element[0].n2 = 44;//parallel planes for gripping
	pClass->VNGraspNormals_.Element[0].n3 = 15;//adjacent planes
	pClass->VNGraspNormals_.Element[0].n4 = 18;//adjacent planes
	pClass->VNGraspNormals_.Element[0].n5 = 0;


	// class banana
	pClass = pClassifier->classArray.Element + 1;
	//pClass->iMetaModel = RVLVN_METAMODEL_BANANA;
	pClass->iMetaModel = RVLVN_METAMODEL_TORUS;
	pClass->iFirstInstance = 12;
	pClass->nInstances = 6;
	pClass->iRefInstance = 17;

	pClass->VNGraspNormals_.n = 1;
	pClass->VNGraspNormals_.Element = new RECOG::VNGraspNormals[pClass->VNGraspNormals_.n];
	pClass->VNGraspNormals_.Element[0].n1 = 55;
	pClass->VNGraspNormals_.Element[0].n2 = 72;
	pClass->VNGraspNormals_.Element[0].n3 = 50;
	pClass->VNGraspNormals_.Element[0].n4 = 57;
	pClass->VNGraspNormals_.Element[0].n5 = 0;


	// class bottle
	pClass = pClassifier->classArray.Element + 2;
	pClass->iMetaModel = RVLVN_METAMODEL_BOTTLE;
	pClass->iFirstInstance = 18;
	pClass->nInstances = 74;
	pClass->iRefInstance = pClass->iFirstInstance;

	pClass->VNGraspNormals_.n = 1;
	pClass->VNGraspNormals_.Element = new RECOG::VNGraspNormals[pClass->VNGraspNormals_.n];
	pClass->VNGraspNormals_.Element[0].n1 = 87;
	pClass->VNGraspNormals_.Element[0].n2 = 72;
	pClass->VNGraspNormals_.Element[0].n3 = 89;
	pClass->VNGraspNormals_.Element[0].n4 = 91;
	pClass->VNGraspNormals_.Element[0].n5 = 61;

	// class bowl
	pClass = pClassifier->classArray.Element + 3;
	pClass->iMetaModel = RVLVN_METAMODEL_BOWL;
	pClass->iFirstInstance = 92;
	pClass->nInstances = 31;
	pClass->iRefInstance = pClass->iFirstInstance;

	pClass->VNGraspNormals_.n = 1;
	pClass->VNGraspNormals_.Element = new RECOG::VNGraspNormals[pClass->VNGraspNormals_.n];
	//pClass->VNGraspNormals_.Element[0].n1 = 0;
	//pClass->VNGraspNormals_.Element[0].n2 = 0;
	//pClass->VNGraspNormals_.Element[0].n3 = 0;
	//pClass->VNGraspNormals_.Element[0].n4 = 0;
	//pClass->VNGraspNormals_.Element[0].n5 = 0;



	// class car
	pClass = pClassifier->classArray.Element + 4;
	pClass->iMetaModel = RVLVN_METAMODEL_CONVEX;
	pClass->iFirstInstance = 123;
	pClass->nInstances = 72;
	//pClass->iRefInstance = 172;
	pClass->iRefInstance = 123;

	pClass->VNGraspNormals_.n = 1;
	pClass->VNGraspNormals_.Element = new RECOG::VNGraspNormals[pClass->VNGraspNormals_.n];
	pClass->VNGraspNormals_.Element[0].n1 = 55;
	pClass->VNGraspNormals_.Element[0].n2 = 22;
	pClass->VNGraspNormals_.Element[0].n3 = 58;
	pClass->VNGraspNormals_.Element[0].n4 = 61;
	pClass->VNGraspNormals_.Element[0].n5 = 0;



	// class donut
	pClass = pClassifier->classArray.Element + 5;
	pClass->iMetaModel = RVLVN_METAMODEL_TORUS;
	pClass->iFirstInstance = 195;
	pClass->nInstances = 10;
	pClass->iRefInstance = pClass->iFirstInstance;

	pClass->VNGraspNormals_.n = 1;
	pClass->VNGraspNormals_.Element = new RECOG::VNGraspNormals[pClass->VNGraspNormals_.n];
	//pClass->VNGraspNormals_.Element[0].n1 = 55;
	//pClass->VNGraspNormals_.Element[0].n2 = 22;
	//pClass->VNGraspNormals_.Element[0].n3 = 58;
	//pClass->VNGraspNormals_.Element[0].n4 = 61;
	//pClass->VNGraspNormals_.Element[0].n5 = 0;


	// class hammer
	pClass = pClassifier->classArray.Element + 6;
	pClass->iMetaModel = RVLVN_METAMODEL_BOTTLE;
	pClass->iFirstInstance = 205;
	pClass->nInstances = 36;
	pClass->iRefInstance = pClass->iFirstInstance;

	pClass->VNGraspNormals_.n = 1;
	pClass->VNGraspNormals_.Element = new RECOG::VNGraspNormals[pClass->VNGraspNormals_.n];
	pClass->VNGraspNormals_.Element[0].n1 = 99;
	pClass->VNGraspNormals_.Element[0].n2 = 85;
	pClass->VNGraspNormals_.Element[0].n3 = 89;
	pClass->VNGraspNormals_.Element[0].n4 = 26;
	pClass->VNGraspNormals_.Element[0].n5 = 92;


	// class mug
	pClass = pClassifier->classArray.Element + 7;
	pClass->iMetaModel = RVLVN_METAMODEL_MUG;
	pClass->iFirstInstance = 241;
	pClass->nInstances = 75;
	pClass->iRefInstance = 245;

	pClass->VNGraspNormals_.n = 1;
	pClass->VNGraspNormals_.Element = new RECOG::VNGraspNormals[pClass->VNGraspNormals_.n];
	pClass->VNGraspNormals_.Element[0].n1 = 11;
	pClass->VNGraspNormals_.Element[0].n2 = 92;
	pClass->VNGraspNormals_.Element[0].n3 = 15;
	pClass->VNGraspNormals_.Element[0].n4 = 18;
	pClass->VNGraspNormals_.Element[0].n5 = 0;
	
	

	// class tetra pak
	pClass = pClassifier->classArray.Element + 8;
	pClass->iMetaModel = RVLVN_METAMODEL_CONVEX;
	pClass->iFirstInstance = 316;
	pClass->nInstances = 26;
	pClass->iRefInstance = pClass->iFirstInstance;

	pClass->VNGraspNormals_.n = 1;
	pClass->VNGraspNormals_.Element = new RECOG::VNGraspNormals[pClass->VNGraspNormals_.n];
	//pClass->VNGraspNormals_.Element[0].n1 = 55;
	//pClass->VNGraspNormals_.Element[0].n2 = 22;
	//pClass->VNGraspNormals_.Element[0].n3 = 58;
	//pClass->VNGraspNormals_.Element[0].n4 = 61;
	//pClass->VNGraspNormals_.Element[0].n5 = 0;


	// class toilet paper
	pClass = pClassifier->classArray.Element + 9;
	//pClass->iMetaModel = RVLVN_METAMODEL_PIPE;
	pClass->iMetaModel = RVLVN_METAMODEL_BOWL;
	pClass->iFirstInstance = 342;
	pClass->nInstances = 9;
	pClass->iRefInstance = pClass->iFirstInstance;

	pClass->VNGraspNormals_.n = 1;
	pClass->VNGraspNormals_.Element = new RECOG::VNGraspNormals[pClass->VNGraspNormals_.n];
	//pClass->VNGraspNormals_.Element[0].n1 = 55;
	//pClass->VNGraspNormals_.Element[0].n2 = 22;
	//pClass->VNGraspNormals_.Element[0].n3 = 58;
	//pClass->VNGraspNormals_.Element[0].n4 = 61;
	//pClass->VNGraspNormals_.Element[0].n5 = 0;

	// common values for all classes

	int iClass;

	for (iClass = 0; iClass < pClassifier->classArray.n; iClass++)
	{
		pClass = pClassifier->classArray.Element + iClass;

		pClass->M.Element = NULL;
		pClass->qMax = NULL;
		pClass->qMin = NULL;
		pClass->A = NULL;
		pClass->b = NULL;
		pClass->q0 = NULL;
		pClass->valid = NULL;
		pClass->nq0 = 1;
	}

	// create VN model for every class.
	if (pClassifier->bTrainingStructure)
	{
		pClassifier->models = pClassifier->metaModels;
	}
	else
	{
		int i;
		for (int iClass = 0; iClass < pClassifier->classArray.n; iClass++)
		{
			pClass = pClassifier->classArray.Element + iClass;
			i = 0;

			std::string validFeaturesFileName = std::string(pClassifier->validFeaturesFileName) + std::to_string(iClass) + ".txt";
			std::ifstream infile(validFeaturesFileName);
			std::string unused;
			while (std::getline(infile, unused))
				++i;
			infile.close();
			infile.open(validFeaturesFileName);
			pClass->valid = new int[i];
			for (int line = 0; line < i; line++)
			{
				infile >> pClass->valid[line];
			}
			infile.close();
			VN* pModel = new VN;
			VN_::Create(pClassifier->metaModels[pClass->iMetaModel], pClass->valid, pClassifier->alignment.convexTemplate66, pClassifier->pMem0, pModel);	
			pClassifier->models.push_back(pModel);
			pClass->iMetaModel = iClass;
			

			char metaModelFileName[] = "metamodel0.dat";

			FILE *fp;

			sprintf(metaModelFileName + strlen(metaModelFileName) - 5, "%1d.dat", iClass);
			fp = fopen(metaModelFileName, "w");

			pModel->SaveFeatures(fp);

			fclose(fp);
		}
		
		
	}



	// special property of class mug

	pClass = pClassifier->classArray.Element + 7;

	pClass->nq0 = 2;
	//pClass->q0 = new float[pClass->nq0 * ];

}

void VN_::OSDDatabaseClasses(VNClassifier *pClassifier)
{
	pClassifier->classArray.n = 4;

	pClassifier->classArray.Element = new RECOG::ClassData[pClassifier->classArray.n];

	RECOG::ClassData *pClass, *pMug;
	int i, j;
	float z;

	// class mug

	pClass = pClassifier->classArray.Element + 0;
	pClass->iMetaModel = RVLVN_METAMODEL_CONVEX;
	//pClass->iFirstInstance = 0;
	//pClass->nInstances = 12;
	//pClass->iRefInstance = pClass->iFirstInstance; //by default
	pClass->bConvex = false;
	pClass->bAxisAlignment = false;
#ifdef RVLVN_CLASSIFY_CTI_VERSION1B
	pClass->faces.n = 6;
	pClass->faces.Element = new int[pClass->faces.n];
	for (j = 0; j < 6; j++)
		pClass->faces.Element[j] = 11 * j;
#else
	pClass->faces.n = 18;
	pClass->faces.Element = new int[pClass->faces.n];
	pClass->faces.Element[0] = 0;
	pClass->faces.Element[1] = 33;
	j = 2;
	for (i = 0; i < pClassifier->sampledUnitSphere.h; i++)
	{
		z = pClassifier->sampledUnitSphere.Element[3 * i + 2];

		if (RVLABS(z) < 1e-6)
			pClass->faces.Element[j++] = i;
	}
#endif
	pMug = pClass;
	pClass->iHypothesisBlocks.Element = pClass->iHypotheisisBlocksMem;
	pClass->iHypothesisBlocks.h = 6;
	pClass->iHypothesisBlocks.w = 1;
	for (i = 0; i < 6; i++)
		pClass->iHypothesisBlocks.Element[i] = i;

	// class bowl

	pClass = pClassifier->classArray.Element + 1;
	pClass->iMetaModel = RVLVN_METAMODEL_CONVEX;
	pClass->bConvex = false;
	pClass->bAxisAlignment = true;
#ifdef RVLVN_CLASSIFY_CTI_VERSION1B
	pClass->faces.n = 6;
	pClass->faces.Element = new int[pClass->faces.n];
	for (j = 0; j < 6; j++)
		pClass->faces.Element[j] = 11 * j;
#else
	pClass->faces.n = 2;
	pClass->faces.Element = new int[pClass->faces.n];
	pClass->faces.Element[0] = 0;
	pClass->faces.Element[1] = 33;
#endif
	pClass->iHypothesisBlocks.Element = pClass->iHypotheisisBlocksMem;
	pClass->iHypothesisBlocks.h = 2;
	pClass->iHypothesisBlocks.w = 3;
	for (i = 0; i < 6; i++)
		pClass->iHypothesisBlocks.Element[i] = i;

	// class box

	pClass = pClassifier->classArray.Element + 2;
	pClass->iMetaModel = RVLVN_METAMODEL_CONVEX;
	pClass->bConvex = true;
	pClass->bAxisAlignment = false;
	pClass->faces.n = 6;
	pClass->faces.Element = new int[pClass->faces.n];
	for (j = 0; j < 6; j++)
		pClass->faces.Element[j] = 11 * j;
	pClass->iHypothesisBlocks.Element = pClass->iHypotheisisBlocksMem;
	pClass->iHypothesisBlocks.h = 1;
	pClass->iHypothesisBlocks.w = 6;
	for (i = 0; i < 6; i++)
		pClass->iHypothesisBlocks.Element[i] = i;

	// class cylinder

	pClass = pClassifier->classArray.Element + 3;
	pClass->iMetaModel = RVLVN_METAMODEL_CONVEX;
	pClass->bConvex = false;
	pClass->bAxisAlignment = false;
#ifdef RVLVN_CLASSIFY_CTI_VERSION1B
	pClass->faces.n = 6;
	pClass->faces.Element = new int[pClass->faces.n];
	for (j = 0; j < 6; j++)
		pClass->faces.Element[j] = 11 * j;
#else
	pClass->faces.n = pMug->faces.n;
	pClass->faces.Element = new int[pClass->faces.n];
	memcpy(pClass->faces.Element, pMug->faces.Element, pClass->faces.n * sizeof(int));
#endif
	pClass->iHypothesisBlocks.Element = pClass->iHypotheisisBlocksMem;
	pClass->iHypothesisBlocks.h = 3;
	pClass->iHypothesisBlocks.w = 2;
	for (i = 0; i < 3; i++)
	{
		pClass->iHypothesisBlocks.Element[pClass->iHypothesisBlocks.w * i] = i;
		pClass->iHypothesisBlocks.Element[pClass->iHypothesisBlocks.w * i + 1] = i + 3;
	}

	// common values for all classes

	int iClass;

	for (iClass = 0; iClass < pClassifier->classArray.n; iClass++)
	{
		pClass = pClassifier->classArray.Element + iClass;

		pClass->M.Element = NULL;
		pClass->qMax = NULL;
		pClass->qMin = NULL;
		pClass->q0 = NULL;
		pClass->nq0 = 1;
		pClass->VNGraspNormals_.Element = NULL;
		pClass->valid = NULL;
	}

	// Copy metamodels to models.

	pClassifier->models = pClassifier->metaModels;
}

bool VN_::keyPressUserFunction(
	Mesh *pMesh,
	SurfelGraph *pSurfels,
	std::string &key,
	void *vpData
	)
{
	VN_::VisualizationData *pData = (VN_::VisualizationData *)vpData;

	VNClassifier *pClassifier = pData->pClassifier;
	Visualizer *pVisualizer = pData->pVisualizer;

	if (key == "i")
	{
		printf("Enter cluster identifier (0-%d): ", pClassifier->SClusters.n);

		int iSelectedCluster;

		scanf("%d", &iSelectedCluster);

		if (iSelectedCluster >= 0 && iSelectedCluster < pClassifier->SClusters.n)
		{
			pClassifier->DisplaySelectedCluster(iSelectedCluster);

			if (iSelectedCluster >= 0)
			{
				pData->iSelectedCluster = iSelectedCluster;

				return true;
			}
			else
				return false;
		}
	}

	return false;
}

bool VN_::mouseRButtonDownUserFunction(
	Mesh *pMesh,
	SurfelGraph *pSurfels,
	int iSelectedPt,
	int iSelectedSurfel,
	void *vpData)
{
	VN_::VisualizationData *pData = (VN_::VisualizationData *)vpData;

	if (!pData->bClusters)
		return false;

	VNClassifier *pClassifier = pData->pClassifier;

	int iSelectedCluster = pClassifier->clusterMap[iSelectedSurfel];

	pClassifier->DisplaySelectedCluster(iSelectedCluster);

	if (iSelectedCluster >= 0)
	{
		pData->iSelectedCluster = iSelectedCluster;

		return true;
	}
	else
		return false;

	//unsigned char color[3];
	//SceneCluster *pCluster;
	//PSGM_::Cluster *pCluster_;

	//if (pData->iSelectedCluster >= 0)
	//{
	//	pClassifier->ResetClusterColor(pData->iSelectedCluster);

	//	if (pSurfels->DisplayData.bVertices)
	//	{
	//		RVLSET3VECTOR(color, 255, 0, 0);

	//		pCluster = pClassifier->SClusters.Element + pData->iSelectedCluster;

	//		pCluster_ = (PSGM_::Cluster *)(pCluster->vpCluster);

	//		pSurfels->PaintVertices(&(pCluster_->iVertexArray), color);
	//	}
	//}

	//

	//if (iCluster >= 0)
	//{
	//	printf("Selected segment: %d\n", iCluster);

	//	pCluster = pClassifier->SClusters.Element + iCluster;

	//	pCluster_ = (PSGM_::Cluster *)(pCluster->vpCluster);

	//	pSurfels->PaintSurfels(pMesh, pVisualizer, pCluster_->iSurfelArray, pData->selectionColor);

	//	if (pSurfels->DisplayData.bVertices)
	//	{
	//		RVLSET3VECTOR(color, 255, 255, 0);

	//		pCluster = pClassifier->SClusters.Element + iCluster;

	//		pCluster_ = (PSGM_::Cluster *)(pCluster->vpCluster);

	//		pSurfels->PaintVertices(&(pCluster_->iVertexArray), color);
	//	}

	//	pData->iSelectedCluster = iCluster;

	//	return true;
	//}
	//else
	//	return false;
}

//NOT FINISHED - MOVED TO PSGM.cpp
//bool VNClassifier::ReferenceFrames(Mesh convexHull)
//{
//	int i, j;
//	float score;
//	float dotProduct, normi, normj;
//	Array<convexHullFacesPairs> convexHullFacesPairArr;
//
//	convexHullFacesPairArr.n = convexHull.faces.n * convexHull.faces.n / 2;
//	convexHullFacesPairArr.Element = new convexHullFacesPairs[convexHullFacesPairArr.n];
//	int br = 0;
//
//	for (i = 0; i < convexHull.faces.n; i++)
//		for (j = i+1; j < convexHull.faces.n; j++)
//		{
//			dotProduct = RVLDOTPRODUCT3(convexHull.faces.Element[i]->N, convexHull.faces.Element[j]->N);
//			if (RVLABS(dotProduct) < 0.7)
//			{
//				score = convexHull.faces.Element[i]->Area * convexHull.faces.Element[j]->Area;
//				convexHullFacesPairArr.Element[br].cost = score;
//				convexHullFacesPairArr.Element[br].faces->a = i;
//				convexHullFacesPairArr.Element[br].faces->b = j;
//				br++;
//			}
//		}
//
//	convexHullFacesPairArr.n = br;
//
//	BubbleSort<convexHullFacesPairs>(convexHullFacesPairArr, true);
//
//	int nBestFacesPairs = 50;
//	float theta;
//	Array<convexHullFacesPairs> convexHullFacesPairArr_;
//	float R[9], v[3], x[3], y[3], z[3];
//
//	convexHullFacesPairArr_.n = br;
//	convexHullFacesPairArr_.Element = new convexHullFacesPairs[convexHullFacesPairArr_.n];
//	br = 0;
//	for (i = 0; i < nBestFacesPairs; i++)
//	{
//
//		z[0] = convexHull.faces.Element[convexHullFacesPairArr.Element[i].faces->a]->N[0];
//		z[1] = convexHull.faces.Element[convexHullFacesPairArr.Element[i].faces->a]->N[1];
//		z[2] = convexHull.faces.Element[convexHullFacesPairArr.Element[i].faces->a]->N[2];
//
//
//		RVLCROSSPRODUCT3(z, convexHull.faces.Element[convexHullFacesPairArr.Element[i].faces->b]->N, x);
//		RVLCROSSPRODUCT3(z, x, y);
//
//		R[0] = x[0];
//		R[3] = x[1];
//		R[6] = x[2];
//
//		R[1] = y[0];
//		R[4] = y[1];
//		R[7] = y[2];
//
//		R[2] = z[0];
//		R[5] = z[1];
//		R[9] = z[2];
//
//		GetAngleAxis(R, v, theta);
//
//		if (theta > 0.35)
//		{
//			convexHullFacesPairArr_.Element[br] = convexHullFacesPairArr.Element[i];
//			br++;
//		}
//
//	}
//
//}
